# Helper methods around auto-generated pstnconfig module

# Standard modules
import sys

# Impl. modules
import pstnconfig as m
import helpersConfiguration as config
import logging
from netborderPyInfra import PyGlobalConfig as globalConfig
from netborderPyInfra import *
log = logging.getLogger("pstnconfig")
##############################################################################
# Helpers for registering IDs for model entities used in PSTN Editor
#
# TODO put in a class/belongs to the controller
#
# ExtJS Tree items must have a unique ID - also used to retrieve associated
#   object
# ID generator using a 1-based count() iterator

import itertools
idgen = itertools.count(1)

# Map of unique id/model entities
_idToModelEntityDict = {}

# Register object using an ID for later edition from TreeView of Editor
# @returns id
def registerObjectForEditor(o):
  # If the object exists in the list, return its current id instead of creating
  # a new one.
  if o in _idToModelEntityDict.values():
    return getObjectId(o)
  else:
    id = idgen.next()
    _idToModelEntityDict[id] = o
    return id

# Scans the object dictionary _idToModelEntityDict to find the parameter object's
# id.
def getObjectId(o):
  for k,v in _idToModelEntityDict.items():
    if v == o:
      return k
  return None

# Get object from Id assigend in Editor tree
def getObjectFromEditorId(id):
  # Converted to int as passed as str to controller method
  return _idToModelEntityDict[int(id)]

# Deletes an object from the dicitonary
def deleteObjectFromEditor(o):
  objID = getObjectId(o)

  if not objID is None:
    del _idToModelEntityDict[objID]

def getNextBoardID(pstnConfig):
  id = 0
  for board in pstnConfig.get_boards().get_sangomaAnalogBoard() + \
      pstnConfig.get_boards().get_sangomaDigitalBoard() + \
      pstnConfig.get_boards().get_sangomaVirtualDigitalBoard():
    if board.get_ID() > id:
      id = board.get_ID()

  return id + 1

def getNextInterfaceID(pstnConfig, boardID):
  board = getBoardObject(pstnConfig,boardID)

  if board:
    return len(board.get_interface()) + 1

  return 1

def getNextWanpipeInterface(pstnConfig, boardID):
  wanpipeIf = 0

  board = getBoardObject(pstnConfig, boardID)

  if board:
    for interface in board.get_interface():
      if interface.get_wanpipeInterface() > wanpipeIf:
        wanpipeIf = interface.get_wanpipeInterface()

  return wanpipeIf + 1


def getNextWanpipe(pstnConfig):
  wanpipe = 0
  for board in pstnConfig.get_boards().get_sangomaDigitalBoard():
    for interface in board.get_interface():
      if interface.get_wanpipe() > wanpipe:
        wanpipe = interface.get_wanpipe()
  for board in pstnConfig.get_boards().get_sangomaAnalogBoard():
    if board.get_wanpipe() > wanpipe:
      wanpipe = board.get_wanpipe()

  return wanpipe + 1

def getNextPeerVirtualInterfaceID(pstnConfig, boardID):
  nextPeerID = 0

  board = getBoardObject(pstnConfig, boardID)

  if board:
    for interface in board.get_interface():
      if interface.get_peerVirtualInterfaceID() > nextPeerID:
        nextPeerID = interface.get_peerVirtualInterfaceID()

  return nextPeerID + 1

def getLastNonVirtualBoard(pstnConfig):
  lastID = getNextBoardID(pstnConfig) - 1
  lastBoard = getBoardObject(pstnConfig,lastID)
  if lastBoard:
    boardType = boardsModelInfo[lastBoard.get_model()]['type']
    while boardType == 'virtual_digital':
      lastID = lastID - 1
      lastBoard = getBoardObject(pstnConfig,lastID)
      if lastBoard is None:
        break
      boardType = boardsModelInfo[lastBoard.get_model()]['type']

  return lastBoard

def getNextPhysConfID(pstnConfig):
  id = 0;
  for config in pstnConfig.get_physicalInterfaceConfigs().get_sangomaAnalogConfig():
    if config.get_ID() > id:
      id = config.get_ID()
  for config in pstnConfig.get_physicalInterfaceConfigs().get_sangomaDigitalConfig():
    if config.get_ID() > id:
      id = config.get_ID()

  return id + 1

def getNextISDNConfID(pstnConfig):
  id = 0;
  isdn = pstnConfig.get_callControl().get_isdn()
  if isdn:
    for config in isdn.get_fasConfig():
      if config.get_ID() > id:
        id = config.get_ID()
    for config in isdn.get_nfasGroup():
      if config.get_ID() > id:
        id = config.get_ID()
    for config in isdn.get_briConfig():
      if config.get_ID() > id:
        id = config.get_ID()
  

  return id + 1
def getNextCASConfID(pstnConfig):
  id = 0;
  cas = pstnConfig.get_callControl().get_cas()
  if cas:
    for config in cas.get_digitalR2Config():
      if config.get_ID() > id:
        id = config.get_ID()
  return id + 1

def getNextAnalogConfID(pstnConfig):
  id = 0;
  analog = pstnConfig.get_callControl().get_analog()
  if analog:
    for config in analog.get_analogFxoConfig():
      if config.get_ID() > id:
        id = config.get_ID()
    for config in analog.get_analogFxsConfig():
      if config.get_ID() > id:
        id = config.get_ID()

  return id + 1

###############################################################################
# Info for each supported board model
boardsModelInfo = { \
    'A101': {'type': 'digital', 'numInterfaces': 1, 'digitalType':'PRI'},
    'A102': {'type': 'digital', 'numInterfaces': 2, 'digitalType':'PRI'},
    'A104': {'type': 'digital', 'numInterfaces': 4, 'digitalType':'PRI'},
    'A108': {'type': 'digital', 'numInterfaces': 8, 'digitalType':'PRI'},
    'A116': {'type': 'digital', 'numInterfaces': 8, 'digitalType':'PRI'},
    'A500': {'type': 'digital', 'numInterfaces': 6, 'digitalType':'BRI'},
    'B500': {'type': 'digital', 'numInterfaces': 4, 'digitalType':'BRI'},
    'B700_digital': {'type': 'digital', 'numInterfaces': 4, 'digitalType':'BRI'},
    'A200': {'type': 'analog'},
    'A400': {'type': 'analog'},
    'B600': {'type': 'analog'},
    'B700_analog': {'type': 'analog'},
    'VIRTUAL_DIGITAL': {'type': 'virtual_digital'},
    }

###############################################################################
# Default settings for some configuration objects

defaultEchoCancellation = {
    'typexx':'HARDWARE',
    'mode':'NORMAL',
    'comfortNoiseMode':'NORMAL',
    'tailDisplacementInMs':'0',
    'doubleTalkBehavior':'OPTIMAL'
    }
defaultAcousticEchoCancellation = {
    'typexx':"HARDWARE", 'mode':"DISABLED"
    }

#defaultLevelControl = {
#    'typexx':"HARDWARE",
#    'direction':"ALL",
#    'mode':"DISABLED"
#    }
defaultLevelControl = config.levelControlCfg

defaultBackgroundNoiseAttenuation = {
    'typexx':"HARDWARE",
    'direction':"FROM_PSTN",
    'mode':"DISABLED",
    'attenuationIndB':"-18"
    }
defaultDTMFRemoval = {
    'typexx':"HARDWARE",
    'direction':"FROM_PSTN",
    'mode':"DISABLED"
    }

defaultInboundParams = {
  'connect':{
    'condition':'ON-OUTBOUNDSIDE-CONNECTED'
  },
  'callerID':{
    'typexx':'BT-Bellcore-FSK-SDMF-MDMF'
  },
  'notifyOutboundSide':{
    'condition':'ON-CALLER-ID',
    'delayInMs':0
  }
}

defaultOutboundParams = {
  'dial':{
    'DTMFDigitDurationInMs':200,
    'interDigitDelayInMs':100,
    'active':'YES',
    'waitForDialTone':'YES',
    'preDialDelayInMs':100,
    'InternationalDialingSpecialCharacter':'+',
    'InternationalDialingSpecialCharacterSubstitute':'011'
  },
  'connectOutboundSide':{
    'condition':'AFTER-DIALING-DESTINATION-NUMBER'
  },
  'connect':{
    'condition':'ON-OUTBOUNDSIDE-CONNECTED',
    'delayInMs':0
  }
}

defaultDisconnectParams = {
  'ringCadenceDetector':{
    'active':'YES',
    'ringCadenceOffDelayMs':7000
  },
  'detectors':[
    {
      'name':'dial',
      'sequence':'dial'
    },
    {
      'name':'busy',
      'sequence':'busy'
    },
    {
      'name':'reorder',
      'sequence':'reorder'
    },
    {
      'name':'disconnect',
      'sequence':'disconnect'
    },
    {
      'name':'batteryDisconnect',
      'sequence':'batteryDisconnect'
    },
    {
      'name':'batteryPolarityInversion',
      'sequence':'batteryPolarityInversion'
    }
  ]
}

defaultT1Config = {
  'lineEncoding':'B8ZS',
  'clocking':'TERMINAL',
  'loopbackMode':'DISABLED',
  'framing':'ESF',
  'LBO':'0dB'
}

defaultAudioParams = {
  'defaultPcmLaw':'PCMU',
  'voicePacketLengthInMs':config.voicePacketLength,
  'toneSetName':'USA',
  'optimalBchannelTxQueueLevel':config.optimalTxBufferLevel
}

defaultBriAudioParams = {
  'defaultPcmLaw':'PCMA',
  'voicePacketLengthInMs':config.voicePacketLength,
  'toneSetName':'USA',
  'optimalBchannelTxQueueLevel':config.optimalTxBufferLevel
}

defaultFasConfig = {
  'displayName':'TODO',
  'termination':'TERMINAL',
  'isdnT310Timer':30000,
  'isdnT203Timer':30000,
  'isdnT302Timer':15000,
  'switchVariant':'5ESS',
  'BChannelNegotiation':'EXCLUSIVE',
  'initiateRestartProcedure':'YES',
  'inbandProgressTonesGeneration':'AS-NEEDED',
  'inbandProgressTonesIndicator':'PROGRESS',
  'callerNameLocation':'IN-SETUP-MSG-USING-DISPLAY-IE',
  'waitFacilityDelayMs':100,
  'overlapDigitsReception':'DISABLED'
}

defaultFxsParams = {
  'regretTimeMs':500,
  'oosBehavior':"REORDER"
}

defaultFxsInboundParams = {
  'minExpectedDigits':4,
  'busyToneDurationMs':5000,
  'interDigitTimeoutMs':4000,
  'dialToneDurationMs':10000,
  'relayEarlyMedia': {
    'value':"ENABLED"
  }
}

defaultFXSDisconnectAnswerNotificationParams = {
  'mode' : 'POLARITY-REVERSAL'
}

defaultHotlineDialingParams = {
  'enable':"NO"
}

###############################################################################
# Writes the pstnConfig as XML to a "file" subclass
def writeConfigToFileAsXML(pstnConfig, outfile):
  outfile.write('<?xml version="1.0" ?>\n')
  pstnConfig.export(outfile, 0)

###############################################################################
# Creates a pstnConfig object with defaults for Sangoma, including a
#   default BIDIR channel group
# @returns pstnConfig object
def createDefaultSangomaPstnConfig():
  # providers
  bChannelsThreads = m.BChannelsThreads.factory(
      count=16, priority="TIME_CRITICAL")
  dChannelsThreads = m.DChannelsThreads.factory(
      count=1, priority="NORMAL")
  sangoma = m.sangoma.factory(BChannelsThreads=bChannelsThreads,
      DChannelsThreads=dChannelsThreads)
  providers = m.providers.factory(sangoma=sangoma)
  # callControl
  channelGroup = m.channelGroup.factory(ID="default", direction="BIDIR",
      outboundHuntingScheme="REVERSE_LINEAR")
  # NOTE: the "channelGroup" property of the generated "channelGroups" class
  #   is a - list - of "channelGroup" objects
  channelGroups = m.channelGroups.factory(channelGroup=[channelGroup])
  callControl = m.callControl.factory(channelGroups=channelGroups)
  # pstncOnfig
  return m.pstnConfig.factory(version='4.4',
      providers=providers,
      boards=m.boards.factory(),
      physicalInterfaceConfigs=m.physicalInterfaceConfigs.factory(),
      callControl=callControl)

###############################################################################
# Helper that will append a new interface to the interface list
def addInterface(interfaceList,
                 ID,
                 displayName="",
                 wanpipe=None,
                 wanpipeInterface=None,
                 peerVirtualInterfaceID=None,
                 type=None,
                 phoneNumber=None,
                 callerName=None,
                 briModType=None):

  interfaceList.append(
    m.interface.factory(
      ID=ID,
      displayName=displayName,
      wanpipe=wanpipe,
      wanpipeInterface=wanpipeInterface,
      peerVirtualInterfaceID=peerVirtualInterfaceID,
      typexx=type,
      phoneNumber=phoneNumber,
      callerName=callerName,
      briModType=briModType))
  


###############################################################################
# @returns Sangoma Digital Board (already added to the PSTN Config)
def addSangomaDigitalBoard( pstnConfig,
                            boardModel,
                            boardID,
                            starting_wanpipe=None,
                            briModType=None):

  interfaceList = []
  isA108 = (boardModel == "A108")
  
  if boardsModelInfo[boardModel]['digitalType'] == "PRI":
    for i in range(1, boardsModelInfo[boardModel]['numInterfaces'] + 1):
      addInterface(
        interfaceList,
        ID=i,
        displayName='B%uI%u' % (boardID, i),
        wanpipe=starting_wanpipe+i-1)
      
  elif boardsModelInfo[boardModel]['digitalType'] == "BRI":

    
    for i in range(1, len(briModType)+1):
      addInterface(
        interfaceList,
        ID=i,
        displayName='B%uI%u_%s' % (boardID, i,briModType[i-1]),
        wanpipe=starting_wanpipe+i-1,
        briModType=briModType[i-1])

  sangomaDigitalBoard = m.sangomaDigitalBoard.factory(
      ID=boardID,
      displayName='B%u - %s_%s' % (boardID, boardModel, 'digital'),
      interfaceNumberingType='DEFAULT',
      model=boardModel,
      primaryClockReference=interfaceList[0].get_ID(),
      interface=interfaceList) # "interface" here is a list

  pstnConfig.get_boards().add_sangomaDigitalBoard(sangomaDigitalBoard)

  return sangomaDigitalBoard

###############################################################################
# @returns Sangoma Digital Board (already added to the PSTN Config)
def addSangomaDigitalBoardHwMode( pstnConfig,
                                  boardModel,
                                  boardID,
                                  displayName,
                                  pciBus,
                                  pciSlot,
                                  interfacesData):
  """Adds a digital board to the config given in parameter using the Hardware
     detection mode (using netborder-list-boards tool)"""
  interfaceList = []

  for data in interfacesData:
    log.info("addSangomaDigitalBoardHwMode.interface: " + str(data))
    if data['active']:
      if(boardsModelInfo[boardModel]['digitalType'] == 'PRI'):
        addInterface(
            interfaceList,
            ID=data['id'],
            displayName=data['displayName'],
            wanpipe=data['wanpipe'])
      else:
        addInterface(
            interfaceList,
            ID=data['id'],
            displayName=data['displayName'],
            wanpipe=data['wanpipe'],
            briModType=data['briModType'])
  
  sangomaDigitalBoard = m.sangomaDigitalBoard.factory(
      ID=boardID,
      displayName=displayName,
      model=boardModel,
      interfaceNumberingType='DEFAULT',
      pciBus=pciBus,
      pciSlot=pciSlot,
      primaryClockReference=interfaceList[0].get_ID(),
      interface=interfaceList) # "interface" here is a list
  pstnConfig.get_boards().add_sangomaDigitalBoard(sangomaDigitalBoard)
  return sangomaDigitalBoard

###############################################################################
# TODO add factory/context where next wanpipe, assigned ids are maintained
# @returns Sangoma Digital T1 Config (already added to the PSTN Config)
def addT1SangomaDigitalConfig(pstnConfig,displayName,configParams, \
    audioParams = defaultAudioParams,
    EchoCancellation = defaultEchoCancellation,
    AcousticEchoCancellation = defaultAcousticEchoCancellation,
    LevelControl = defaultLevelControl,
    BackgroundNoiseAttenuation = defaultBackgroundNoiseAttenuation,
    DTMFRemoval = defaultDTMFRemoval):

  # Add DChannel config to the list
  configParams['DChannelT1'] = \
    m.DChannelT1.factory(mtuSize='2048', channelIndex=24, \
                         RecordToPcapFile='NO', \
                         PcapFilesFolder='/logs/dchan-pcap-recording/')

  # Create a digitalT1 from the parameter list
  digitalT1 = m.digitalT1.factory(**configParams)

  # Create the audio element and all subElement
  echoCancellation = m.EchoCancellation.factory(**EchoCancellation)
  acousticEcho = m.AcousticEchoCancellation.factory(**AcousticEchoCancellation)
  levelControl = m.LevelControl.factory(**LevelControl)
  backgroundNoise = m.BackgroundNoiseAttenuation.factory(**BackgroundNoiseAttenuation)
  dtmfRemoval = m.DTMFRemoval.factory(**DTMFRemoval)

  audioParams['EchoCancellation'] = echoCancellation
  audioParams['AcousticEchoCancellation'] = acousticEcho
  audioParams['LevelControl'] = levelControl
  audioParams['BackgroundNoiseAttenuation'] = backgroundNoise
  audioParams['DTMFRemoval'] = dtmfRemoval
  audio = m.audio.factory(**audioParams)

  # Create the digital config
  sangomaDigitalConfig = m.sangomaDigitalConfig.factory()

  # Set the displayName to the given name and assign the digitalT1
  # and the audioConfig we just created to the config.
  sangomaDigitalConfig.set_displayName(displayName)
  sangomaDigitalConfig.set_digitalT1(digitalT1)
  sangomaDigitalConfig.set_audio(audio)
  sangomaDigitalConfig.set_ID(getNextPhysConfID(pstnConfig))

  # add empty applyTo section. If this section doesn't exists, gateway won't start.
  sangomaDigitalConfig.set_applyTo(m.applyTo.factory())

  # Finally add the whole package to physical configuration list of the current
  # configuration.
  pstnConfig.get_physicalInterfaceConfigs().add_sangomaDigitalConfig(sangomaDigitalConfig)

  return sangomaDigitalConfig

###############################################################################
# @returns Sangoma Digital J1 Config (already added to the PSTN Config)
def addJ1SangomaDigitalConfig(pstnConfig,displayName,configParams, \
    audioParams = defaultAudioParams,
    EchoCancellation = defaultEchoCancellation,
    AcousticEchoCancellation = defaultAcousticEchoCancellation,
    LevelControl = defaultLevelControl,
    BackgroundNoiseAttenuation = defaultBackgroundNoiseAttenuation,
    DTMFRemoval = defaultDTMFRemoval):

  # Add DChannel config to the list
  configParams['DChannelJ1'] = \
    m.DChannelJ1.factory(mtuSize='2048', channelIndex=24, \
                         RecordToPcapFile='NO', \
                         PcapFilesFolder='/logs/dchan-pcap-recording/')

  # Create a digitalJ1 from the parameter list
  digitalJ1 = m.digitalJ1.factory(**configParams)

  # Create the audio element and all subElement
  echoCancellation = m.EchoCancellation.factory(**EchoCancellation)
  acousticEcho = m.AcousticEchoCancellation.factory(**AcousticEchoCancellation)
  levelControl = m.LevelControl.factory(**LevelControl)
  backgroundNoise = m.BackgroundNoiseAttenuation.factory(**BackgroundNoiseAttenuation)
  dtmfRemoval = m.DTMFRemoval.factory(**DTMFRemoval)

  audioParams['EchoCancellation'] = echoCancellation
  audioParams['AcousticEchoCancellation'] = acousticEcho
  audioParams['LevelControl'] = levelControl
  audioParams['BackgroundNoiseAttenuation'] = backgroundNoise
  audioParams['DTMFRemoval'] = dtmfRemoval
  audio = m.audio.factory(**audioParams)

  # Create the digital config
  sangomaDigitalConfig = m.sangomaDigitalConfig.factory()

  # Set the displayName to the given name and assign the digitalJ1
  # and the audioConfig we just created to the config.
  sangomaDigitalConfig.set_displayName(displayName)
  sangomaDigitalConfig.set_digitalJ1(digitalJ1)
  sangomaDigitalConfig.set_audio(audio)
  sangomaDigitalConfig.set_ID(getNextPhysConfID(pstnConfig))

  # add empty applyTo section. If this section doesn't exists, gateway won't start.
  sangomaDigitalConfig.set_applyTo(m.applyTo.factory())

  # Finally add the whole package to physical configuration list of the current
  # configuration.
  pstnConfig.get_physicalInterfaceConfigs().add_sangomaDigitalConfig(sangomaDigitalConfig)

  return sangomaDigitalConfig

# This method is exactly identical to addT1SangomaDigitalConfig except that the
# work is done for an E1 interface.
# @returns Sangoma Digital E1 Config (already added to the PSTN Config)
def addE1SangomaDigitalConfig(pstnConfig,displayName,configParams, \
    audioParams = defaultAudioParams,
    EchoCancellation = defaultEchoCancellation,
    AcousticEchoCancellation = defaultAcousticEchoCancellation,
    LevelControl = defaultLevelControl,
    BackgroundNoiseAttenuation = defaultBackgroundNoiseAttenuation,
    DTMFRemoval = defaultDTMFRemoval):

  
  if configParams['E1Signaling'] == "CCS":
    configParams['DChannelE1'] = \
      m.DChannelE1.factory(mtuSize='2048', channelIndex=16, \
                           RecordToPcapFile='NO', \
                           PcapFilesFolder='/logs/dchan-pcap-recording/')

  digitalE1 = m.digitalE1.factory(**configParams)

  echoCancellation = m.EchoCancellation.factory(**EchoCancellation)
  acousticEcho = m.AcousticEchoCancellation.factory(**AcousticEchoCancellation)
  levelControl = m.LevelControl.factory(**LevelControl)
  backgroundNoise = m.BackgroundNoiseAttenuation.factory(**BackgroundNoiseAttenuation)
  dtmfRemoval = m.DTMFRemoval.factory(**DTMFRemoval)

  audioParams['EchoCancellation'] = echoCancellation
  audioParams['AcousticEchoCancellation'] = acousticEcho
  audioParams['LevelControl'] = levelControl
  audioParams['BackgroundNoiseAttenuation'] = backgroundNoise
  audioParams['DTMFRemoval'] = dtmfRemoval
  audio = m.audio.factory(**audioParams)

  sangomaDigitalConfig = m.sangomaDigitalConfig.factory()

  sangomaDigitalConfig.set_displayName(displayName)
  sangomaDigitalConfig.set_digitalE1(digitalE1)
  sangomaDigitalConfig.set_audio(audio)
  sangomaDigitalConfig.set_ID(getNextPhysConfID(pstnConfig))

  # add empty applyTo section. If this section doesn't exists, gateway won't start.
  sangomaDigitalConfig.set_applyTo(m.applyTo.factory())

  pstnConfig.get_physicalInterfaceConfigs().add_sangomaDigitalConfig(sangomaDigitalConfig)

  return sangomaDigitalConfig

###############################################################################
# TODO add factory/context where next wanpipe, assigned ids are maintained
# @returns Sangoma Digital T1 Config (already added to the PSTN Config)
def addBRISangomaDigitalConfig(pstnConfig,displayName,configParams, \
    audioParams = defaultBriAudioParams,
    EchoCancellation = defaultEchoCancellation,
    AcousticEchoCancellation = defaultAcousticEchoCancellation,
    LevelControl = defaultLevelControl,
    BackgroundNoiseAttenuation = defaultBackgroundNoiseAttenuation,
    DTMFRemoval = defaultDTMFRemoval):

  # Add DChannel config to the list
  configParams['DChannelBRI'] = \
    m.DChannelT1.factory(mtuSize='2048', channelIndex=3, \
                         RecordToPcapFile='NO', \
                         PcapFilesFolder='/logs/dchan-pcap-recording/')

  # Create a digitalT1 from the parameter list
  digitalBRI = m.digitalBRI.factory(**configParams)

  # Create the audio element and all subElement
  echoCancellation = m.EchoCancellation.factory(**EchoCancellation)
  acousticEcho = m.AcousticEchoCancellation.factory(**AcousticEchoCancellation)
  levelControl = m.LevelControl.factory(**LevelControl)
  backgroundNoise = m.BackgroundNoiseAttenuation.factory(**BackgroundNoiseAttenuation)
  dtmfRemoval = m.DTMFRemoval.factory(**DTMFRemoval)

  audioParams['EchoCancellation'] = echoCancellation
  audioParams['AcousticEchoCancellation'] = acousticEcho
  audioParams['LevelControl'] = levelControl
  audioParams['BackgroundNoiseAttenuation'] = backgroundNoise
  audioParams['DTMFRemoval'] = dtmfRemoval
  audio = m.audio.factory(**audioParams)

  # Create the digital config
  sangomaDigitalConfig = m.sangomaDigitalConfig.factory()

  # Set the displayName to the given name and assign the digitalBRI
  # and the audioConfig we just created to the config.
  sangomaDigitalConfig.set_displayName(displayName)
  sangomaDigitalConfig.set_digitalBRI(digitalBRI)
  sangomaDigitalConfig.set_audio(audio)
  sangomaDigitalConfig.set_ID(getNextPhysConfID(pstnConfig))

  # add empty applyTo section. If this section doesn't exists, gateway won't start.
  sangomaDigitalConfig.set_applyTo(m.applyTo.factory())

  # Finally add the whole package to physical configuration list of the current
  # configuration.
  pstnConfig.get_physicalInterfaceConfigs().add_sangomaDigitalConfig(sangomaDigitalConfig)

  return sangomaDigitalConfig

###############################################################################
# TODO add factory/context where next wanpipe, assigned ids are maintained
# Config for ISDN span not in an NFAS group
# @returns FAS Config (already added to the PSTN Config)
def addFasConfig(pstnConfig, **configParams):

  isdn = pstnConfig.get_callControl().get_isdn()
  if isdn is None:
    isdn = m.isdn.factory()
    pstnConfig.get_callControl().set_isdn(isdn)

  fasConfig = m.fasConfig.factory(**configParams)
  fasConfig.set_ID(getNextISDNConfID(pstnConfig))

  applyTo = m.applyTo.factory()
  fasConfig.set_applyTo(applyTo)

  isdn.add_fasConfig(fasConfig)

  return fasConfig

###############################################################################
# copied from addFasConfig
# @returns BRI Config (already added to the PSTN Config)
def addBRIConfig(pstnConfig, **configParams):

  isdn = pstnConfig.get_callControl().get_isdn()
  if isdn is None:
    isdn = m.isdn.factory()
    pstnConfig.get_callControl().set_isdn(isdn)

  briConfig = m.briConfig.factory(**configParams)
  briConfig.set_ID(getNextISDNConfID(pstnConfig))

  applyTo = m.applyTo.factory()
  briConfig.set_applyTo(applyTo)

  isdn.add_briConfig(briConfig)

  return briConfig

def addR2Config(pstnConfig, **configParams):

  cas = pstnConfig.get_callControl().get_cas()
  if cas is None:
    cas = m.cas.factory()
    pstnConfig.get_callControl().set_cas(cas)

  digitalR2Config = m.digitalR2Config.factory(**configParams)
  digitalR2Config.set_ID(getNextCASConfID(pstnConfig))

  applyTo = m.applyTo.factory()
  digitalR2Config.set_applyTo(applyTo)

  cas.add_digitalR2Config(digitalR2Config)

  return digitalR2Config

###############################################################################
# TODO add factory/context where next wanpipe, assigned ids are maintained
# Config for ISDN span not in an NFAS group
# @returns FAS Config (already added to the PSTN Config)
def addNfasGroup(pstnConfig, **configParams):

  isdn = pstnConfig.get_callControl().get_isdn()
  if isdn is None:
    isdn = m.isdn.factory()
    pstnConfig.get_callControl().set_isdn(isdn)

  nfasGroup = m.nfasGroup.factory(**configParams)
  nfasGroup.set_ID(getNextISDNConfID(pstnConfig))

  members = m.members.factory()
  nfasGroup.set_members(members)

  isdn.add_nfasGroup(nfasGroup)

  return nfasGroup

###############################################################################
# @returns Sangoma Analog Board (already added to the PSTN Config)
def addSangomaAnalogBoard(pstnConfig,
                          boardModel,
                          boardID,
                          wanpipe,
                          numFxoInterfaces=0,
                          numFxsInterfaces=0,
                          phoneNumbers=None,
                          callerNames=None):

  interfaceList = []

  j = 0
  # NOTE typexx is used instead of "type" by generatorDS to avaoid reserved
  #   Python keyword
  for i in range(0, numFxoInterfaces):
    addInterface(
      interfaceList,
      ID=i+1,
      displayName='B%uI%u-FXO' % (boardID, i + 1),
      type='FXO',
      wanpipeInterface=i+1,
      phoneNumber=phoneNumbers[j]
    )
    j = j + 1

  for i in range(numFxoInterfaces, numFxsInterfaces+numFxoInterfaces):
    addInterface(
      interfaceList,
      ID=i+1,
      displayName='B%uI%u-FXS' % (boardID, i + 1),
      type='FXS',
      wanpipeInterface=i+1,
      phoneNumber=phoneNumbers[j],
      callerName=callerNames[j]
    )
    addChannelGroup(
      pstnConfig,
      ID=phoneNumbers[j],
      direction="BIDIR",
      outboundHuntingScheme="REVERSE_LINEAR"
    )

    j = j + 1

  sangomaAnalogBoard = m.sangomaAnalogBoard.factory(
    ID=boardID,
    wanpipe=wanpipe,
    displayName='B%u - %s_%s' % (boardID, boardModel, 'analog'),
    model=boardModel,
    interface=interfaceList) # "interface" here is a list

  pstnConfig.get_boards().add_sangomaAnalogBoard(sangomaAnalogBoard)

  return sangomaAnalogBoard

###############################################################################
# Adds a board when Hardware Detection Mode is enabled
def addSangomaAnalogBoardHwMode(pstnConfig,
                                boardModel,
                                boardID,
                                wanpipe,
                                displayName,
                                pciBus,
                                pciSlot,
                                interfacesData):
  """Adds an analog board to the config given in parameter using the Hardware
     detection mode (using netborder-list-boards tool)"""
  interfaceList = []

  for data in interfacesData:
    if data['active']:
      if data['type'] == "FXO":
        addInterface(
          interfaceList,
          ID=data['id'],
          displayName=data['displayName'],
          type=data['type'],
          wanpipeInterface=data['wanpipeInterface'],
          phoneNumber=data['phoneNumber']
        )
      elif data['type'] == "FXS":
        addInterface(
          interfaceList,
          ID=data['id'],
          displayName=data['displayName'],
          type=data['type'],
          wanpipeInterface=data['wanpipeInterface'],
          phoneNumber=data['phoneNumber'],
          callerName=data['callerName']
        )
        addChannelGroup(
          pstnConfig,
          ID=data['phoneNumber'],
          direction="BIDIR",
          outboundHuntingScheme="REVERSE_LINEAR"
        )

  sangomaAnalogBoard = m.sangomaAnalogBoard.factory(
    ID=boardID,
    wanpipe=wanpipe,
    displayName=displayName,
    model=boardModel,
    pciBus=pciBus,
    pciSlot=pciSlot,
    interface=interfaceList) # "interface" here is a list

  pstnConfig.get_boards().add_sangomaAnalogBoard(sangomaAnalogBoard)

  return sangomaAnalogBoard


###############################################################################
# audioParams are "audio" element factory arguments
# @returns Sangoma Analog Config (already added to the PSTN Config)
def addSangomaAnalogConfig(
    pstnConfig,
    displayName,
    variant,
    clocking,
    audioParams = defaultAudioParams,
    EchoCancellation = defaultEchoCancellation,
    AcousticEchoCancellation = defaultAcousticEchoCancellation,
    LevelControl = defaultLevelControl,
    BackgroundNoiseAttenuation = defaultBackgroundNoiseAttenuation,
    DTMFRemoval = defaultDTMFRemoval,
    fxoRxGain = 0.0,
    fxoTxGain = 0.0,
    fxsRxGain = 0.0,
    fxsTxGain = 0.0,
    fxoBatteryDebounceTimeInMs = 64,
    fxoBatteryThresholdInVolts = 3,
    analogFxoFakePolarityEnabled = "NO"
    
    ):

  sangomaAnalogConfig = m.sangomaAnalogConfig.factory()

  echoCancellation = m.EchoCancellation.factory(**EchoCancellation)
  acousticEcho = m.AcousticEchoCancellation.factory(**AcousticEchoCancellation)
  levelControl = m.LevelControl.factory(**LevelControl)
  backgroundNoise = m.BackgroundNoiseAttenuation.factory(**BackgroundNoiseAttenuation)
  dtmfRemoval = m.DTMFRemoval.factory(**DTMFRemoval)

  audioParams['EchoCancellation'] = echoCancellation
  audioParams['AcousticEchoCancellation'] = acousticEcho
  audioParams['LevelControl'] = levelControl
  audioParams['BackgroundNoiseAttenuation'] = backgroundNoise
  audioParams['DTMFRemoval'] = dtmfRemoval
  audio = m.audio.factory(**audioParams)

  sangomaAnalogConfig.set_audio(audio)
  analog = m.analog.factory(variant=variant,clocking=clocking, \
                            fxoRxGainIndB=fxoRxGain, fxoTxGainIndB=fxoTxGain, \
                            fxsRxGainIndB=fxsRxGain, fxsTxGainIndB=fxsTxGain, \
                            fxoBatteryDebounceTimeInMs=fxoBatteryDebounceTimeInMs, \
                            fxoBatteryThresholdInVolts=fxoBatteryThresholdInVolts, \
                            analogFxoFakePolarityEnabled=analogFxoFakePolarityEnabled)
  sangomaAnalogConfig.set_displayName(displayName)
  sangomaAnalogConfig.set_analog(analog)
  sangomaAnalogConfig.set_ID(getNextPhysConfID(pstnConfig))

  # add empty applyTo section. If this section doesn't exists, gateway won't start.
  sangomaAnalogConfig.set_applyTo(m.applyTo.factory())

  pstnConfig.get_physicalInterfaceConfigs().\
      add_sangomaAnalogConfig(sangomaAnalogConfig)

  return sangomaAnalogConfig

###############################################################################
# @returns Sangoma Analog Fxo Config (already added to the PSTN Config)
def addAnalogFxoConfig(pstnConfig, configParams,
    inboundCallParams=defaultInboundParams,
    outboundCallParams=defaultOutboundParams,
    disconnectSupervisionParams=defaultDisconnectParams):
  # get analog from callControl
  analog = pstnConfig.get_callControl().get_analog()
  if analog is None:
    analog = m.analog.factory()
    pstnConfig.get_callControl().set_analog(analog)

  # create inboundCallHandling
  inboundCall = m.inboundCallHandling.factory()
  #   create connect :
  #     - delayInMs
  #     - condition
  connect = m.connect.factory(**inboundCallParams['connect'])
  #   create callerID
  #     - typexx : BT-Bellcore-FSK-SDMF-MDMF is the only supported yet.
  callerID = m.callerID.factory(**inboundCallParams['callerID'])
  #   create notifyOutboundSide
  #     - numberOfRings
  #     - delayInMs
  #     - condition
  notifyOutboundSide = \
    m.notifyOutboundSide.factory(**inboundCallParams['notifyOutboundSide'])

  inboundCall.set_connect(connect)
  inboundCall.set_callerID(callerID)
  inboundCall.set_notifyOutboundSide(notifyOutboundSide)

  # create outboundCallHandling
  outboundCall = m.outboundCallHandling.factory()
  #   create dial
  #     - DTMFDigitDurationInMs
  #     - interDigitDelayInMs
  #     - waitForDialTone
  #     - preDialDelayInMs
  dial = m.dial.factory(**outboundCallParams['dial'])
  #   create connectSupervision
  connectOutboundSide = m.connectOutboundSide.factory(\
      **outboundCallParams['connectOutboundSide'])
  outboundConnect = m.connect.factory(**outboundCallParams['connect'])
  connectSupervision = m.connectSupervision.factory(
      connectOutboundSide=connectOutboundSide, connect=outboundConnect)

  outboundCall.add_dial(dial)
  outboundCall.set_connectSupervision([connectSupervision])

  # create disconnectSupervision
  disconnectSupervision = m.disconnectSupervision.factory()
  #   create disconnectDetectors
  disconnectDetectors = m.disconnectDetectors.factory()

  # add default ringCadenceDetector
  disconnectDetectors.set_ringCadenceDetector(\
      m.ringCadenceDetector.factory(\
        **disconnectSupervisionParams['ringCadenceDetector']))

  #   create each non-cadence detector
  for detectorValues in disconnectSupervisionParams['detectors']:
    disconnectDetectors.add_detector(m.detector.factory(**detectorValues))

  disconnectSupervision.set_disconnectDetectors(disconnectDetectors)

  # create analogFxoConfig
  analogFxo = m.analogFxoConfig.factory(**configParams)
  analogFxo.set_inboundCallHandling(inboundCall)
  analogFxo.set_outboundCallHandling(outboundCall)
  analogFxo.set_disconnectSupervision(disconnectSupervision)
  analogFxo.set_ID(getNextAnalogConfID(pstnConfig))
  analogFxo.set_minimumOnHookDurationMs(2000)

  # Add empty applyTo section.
  # If this section doesn't exists, gateway won't start.
  analogFxo.set_applyTo(m.applyTo.factory())

  # add analogFxoConfig to analog section of callControl
  analog.add_analogFxoConfig(analogFxo)

  return analogFxo

###############################################################################
# @returns Sangoma Analog Fxs Config (already added to the PSTN Config)
def addAnalogFxsConfig(pstnConfig,configParams,
    inboundCallParams=defaultFxsInboundParams,
    disconnectAnswerParams=defaultFXSDisconnectAnswerNotificationParams,
    hotlineDialingParams=defaultHotlineDialingParams):
  # Get analog from callControl
  analog = pstnConfig.get_callControl().get_analog()
  if analog is None:
    analog = m.analog.factory()
    pstnConfig.get_callControl().set_analog(analog)

  # Create inboundCallHandling
  inboundCall = m.inboundCallHandling.factory(**inboundCallParams)

  #   Create relayEarlyMedia :
  #     - value by default is DISABLE
  relayEarlyMedia = \
    m.relayEarlyMedia.factory(**inboundCallParams['relayEarlyMedia'])

  inboundCall.set_relayEarlyMedia(relayEarlyMedia)

  # Create disconnectAnswerNotification section
  disconnectAnswer = \
    m.disconnectAnswerNotification.factory(**disconnectAnswerParams)

  # Create Hotline Dialing section
  hotlineDialing = m.hotlineDialing.factory(**hotlineDialingParams)

  # Create analogFxsConfig
  analogFxs = m.analogFxsConfig.factory(**configParams)
  analogFxs.set_inboundCallHandling(inboundCall)
  analogFxs.set_disconnectAnswerNotification(disconnectAnswer)
  analogFxs.set_hotlineDialing(hotlineDialing)
  analogFxs.set_ID(getNextAnalogConfID(pstnConfig))

  # Add empty applyTo section.
  # If this section doesn't exists, gateway won't start.
  analogFxs.set_applyTo(m.applyTo.factory())

  # Add analogFxsConfig to analog section of callControl
  analog.add_analogFxsConfig(analogFxs)

  return analogFxs

###############################################################################
# Add a virtual digital board
# numInterfaces can be of any value, as long as it is an even number
# The first interface will be peered to the last interface, the second to the
#   one before the last, etc.
# @returns Sangoma Virtual Digital Board (already added to the PSTN Config)
def addSangomaVirtualDigitalBoard(pstnConfig, boardID, numInterfaces):

  # Must be an even number of interfaces
  assert not numInterfaces % 2

  # DChannels thread priority should be HIGHEST with virtual boards
  pstnConfig.get_providers().get_sangoma().\
    get_DChannelsThreads().set_priority("HIGHEST")

  interfaceList = []
  for i in range(1, numInterfaces + 1):
    interfaceList.append(m.interface.factory(
      ID=i,
      displayName='B%uI%u' % (boardID, i),
      peerVirtualInterfaceID=numInterfaces-i+1))

  sangomaVirtualDigitalBoard = m.sangomaVirtualDigitalBoard.factory(
      ID=boardID,
      displayName='B%u - virtual' % (boardID),
      model='VIRTUAL_DIGITAL',
      interface=interfaceList) # "interface" here is a list

  pstnConfig.get_boards().add_sangomaVirtualDigitalBoard(
      sangomaVirtualDigitalBoard)

  return sangomaVirtualDigitalBoard

###############################################################################
# Delete a board from the config by removing all it's reference
# in the configs, removing it from the boards section and also
# by deleting the object in the registered objec list
def deleteBoard(pstnConfig, boardID):
  board = getBoardObject(pstnConfig, boardID)
  boardType = boardsModelInfo[board.get_model()]['type']

  bRv = deleteBoardFromPhysConf(pstnConfig, boardID, boardType)
  bRv = bRv and deleteBoardFromCallControl(pstnConfig, boardID, boardType)
  bRv = bRv and deleteBoardFromBoards(pstnConfig, boardID, boardType)

  # When board is analog, if it has FXS interfaces, we need to delete the
  # channel groups as well
  if bRv:
    if boardType == "analog":
      for interface in board.get_interface():
        if interface.get_type() == "FXS":
          deleteChannelGroup(pstnConfig, interface.get_phoneNumber())

  deleteObjectFromEditor(board)

###############################################################################
# Delete the board from the boards list.
def deleteBoardFromBoards(pstnConfig, boardID, boardType):
  if boardType == 'digital':
    boardList = pstnConfig.get_boards().get_sangomaDigitalBoard()
  elif boardType == 'virtual_digital':
    boardList = pstnConfig.get_boards().get_sangomaVirtualDigitalBoard()
  elif boardType == 'analog':
    boardList = pstnConfig.get_boards().get_sangomaAnalogBoard()

  for board in boardList:
    if board.get_ID() == boardID:
      boardList.remove(board)

  return True


###############################################################################
# Scans the Physical config corresponding to the board type and
# removes all the references to the given board id.
# @param pstnConfig: the current config
# @param boardID: the ID of the board to be removed
# @param boardType: the type (analog,digital,virtual_digital) of the board
def deleteBoardFromPhysConf(pstnConfig, boardID, boardType):
  if boardType == 'digital' or boardType == 'virtual_digital':
    configList = \
      pstnConfig.get_physicalInterfaceConfigs().get_sangomaDigitalConfig()
  elif boardType == 'analog':
    configList = \
      pstnConfig.get_physicalInterfaceConfigs().get_sangomaAnalogConfig()
  else:
    return False

  for config in configList:
    deleteList = []
    refs = getInterfaceRefsForConfig(config)
    for interfaceRef in refs:
      if interfaceRef.get_boardID() == boardID:
        deleteList.append(interfaceRef)
    for el in deleteList:
      refs.remove(el)

  return True

###############################################################################
# Scans the Call control config corresponding to the board type and
# removes all the references to the given board id.
# @param pstnConfig: the current config
# @param boardID: the ID of the board to be removed
# @param boardType: the type (analog,digital,virtual_digital) of the board
def deleteBoardFromCallControl(pstnConfig, boardID, boardType):
  # get the config list corresponding to the board type
  if boardType == 'digital' or boardType == 'virtual_digital':
    isdn = pstnConfig.get_callControl().get_isdn()

    if isdn:
      configList = isdn.get_nfasGroup()
    else:
      configList = []

    # for the digital and virtual boards, we need to delete reference from
    # the nfasGroup and the fasConfig.
    for nfas in configList:
      deleteList = []
      refs = getInterfaceRefsForNfasGroup(nfas)
      for interfaceRef in refs:
        if interfaceRef.get_boardID() == boardID:
          deleteList.append(interfaceRef)
      for el in deleteList:
        refs.remove(el)

    if isdn:
      configList = isdn.get_fasConfig()
    else:
      configList = []
  elif boardType == 'analog':
    analog = pstnConfig.get_callControl().get_analog()
    if analog:
      configList = analog.get_analogFxoConfig() + analog.get_analogFxsConfig()
    else:
      configList = []
  else:
    return False

  for config in configList:
    deleteList = []
    refs = getInterfaceRefsForConfig(config)
    for interfaceRef in refs:
      if interfaceRef.get_boardID() == boardID:
        deleteList.append(interfaceRef)
    for el in deleteList:
      refs.remove(el)


  return True

def deleteInterface(pstnConfig, boardID, interfaceID):
  board = getBoardObject(pstnConfig, boardID)
  interface = getInterfaceObject(pstnConfig, boardID, interfaceID)

  if interface and interface.get_type() == "FXS":
    deleteChannelGroup(pstnConfig, interface.get_phoneNumber())

  deleteInterfaceFromPhysConf(pstnConfig, boardID, interfaceID)
  deleteInterfaceFromCallControl(pstnConfig, boardID, interfaceID)

  board.get_interface().remove(interface)

  deleteObjectFromEditor(interface)

def deleteInterfaceFromPhysConf(pstnConfig, boardID, interfaceID):

  # Analog physical configurations apply directly to a board, not to specific
  # interfaces
  for physConf in \
      pstnConfig.get_physicalInterfaceConfigs().get_sangomaDigitalConfig():
    deleteList = []
    refs = getInterfaceRefsForConfig(physConf)
    for interfaceRef in refs:
      if interfaceRef.get_boardID() == boardID \
          and interfaceRef.get_interfaceID() == interfaceID:
        deleteList.append(interfaceRef)
    for el in deleteList:
      refs.remove(el)


def deleteInterfaceFromCallControl(pstnConfig, boardID, interfaceID):
  isdn = pstnConfig.get_callControl().get_isdn()
  analog = pstnConfig.get_callControl().get_analog()

  nfasGroup = []
  fasConfig = []
  if isdn:
    nfasGroup = isdn.get_nfasGroup()
    fasConfig = isdn.get_fasConfig()

  analogFxo = []
  analogFxs = []
  if analog:
    analogFxo = analog.get_analogFxoConfig()
    analogFxs = analog.get_analogFxsConfig()

  for config in nfasGroup + fasConfig + analogFxo + analogFxs:
    deleteList = []
    if type(config) == m.nfasGroup:
      refs = getInterfaceRefsForNfasGroup(config)
    else:
      refs = getInterfaceRefsForConfig(config)
    for interfaceRef in refs:
      if interfaceRef.get_boardID() == boardID \
          and interfaceRef.get_interfaceID() == interfaceID:
        deleteList.append(interfaceRef)
    for el in deleteList:
      refs.remove(el)

def deleteChannelGroup(pstnConfig, chanGroupID):
  channelGroup = getChannelGroupObject(pstnConfig, chanGroupID)

  if channelGroup is None:
    return

  channelGroups = \
    pstnConfig.get_callControl().get_channelGroups().get_channelGroup()

  deleteObjectFromEditor(channelGroup)

  channelGroups.remove(channelGroup)


###############################################################################
# Get list of interfaceRefs for which nfas group applies to
def getInterfaceRefsForNfasGroup(nfasGroup):
  members = nfasGroup.get_members()
  # Construct members object if none exists
  if members is None:
    members = m.members.factory()
    nfasGroup.set_members(members)
  return members.get_interfaceRef()

###############################################################################
# Add interface to the members section of a nfas group
def addInterfaceToNfasMembers(nfasGroup, board, interface, nfasInterfaceID,\
    DChannelMode, BChannelRangeList=None):
  # NOTE not checking if board is already there
  getInterfaceRefsForNfasGroup(nfasGroup).append(m.interfaceRef.factory(
    boardID=board.get_ID(),
    interfaceID=interface.get_ID(),
    nfasInterfaceID=nfasInterfaceID,
    DChannelMode=DChannelMode,
    BChannelRange=BChannelRangeList))

###############################################################################
# Get list of interfaceRefs for which config applies to
def getInterfaceRefsForConfig(config):
  applyTo = config.get_applyTo()
  # Construct applyTo object if none exists
  if applyTo is None:
    applyTo = m.applyTo.factory()
    config.set_applyTo(applyTo)
  return applyTo.get_interfaceRef()

###############################################################################
# Apply config to board (not enumerating its interfaces)
def applyConfigToBoard(config, board):
  # NOTE not checking if board is already there
  getInterfaceRefsForConfig(config).append(m.interfaceRef.factory(
    boardID=board.get_ID()))

###############################################################################
# Apply config to interface
def applyConfigToInterface(config, board, interface, BChannelRangeList=None,
    channelGroupID=None):
  # NOTE not checking if interface is already there
  getInterfaceRefsForConfig(config).append(m.interfaceRef.factory(
    boardID=board.get_ID(),
    interfaceID=interface.get_ID(),
    BChannelRange=BChannelRangeList,
    channelGroupID=channelGroupID))

###############################################################################
# Apply config to all interfaces of a board
def applyConfigToBoardInterfaces(config, board, BChannelRangeList=None,
    channelGroupID=None):
  for interface in board.get_interface():
    applyConfigToInterface(config, board, interface, BChannelRangeList,
        channelGroupID)

###############################################################################
# Add a channel group to the callControl
def addChannelGroup(pstnConfig, **configParams):
  channelGroups = pstnConfig.get_callControl().get_channelGroups()
  if channelGroups is None:
    channelGroups = m.channelGroups.factory()
    pstnConfig.get_callControl().set_channelGroups(channelGroups)

  if not validateChannelGroupID(pstnConfig, configParams['ID']):
    return False

  channelGroup = m.channelGroup.factory(**configParams)
  channelGroups.add_channelGroup(channelGroup)

  return True

###############################################################################
# Make sure that the ID we are trying to use is not already taken by another
# channel group.
def validateChannelGroupID(pstnConfig, ID):
  channelGroups = pstnConfig.get_callControl().get_channelGroups()

  for channelGroup in channelGroups.get_channelGroup():
    if channelGroup.get_ID() == ID:
      return False

  return True


#########################################################################
# Returns the instance of the board matching the given id.
def getBoardObject(pstnConfig, boardID):
  for b in pstnConfig.get_boards().get_sangomaAnalogBoard():
    if b.get_ID() == boardID:
      return b
  for b in pstnConfig.get_boards().get_sangomaDigitalBoard():
    if b.get_ID() == boardID:
      return b
  for b in pstnConfig.get_boards().get_sangomaVirtualDigitalBoard():
    if b.get_ID() == boardID:
      return b
  return None

##########################################################################
# Returns the interface  corresponding to the
# given board and interface  id.
# boardID is expected to be an integer
# interfaceID is expected to be an integer
def getInterfaceObject(pstnConfig, boardID, interfaceID):
  board = getBoardObject(pstnConfig, boardID)

  for interface in board.get_interface():
    if interface.get_ID() == interfaceID:
      return interface

##########################################################################
# Returns the channel group object matching the ID given in parameter
# chanGroupID or None if no channel groups were found.
def getChannelGroupObject(pstnConfig, chanGroupID):
  channelGroups = \
    pstnConfig.get_callControl().get_channelGroups().get_channelGroup()

  for channelGroup in channelGroups:
    if channelGroup.get_ID() == chanGroupID:
      return channelGroup

  return None

##########################################################################
# Returns the config corresponding to the given id.
# configID is expected to be an integer
def getConfigObject(pstnConfig, configID):
  for config in pstnConfig.get_physicalInterfaceConfigs().get_sangomaDigitalConfig():
    if config.get_ID() == configID:
      return config
  for config in pstnConfig.get_physicalInterfaceConfigs().get_sangomaAnalogConfig():
    if config.get_ID() == configID:
      return config

##########################################################################
# Returns the call control config corresponding to the given id.
# configID is expected to be an integer
def getISDNConfig(pstnConfig, configID):
  callControl = pstnConfig.get_callControl()
  isdn = callControl.get_isdn()

  if isdn:
    for fas in isdn.get_fasConfig():
      if fas.get_ID() == configID:
        return fas
    for nfas in isdn.get_nfasGroup():
      if nfas.get_ID() == configID:
        return nfas
    for bri in isdn.get_briConfig():
      if bri.get_ID() == configID:
        return bri
      
##########################################################################
# Returns the call control config corresponding to the given id.
# configID is expected to be an integer
def getCASConfig(pstnConfig, configID):
  callControl = pstnConfig.get_callControl()
  cas = callControl.get_cas()

  if cas:
    for r2 in cas.get_digitalR2Config():
      if r2.get_ID() == configID:
        return r2

##########################################################################
# Returns the call control config corresponding to the given id.
# configID is expected to be an integer
def getAnalogConfig(pstnConfig, configID):
  callControl = pstnConfig.get_callControl()
  analog = callControl.get_analog()

  if analog:
    for fxo in analog.get_analogFxoConfig():
      if fxo.get_ID() == configID:
        return fxo
    for fxs in analog.get_analogFxsConfig():
      if fxs.get_ID() == configID:
        return fxs

##########################################################################
# Returns true if the board has at least one interface of the type given
# in parameters
def boardHasInterfaces(board,type):
  for interface in board.get_interface():
    if interface.get_type() == type:
      return True

  return False

##########################################################################
# Returns true if the board has at least one FXO interface
def boardHasFxoInterfaces(board):
  return boardHasInterfaces(board,"FXO")

##########################################################################
# Returns true if the board has at least one FXS interface
def boardHasFxsInterfaces(board):
  return boardHasInterfaces(board,"FXS")

##########################################################################
# Sets the new values on the given object attributes according to the
# dataList which contains the key:pair attributeName:newValue.
# Params:
#   - object: instance of the config object
#   - dataList: dictionnary made of the (attributeName,value) tuples
def saveAttributes(object, dataList):
  for attribute, value in dataList.items():
    if attribute != 'nodeId':
      method = getattr(object,"set_%s" % attribute)
      method(value)

###############################################################################
# TODO Move together with object registry for editor elsewhere
def getJSONForEditor(in_pstnConfig):
  # Used to sort the lists
  from operator import itemgetter
  # TreePanel expects a list of Nodes
  jsonTree = []

  if in_pstnConfig is None:
    return jsonTree

  jsonBoards = []
  # Digital boards
  for board in in_pstnConfig.get_boards().get_sangomaDigitalBoard():
    jsonInterfaces = []
    for interface in board.get_interface():
      jsonInterfaces.append({
        'text' : interface.get_displayName(),
        'id'   : registerObjectForEditor(interface),
        'leaf' : 'true',
        'cls'  : 'file',
        'frm'  : 'frmDigitalInterface'
        })

    jsonInterfaces = sorted(jsonInterfaces, key=itemgetter('text'))

    jsonBoards.append({
      'text'      : board.get_displayName(),
      'id'        : registerObjectForEditor(board),
      'leaf'      : 'false',
      'cls'       : 'x-tree-node-collapsed',
      'frm'       : 'panelDigitalBoard',
      'children'  : jsonInterfaces
      })

  # Analog boards
  for board in in_pstnConfig.get_boards().get_sangomaAnalogBoard():
    jsonInterfaces = []
    for interface in board.get_interface():
      jsonInterfaces.append({
        'text' : interface.get_displayName(),
        'id'   : registerObjectForEditor(interface),
        'leaf' : 'true',
        'cls'  : 'file',
        'frm'  : 'frmAnalogInterface'
        })

    jsonInterfaces = sorted(jsonInterfaces, key=itemgetter('text'))

    jsonBoards.append({
      'text'      : board.get_displayName(),
      'id'        : registerObjectForEditor(board),
      'leaf'      : 'false',
      'cls'       : 'x-tree-node-collapsed',
      'frm'       : 'panelAnalogBoard',
      'children'  : jsonInterfaces
      })

  # Virtual Digital boards
  for board in in_pstnConfig.get_boards().get_sangomaVirtualDigitalBoard():
    jsonInterfaces = []
    for interface in board.get_interface():
      jsonInterfaces.append({
        'text' : interface.get_displayName(),
        'id'   : registerObjectForEditor(interface),
        'leaf' : 'true',
        'cls'  : 'file',
        'frm'  : 'frmVirtualDigitalInterface'
        })

    jsonInterfaces = sorted(jsonInterfaces, key=itemgetter('text'))

    jsonBoards.append({
      'text'      : board.get_displayName(),
      'id'        : registerObjectForEditor(board),
      'leaf'      : 'false',
      'cls'       : 'x-tree-node-collapsed',
      'frm'       : 'panelVirtualDigitalBoard',
      'children'  : jsonInterfaces
      })

  jsonBoards = sorted(jsonBoards, key=itemgetter('text'))

  # All Boards
  jsonTree.append({
    'text'      : 'Boards',
    'id'        : 'boards', # ID for display only
    'leaf'      : 'false',
    'cls'       : 'x-tree-node-collapsed',
    'frm'       : 'gridBoards',
    'children'  : jsonBoards
    })

  # PhysicalInterfaceConfig
  jsonPhysicalInterfaceConfigs = []
  # Digital
  for sangomaConfig in in_pstnConfig.\
      get_physicalInterfaceConfigs().get_sangomaDigitalConfig():
    displayName = sangomaConfig.get_displayName()
    if sangomaConfig.get_digitalT1() is not None: #T1 board
      jsonPhysicalInterfaceConfigs.append({
        'text'  : displayName,
        'id'    : registerObjectForEditor(sangomaConfig),
        'leaf'  : 'true',
        'cls'   : 'file',
        'frm'   : 'frmDigitalT1PhysConf'
      })
    elif sangomaConfig.get_digitalE1() is not None: #E1 board
      jsonPhysicalInterfaceConfigs.append({
        'text'  : displayName,
        'id'    : registerObjectForEditor(sangomaConfig),
        'leaf'  : 'true',
        'cls'   : 'file',
        'frm'   : 'frmDigitalE1PhysConf'
      })
    elif sangomaConfig.get_digitalJ1() is not None: #J1 board
      jsonPhysicalInterfaceConfigs.append({
        'text'  : displayName,
        'id'    : registerObjectForEditor(sangomaConfig),
        'leaf'  : 'true',
        'cls'   : 'file',
        'frm'   : 'frmDigitalJ1PhysConf'
      })
    elif sangomaConfig.get_digitalBRI() is not None: #BRI board
      jsonPhysicalInterfaceConfigs.append({
        'text'  : displayName,
        'id'    : registerObjectForEditor(sangomaConfig),
        'leaf'  : 'true',
        'cls'   : 'file',
        'frm'   : 'frmDigitalBRIPhysConf'
      })
  # Analog
  for sangomaConfig in in_pstnConfig.\
      get_physicalInterfaceConfigs().get_sangomaAnalogConfig():
    displayName = sangomaConfig.get_displayName()
    jsonPhysicalInterfaceConfigs.append({
      'text'  : displayName,
      'id'    : registerObjectForEditor(sangomaConfig),
      'leaf'  : 'true',
      'cls'   : 'file',
      'frm'   : 'frmAnalogPhysConf'
      })

  jsonPhysicalInterfaceConfigs = sorted(jsonPhysicalInterfaceConfigs,key=itemgetter('id'))

  jsonTree.append({
    'text'      : 'Physical Configurations',
    'id'        : 'physicalConf',
    'leaf'      : 'false',
    'cls'       : 'x-tree-node-collapsed',
    'frm'       : 'gridPhysicalConfig',
    'children'  : jsonPhysicalInterfaceConfigs
  })

  # CallControl
  #   Channel Groups
  jsonChannelGroups = []
  for channelGroup in in_pstnConfig.get_callControl().\
      get_channelGroups().get_channelGroup():
    jsonChannelGroups.append({
      'text'  : channelGroup.get_ID(),
      'id'    : registerObjectForEditor(channelGroup),
      'leaf'  : 'true',
      'cls'   : 'file',
      'frm'   : 'frmChannelGroup'
      })

  #   ISDN configuration group
  #     FAS config groups
  jsonISDNConfigs = []
  isdn = in_pstnConfig.get_callControl().get_isdn()
  if isdn:
    for fasConfig in in_pstnConfig.get_callControl().\
        get_isdn().get_fasConfig():
      displayName = fasConfig.get_displayName()
      jsonISDNConfigs.append({
        'text'  : displayName,
        'id'    : registerObjectForEditor(fasConfig),
        'leaf'  : 'true',
        'cls'   : 'file',
        'frm'   : 'frmIsdnFas'
        })

    # NFAS config groups
    for nfasGroup in in_pstnConfig.get_callControl().\
        get_isdn().get_nfasGroup():
      displayName = nfasGroup.get_displayName()
      jsonISDNConfigs.append({
        'text'  : displayName,
        'id'    : registerObjectForEditor(nfasGroup),
        'leaf'  : 'true',
        'cls'   : 'file',
        'frm'   : 'frmIsdnNfas'
        })
      
    # BRI config
    for briConfig in in_pstnConfig.get_callControl().\
        get_isdn().get_briConfig():
      displayName = briConfig.get_displayName()
      jsonISDNConfigs.append({
        'text'  : displayName,
        'id'    : registerObjectForEditor(briConfig),
        'leaf'  : 'true',
        'cls'   : 'file',
        'frm'   : 'frmIsdnBri'
        })
      
  # CAS configuration
  jsonCASConfigs = []

  if not globalConfig.getParameter("netborder.oamWebServer.hideR2Options"):
    cas = in_pstnConfig.get_callControl().get_cas()
    if cas:
      for digitalR2Config in in_pstnConfig.get_callControl().get_cas().get_digitalR2Config():
        displayName = digitalR2Config.get_displayName()
        jsonCASConfigs.append({
          'text'  : displayName,
          'id'    : registerObjectForEditor(digitalR2Config),
          'leaf'  : 'true',
          'cls'   : 'file',
          'frm'   : 'frmCasR2'
          })
  
  
  #   Analog configuration group
  jsonAnalogConfigs = []
  analog = in_pstnConfig.get_callControl().get_analog()
  if analog:
    #     Fxo config
    for fxo in in_pstnConfig.get_callControl().\
        get_analog().get_analogFxoConfig():
      displayName = fxo.get_displayName()
      jsonAnalogConfigs.append({
        'text'  : displayName,
        'id'    : registerObjectForEditor(fxo),
        'leaf'  : 'true',
        'cls'   : 'file',
        'frm'   : 'frmAnalogFxo'
        })

    #     Fxs config
    for fxs in in_pstnConfig.get_callControl().\
        get_analog().get_analogFxsConfig():
      displayName = fxs.get_displayName()
      jsonAnalogConfigs.append({
        'text'  : displayName,
        'id'    : registerObjectForEditor(fxs),
        'leaf'  : 'true',
        'cls'   : 'file',
        'frm'   : 'frmAnalogFxs'
        })

  jsonChannelGroups = sorted(jsonChannelGroups,key=itemgetter('id'))
  jsonCallControlNode = {
    'text'      : 'Call Control',
    'id'        : 'callControl', # ID for display only
    'leaf'      : 'false',
    'cls'       : 'x-tree-node-collapsed',
    'children'  : [{
                    'text'      : 'Channel Groups',
                    'id'        : 'channelGroup', # ID for display only
                    'leaf'      : 'false',
                    'cls'       : 'x-tree-node-collapsed',
                    'frm'       : 'gridChannelGroup',
                    'children'  : jsonChannelGroups
                  }]
    }

  jsonISDNConfigs = sorted(jsonISDNConfigs,key=itemgetter('id'))
  jsonCallControlNode['children'].append({
    'text'      : 'ISDN configurations',
    'id'        : 'isdnConfig', # ID for display only
    'leaf'      : 'false',
    'cls'       : 'x-tree-node-collapsed',
    'frm'       : 'gridISDNConfiguration',
    'children'  : jsonISDNConfigs
    })

  if not globalConfig.getParameter("netborder.oamWebServer.hideR2Options"):
    jsonCASConfigs = sorted(jsonCASConfigs,key=itemgetter('id'))
    jsonCallControlNode['children'].append({
      'text'      : 'CAS configurations',
      'id'        : 'casConfig', # ID for display only
      'leaf'      : 'false',
      'cls'       : 'x-tree-node-collapsed',
      'frm'       : 'gridCASConfiguration',
      'children'  : jsonCASConfigs
      })

  jsonAnalogConfigs = sorted(jsonAnalogConfigs,key=itemgetter('id'))
  jsonCallControlNode['children'].append({
    'text'      : 'Analog configurations',
    'id'        : 'analogConfig', # ID for display only
    'leaf'      : 'false',
    'cls'       : 'x-tree-node-collapsed',
    'frm'       : 'gridAnalogConfiguration',
    'children'  : jsonAnalogConfigs
    })

  jsonTree.append(jsonCallControlNode)

  return jsonTree

def containsT1Board (in_pstnConfig):
  for sangomaConfig in in_pstnConfig.\
      get_physicalInterfaceConfigs().get_sangomaDigitalConfig():
    if sangomaConfig.get_digitalE1() is None: #T1 board
      return True
  return False

###############################################################################
# Runtests
if __name__ == '__main__':

    def runtest(testName):

      if testName == 'sangomaDefault':
        writeConfigToFileAsXML(createDefaultSangomaPstnConfig(), sys.stdout)

      elif testName == 'oneA104board':
        pstnConfig = createDefaultSangomaPstnConfig()
        addSangomaDigitalBoard(pstnConfig, boardModel='A104',
            boardID=1, starting_wanpipe=1)
        writeConfigToFileAsXML(pstnConfig, sys.stdout)

      elif testName == 'oneA104boardWithPhysicalAndFasConfig':

        pstnConfig = createDefaultSangomaPstnConfig()

        digitalBoard = addSangomaDigitalBoard(pstnConfig, boardModel='A104',
            boardID=1, starting_wanpipe=1)

        physicalDigitalConfig = addT1SangomaDigitalConfig(pstnConfig,
            'Digital Config',
            defaultT1Config,
            defaultAudioParams)

        # Indicate physical config applies to all interfaces of board
        applyConfigToBoardInterfaces(physicalDigitalConfig, digitalBoard)

        fasConfig = addFasConfig(pstnConfig, **defaultFasConfig)

        # Indicate FAS config applies to all channels of board
        BChannelRange = m.BChannelRange.factory(firstChannel=1,
            lastChannel=23, channelGroupID='default')
        applyConfigToBoardInterfaces(fasConfig, digitalBoard, [BChannelRange])

        writeConfigToFileAsXML(pstnConfig, sys.stdout)

      elif testName == 'oneA200boardWithConfig':

        pstnConfig = createDefaultSangomaPstnConfig()

        analogBoard = addSangomaAnalogBoard(pstnConfig, boardModel='A200',
            boardID=1, wanpipe=1, numFxoInterfaces=4, numFxsInterfaces=0,
            phoneNumbers=["1111","2222","3333","4444"])

        analogPhysicalConfig = addSangomaAnalogConfig(pstnConfig,
          'Analog Config',
          'FCC',
          "INTERNAL",
          defaultAudioParams)

        # Indicate physical config applies to board (analog physical configs
        #   are per board)
        applyConfigToBoard(analogPhysicalConfig, analogBoard)

        configParams = {}
        configParams['displayName'] = 'FXO-config1'
        analogFxoConfig = addAnalogFxoConfig(pstnConfig, configParams)

        # Indicate FXO config applies to all interfaces of board
        applyConfigToBoardInterfaces(analogFxoConfig, analogBoard,
            channelGroupID='default')

        writeConfigToFileAsXML(pstnConfig, sys.stdout)

      elif testName == 'oneVirtualDigitalBoardWithConfig':

        # Must be even
        NUM_INTERFACES = 16

        pstnConfig = createDefaultSangomaPstnConfig()

        virtualDigitalBoard = addSangomaVirtualDigitalBoard(pstnConfig,
            boardID=1, numInterfaces=NUM_INTERFACES)

        # Pretty much all physical config ignored with virtual boards,
        #   but must still be provided (giving defaults)
        physicalDigitalConfig = addT1SangomaDigitalConfig(pstnConfig,
            'Digital Config',
            defaultT1Config,
            defaultAudioParams)

        # Indicate physical config applies to all interfaces of board
        applyConfigToBoardInterfaces(physicalDigitalConfig, virtualDigitalBoard)

        # Virtual boards must have one set of interfaces using TERMINAL
        #   termination peered to interfaces using NETWORK termination
        fasConfigParams = defaultFasConfig
        fasConfigParams['displayName'] = 'Terminal'
        fasConfigParams['termination'] = 'TERMINAL'
        fasConfigParams['isdnT310Timer'] = 30000
        fasConfigParams['isdnT203Timer'] = 30000
        fasConfigParams['isdnT302Timer'] = 15000
        fasTerminalConfig = addFasConfig(pstnConfig, **fasConfigParams)

        fasConfigParams['displayName'] = 'Network'
        fasConfigParams['termination'] = 'NETWORK'
        fasConfigParams['isdnT310Timer'] = 10000
        fasConfigParams['isdnT203Timer'] = 30000
        fasConfigParams['isdnT302Timer'] = 15000
        fasNetworkConfig = addFasConfig(pstnConfig, **fasConfigParams)

        BChannelRange = m.BChannelRange.factory(firstChannel=1,
            lastChannel=23, channelGroupID='default')

        # Apply TERMINAL config to first half of interfaces
        for interface in virtualDigitalBoard.get_interface()[:NUM_INTERFACES/2]:
          applyConfigToInterface(fasTerminalConfig, virtualDigitalBoard,
              interface, [BChannelRange])

        # Apply NETWORK config to second half of interfaces
        for interface in virtualDigitalBoard.get_interface()[NUM_INTERFACES/2:]:
          applyConfigToInterface(fasNetworkConfig, virtualDigitalBoard,
              interface, [BChannelRange])

        writeConfigToFileAsXML(pstnConfig, sys.stdout)

      elif testName == 'getJSONForEditor':

        pstnConfig = m.parse('test-input/pstn-config-digital-only.xml')
        print getJSONForEditor(pstnConfig)

      else:

        print 'Invalid runtest "%s"' % testName


    args = sys.argv[1:]
    if len(args) == 1:
      runtest(args[0])
    else:
      print 'Must pass runtest name as argument'
## !!MIRROR FILE, DO NOT MODIFY!!
