import elixir
from datetime import datetime

##########
# Model
##########

class Property(elixir.Entity):

  elixir.using_options(tablename='properties')

  #NOTE Using the name as primary key would make the properties_values clearer,
  # but increases the db size (and may slow down joins)
  #name = elixir.Field(elixir.Text, primary_key=True)

  name = elixir.Field(elixir.Text)
  useParamDB = elixir.Field(elixir.Boolean)

  # TODO incorporate ParamDB here

class PropertyValue(elixir.Entity):

  elixir.using_options(tablename='properties_values')
  property = elixir.ManyToOne('Property')
  value = elixir.Field(elixir.Text) # value as a string
  hostname = elixir.Field(elixir.Text)
  propertyset = elixir.Field(elixir.Text)
  created = elixir.Field(elixir.DateTime, default=datetime.now)
  lastUpdate = elixir.Field(elixir.DateTime, default=datetime.now)
  comment = elixir.Field(elixir.Text)
  def __repr__(self):
    return '<Property: Name: %s, Value: %s , Hostname: %s , PropertySet: %s ,' \
        ' UseParamDB: %s, Comment: %s>' % \
        (self.property.name, self.value, self.hostname, self.propertyset, \
         self.property.useParamDB, self.comment)

##########
# Helpers
##########

# returns list of property name, value tuples
def getPropertyNameValuesForGivenSetsAndHostnames(dburi, sets={}, hostnames={}):
  import sqlalchemy
  elixir.metadata.bind = sqlalchemy.create_engine(dburi, echo=False)
  elixir.setup_all()
  
  propertyNameValues = []
  for p in PropertyValue.query.all():
    propertyNameValues.append((p.property.name, p.value))

  return propertyNameValues

##########
# Test
##########

if __name__ == '__main__':
  import sqlalchemy
  engine = sqlalchemy.create_engine('sqlite:///test-configs/configuration1.s3db', 
      echo=False)
  elixir.metadata.bind = engine
  elixir.setup_all()
  elixir.create_all()

  p = Property(name='a.b', useParamDB = True)
  p1 = PropertyValue(property=p, value='c')
  p2 = PropertyValue(property=p, value='d')
  assert p1.property.name == 'a.b'
  assert p1.value == 'c'
  assert p1.hostname == None
  assert p1.propertyset == None
  assert p1.property.useParamDB == True
  assert p1.comment == None
  assert p2.property.name == 'a.b'
  assert p2.value == 'd'
  assert p2.hostname == None
  assert p2.propertyset == None
  assert p2.property.useParamDB == True
  assert p2.comment == None
  
  elixir.session.flush()
  elixir.session.commit()
## !!MIRROR FILE, DO NOT MODIFY!!
