#
#  @file   Converter.py
#  @author Jean-Philippe Boily, jr. eng.
#  @date   02/17/2011
#
#  Description:
#
#  Wrapper class around the SmartUpdate.py process to launch the process using
#  netborder-py-wrapper-app executable.
#
#  Copyright (C) 2011 Sangoma Technologies.
#  All Rights Reserved.
#
#  This code is Sangoma Technologies Confidential Property.
#  Use of and Access to this code is covered by a previously executed
#  Non-disclosure Agreement between Sangoma Technologies and the Recipient.
#  This code is being supplied for evaluation purposes only and is not to be
#  used for any other purpose.
#
#

__author__="Jean-Philippe Boily, jr. eng."
__date__ ="$17-Feb-2011 10:43:46 AM$"

#-----------------------------------------------------------------------------#
# imports
#-----------------------------------------------------------------------------#
import os
import sys

# Threading must be imported very early when the process is started,
# otherwise we will run into the following exception
#  File
#"/usr/local/python2.5/lib/python2.5/threading.py", line
#540, in __delete
#    del _active[_get_ident()]
# "More" explanations are given at:
# http://bugs.python.org/issue1562822
import threading
import netborderPyInfra

from paraxip.infra.PyLoggingFunctions import *
from paraxip.smart_update.SmartUpdate import *

#-----------------------------------------------------------------------------#
# global variables
#-----------------------------------------------------------------------------#
fileScopeLogger = Logger("netborder.smartUpdate.SmartUpdateJRunnable")

class SmartUpdateJRunnable:

  def __init__(self):
    log_info(fileScopeLogger, 'SmartUpdateJRunnable ctor')
    self.__importCfgPath = netborderPyInfra.PyGlobalConfig.getParameter(
      "netborder.smartUpdate.importCfgPath")


  # JRunnable I/F
  def run_i(self):
    # Set current directory to netborder.Installation.Directory (so SmartUpdate
    # logs can be sent to the proper location)
    strInstallDir = \
      netborderPyInfra.PyGlobalConfig.getParameter(
        "netborder.Installation.Directory"
      )
    if strInstallDir is None:
      log_error(
        fileScopeLogger,
        "netborder.Installation.Directory parameter not set. Cannot " \
        "determine installation path. Ensure that " \
        "netborder.Installation.Directory parameter is set."
      )
      return
    else:
      os.chdir(strInstallDir)

    log_debug(fileScopeLogger, "Python syspath = %s" % sys.path)

    log_info(
      fileScopeLogger, "importing configuration: %s" % self.__importCfgPath)

    rc = importCfg(self.__importCfgPath)
    if not rc == 0:
      log_error(
        fileScopeLogger,
        "SmartUpdate process returned with code [%s]" % rc
        )

    sys.exit(0)

  def terminate(self):
    log_info(fileScopeLogger, 'SmartUpdateJRunnable stopping...')

if __name__ == "__main__":
  t = SmartUpdateJRunnable()
## !!MIRROR FILE, DO NOT MODIFY!!
