import turbogears
import simplejson
import os
import logging

from turbogears import controllers, expose
from turbogears import flash, redirect, url
from pstnconfig import model
from paraxip.gateway_gui import extjs
from operator import itemgetter

log = logging.getLogger("pstnconfig.controllers")

class CallControlController(controllers.Controller):
#################################################
#                                               #
# Channel Group methods (Add/delete/get/save)   #
#                                               #
#################################################
  ##########################################################################
  @expose()
  def addChannelGroup(self, **data):
    log.info("Adding Channel Group to config: %s" % data)
    channelData = simplejson.loads(data['channelGroup'])

    if channelData['direction'] == 'IN':
      model.addChannelGroup(model.currentConfig,
                            ID=channelData['id'],
                            direction=channelData['direction'])
    else:
      model.addChannelGroup(model.currentConfig,
                      ID=channelData['id'],
                      outboundHuntingScheme=channelData['outboundHuntingScheme'],
                      direction=channelData['direction'])

    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))

  ##########################################################################
  @expose()
  def deleteChannelGroup(self, **data):
    log.info("Deleting Channel Group: %s" % data)
    for selection in simplejson.loads(data['channelGroups']):
      channelGroup = model.getObjectFromEditorId(selection['nodeId'])
      model.currentConfig.get_callControl().get_channelGroups().get_channelGroup().remove(channelGroup)
      model.deleteObjectFromEditor(channelGroup)

    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))

  ##########################################################################
  # Returns the data for the interface associated to the given id.
  # The fields to be exposed to the user are:
  #   - registered id (hidden field).
  #   - The display name of the board
  #   - The wanpipeInterface number of this interface
  @expose() # Ext.form.Action.load - expects JSON
  def getFrmChannelGroupData(self, id):
    log.info("Getting channel group data for channel id: %s" % id)
    # Get the instance of the interface
    channelGroup =  model.getObjectFromEditorId(id)
    data = {
      'id':channelGroup.get_ID(),
      'nodeId':id,
      'direction':channelGroup.get_direction(),
      'outboundHuntingScheme':channelGroup.get_outboundHuntingScheme()
    }             
    log.info("Loading Channel group data: %s" % data)
    return(str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Saves the new informations of the channel group.
  # All values are editable.
  @expose()
  def saveChannelGroup (self, **data):
    log.info("Saving channel group: %s" % data)
    channelData = simplejson.loads(data['channelGroup'])
    channelGroup = model.getObjectFromEditorId(channelData['nodeId'])

    if channelGroup:
      if channelData['direction'] == "IN":
        channelGroup.set_outboundHuntingScheme(None)
      model.saveAttributes(channelGroup,channelData)

    # Write current config to file using default file location
    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))

  ##########################################################################
  # Returns a list of all the channel group configs available
  # in the system.
  # The fields to be exposed to the user are:
  #   - Real ID of the board in the config.
  #   - The display name of the configs
  @expose() # Ext.form.Action.load - expects JSON
  def getGridChannelGroupData(self, id):
    # Get the physicalConfig for digital boards
    data = []
    for channelGroup in model.currentConfig.get_callControl().\
              get_channelGroups().get_channelGroup():
      data.append({
        'id': channelGroup.get_ID(),
        'direction': channelGroup.get_direction(), 
        'outboundHuntingScheme': channelGroup.get_outboundHuntingScheme(),
        'nodeId':model.getObjectId(channelGroup)
      })     
    data = sorted(data, key=itemgetter('id'))
    return(str(extjs.FormActionLoadResponse(success=True, data=data)))

#################################################
#                                               #
# ISDN Config methods (Add/delete/get/save)     #
#                                               #
#################################################
  ##########################################################################
  @expose()
  def addIsdnConfig(self, **data):
    isdnData = simplejson.loads(data['isdnConfig'])

    if isdnData['isdnConfigChoice'] == 'FAS':
      # Default switch variant set to 5ESS for T1 and ETSI for E1.
      if model.containsT1Board(model.currentConfig):
        log.debug("Adding isdn config with T1, switch variant set to 5ESS")
        strSwitchVariant = '5ESS'
      else:
        log.debug("Adding isdn config with E1, switch variant set to ETSI")
        strSwitchVariant = 'ETSI'
      model.addFasConfig(model.currentConfig,
          displayName='new_FAS_%s' % model.getNextISDNConfID(model.currentConfig),
          termination='TERMINAL',
          isdnT310Timer=30000,
          isdnT203Timer=30000,
          isdnT302Timer=15000,
          switchVariant=strSwitchVariant,
          BChannelNegotiation='EXCLUSIVE',
          initiateRestartProcedure='YES',
          inbandProgressTonesGeneration='AS-NEEDED',
          inbandProgressTonesIndicator='PROGRESS',
          callerNameLocation='IN-SETUP-MSG-USING-DISPLAY-IE',
          waitFacilityDelayMs=100,
          overlapDigitsReception='DISABLED')
    elif isdnData['isdnConfigChoice'] == 'NFAS':
      model.addNfasGroup(model.currentConfig,
          displayName='new_NFAS_%s' % model.getNextISDNConfID(model.currentConfig),
          termination='TERMINAL',
          isdnT310Timer=30000,
          isdnT203Timer=30000,
          isdnT302Timer=15000,
          switchVariant='5ESS',
          BChannelNegotiation='EXCLUSIVE',
          initiateRestartProcedure='YES',
          inbandProgressTonesGeneration='AS-NEEDED',
          inbandProgressTonesIndicator='PROGRESS',
          callerNameLocation='IN-SETUP-MSG-USING-DISPLAY-IE',
          waitFacilityDelayMs=100,
          overlapDigitsReception='DISABLED')
    elif isdnData['isdnConfigChoice'] == 'BRI':
      model.addBRIConfig(model.currentConfig,
          displayName='new_BRI_%s' % model.getNextISDNConfID(model.currentConfig),
          termination='NETWORK',
          isdnT310Timer=120000,
          isdnT203Timer=30000,
          isdnT302Timer=15000,
          switchVariant="ETSI",
          BChannelNegotiation='EXCLUSIVE',
          initiateRestartProcedure='NO',
          inbandProgressTonesGeneration='AS-NEEDED',
          inbandProgressTonesIndicator='PROGRESS',
          callerNameLocation='IN-SETUP-MSG-USING-DISPLAY-IE',
          waitFacilityDelayMs=100,
          topology='MULTIPOINT',
          multipleSubscriberNumbers='',
          serviceType='VOICE',
          aocBehavior='NONE',
          overlapDigitsReception='DISABLED')

    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))

  ##########################################################################
  
  #return the number of availables configs
  #ISDN + CAS (Fas/NFAS/BRI/R2)
  def __getNumberOfAvailablesConfigs(self):
    isdn = model.currentConfig.get_callControl().get_isdn()
    cas = model.currentConfig.get_callControl().get_cas()
   
    nbConfigs = 0
    if isdn:
      nbConfigs += len(isdn.get_fasConfig()) + len(isdn.get_nfasGroup()) + \
      len(isdn.get_briConfig())
  
    if cas:
      nbConfigs += len(cas.get_digitalR2Config())

    return nbConfigs
        
  ##########################################################################
  @expose()
  def deleteIsdnConfig(self, **data):
  
    configsToDelete = simplejson.loads(data['isdnConfigs'])
    
    nbConfigsTotal = self.__getNumberOfAvailablesConfigs()
       
    if(nbConfigsTotal - len(configsToDelete) <= 0):
      return(str(extjs.FormActionSubmitResponse(success=False)))
      
    for selection in configsToDelete:
      isdnConfig = model.getObjectFromEditorId(selection['nodeId'])
      if type(isdnConfig) == model.fasConfig:
        model.currentConfig.get_callControl().get_isdn().get_fasConfig().remove(isdnConfig)
      elif type(isdnConfig) == model.nfasGroup:
        model.currentConfig.get_callControl().get_isdn().get_nfasGroup().remove(isdnConfig)
      elif type(isdnConfig) == model.briConfig:
        model.currentConfig.get_callControl().get_isdn().get_briConfig().remove(isdnConfig)
      model.deleteObjectFromEditor(isdnConfig)

    #If it was the last analog config, erase the <analog></analgo> from the config
    if len(model.currentConfig.get_callControl().get_isdn().get_fasConfig()) <= 0 and \
      len(model.currentConfig.get_callControl().get_isdn().get_nfasGroup()) <= 0 and \
      len(model.currentConfig.get_callControl().get_isdn().get_briConfig()) <= 0:
      model.currentConfig.get_callControl().set_isdn(None)
    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))
  
#################################################
#                                               #
# CAS Config methods (Add/delete/get/save)     #
#                                               #
#################################################
  ##########################################################################
  @expose()
  def addCasConfig(self, **data):
    casData = simplejson.loads(data['casConfig'])

    if casData['casConfigChoice'] == 'R2':
    
      model.addR2Config(
          model.currentConfig
          ,displayName='new_R2_%s' % model.getNextCASConfID(model.currentConfig)
          ,switchVariant='Mexico'
          ,get_ani_first="FALSE"
          ,max_ani=10
          ,max_dnis=10
          ,category="National Subscriber"
          ,skip_category="FALSE"
          ,allow_collect_calls="FALSE"
          ,charge_calls="FALSE"
          ,immediate_accept="FALSE"
          ,double_answer="FALSE"
          ,forced_release="FALSE"
          ,mfback_timeout=2000
          ,call_files="FALSE")       
   
    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))
  
  ##########################################################################
  @expose()
  def deleteCasConfig(self, **data):
    
    configsToDelete = simplejson.loads(data['casConfigs'])
    nbConfigsTotal = self.__getNumberOfAvailablesConfigs()
       
    if(nbConfigsTotal - len(configsToDelete) <= 0):
      return(str(extjs.FormActionSubmitResponse(success=False)))
      
    for selection in configsToDelete:
      casConfig = model.getObjectFromEditorId(selection['nodeId'])
      if type(casConfig) == model.digitalR2Config:
        model.currentConfig.get_callControl().get_cas().get_digitalR2Config().remove(casConfig)
      model.deleteObjectFromEditor(casConfig)

    if len(model.currentConfig.get_callControl().get_cas().get_digitalR2Config()) <= 0 and \
      len(model.currentConfig.get_callControl().get_cas().get_digitalR2Config()) <= 0:
      model.currentConfig.get_callControl().set_cas(None)
    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))
    
        
  ##########################################################################
  # Returns a list of all the fas and nfasGroup configs available
  # in the system.
  # The fields to be exposed to the user are:
  #   - Real ID of the board in the config.
  #   - The display name of the configs
  @expose() # Ext.form.Action.load - expects JSON
  def getGridISDNConfigurationData(self, id):
    data = []
    isdn = model.currentConfig.get_callControl().get_isdn()
    if isdn:
      for fas in (isdn.get_fasConfig() + isdn.get_nfasGroup() + isdn.get_briConfig()):
        data.append({
          'id':fas.get_ID(),
          'displayName':fas.get_displayName(),
          'nodeId':model.getObjectId(fas)
        })
    data = sorted(data, key=itemgetter('id'))
    return(str(extjs.FormActionLoadResponse(success=True, data=data)))
  
   ##########################################################################
  # Returns a list of all the cas configs available in the system.
  # The fields to be exposed to the user are:
  #   - Real ID of the board in the config.
  #   - The display name of the configs
  @expose() # Ext.form.Action.load - expects JSON
  def getGridCASConfigurationData(self, id):
    data = []
    cas = model.currentConfig.get_callControl().get_cas()
    if cas:
      for r2 in cas.get_digitalR2Config():
        data.append({
          'id':r2.get_ID(),
          'displayName':r2.get_displayName(),
          'nodeId':model.getObjectId(r2)
        })
    data = sorted(data, key=itemgetter('id'))
    return(str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Returns the value of all the attributes of the fas config.
  # The fields to be exposed to the user are:
  #   - Real ID of the config.
  #   - The display name of the config
  #   - Termination
  #   - Switch Variant
  #   - B-Channel Negotiation
  #   - Tone Set Attributes
  #   - Initiate Restart Procedure
  #   - Inband Progress Tones Generation
  #   - Inband Progress Tones Indicator
  @expose() # Ext.form.Action.load - expects JSON
  def getFrmIsdnFasData(self, id):
    fasConfig = model.getObjectFromEditorId(id)

    data = {
      'idFas_text': fasConfig.get_ID(),
      'nodeId':id,
      'nameFas_text': fasConfig.get_displayName(),
      'terminationFas_select': fasConfig.get_termination(),
      'switchVariantFas_select': fasConfig.get_switchVariant(),
      'BChannelNegotiationFas_select':fasConfig.get_BChannelNegotiation(),
      'initRestartProcFas_select':fasConfig.get_initiateRestartProcedure(),
      'inbandProgTonesGenFas_select':fasConfig.get_inbandProgressTonesGeneration(),
      'inbandProgTonesIndFas_select':fasConfig.get_inbandProgressTonesIndicator(),
      'callerNameLocationFas_select':fasConfig.get_callerNameLocation(),
      'waitFacilityDelayMsFas_numfield':fasConfig.get_waitFacilityDelayMs(),
      'overlapDigitsReceptionFas_select':fasConfig.get_overlapDigitsReception(),
      'minNumberOfDigitsFas_numfield':fasConfig.get_minNumberOfDigits(),
      'maxNumberOfDigitsFas_numfield':fasConfig.get_maxNumberOfDigits(),
      'endCollectionDigitsFas_text':fasConfig.get_endCollectionDigits(),
      'isdnT310TimerFas_text':fasConfig.get_isdnT310Timer(),
      'isdnT203TimerFas_text':fasConfig.get_isdnT203Timer(),
      'isdnT302TimerFas_text':fasConfig.get_isdnT302Timer()
    }                                  

    return(str(extjs.FormActionLoadResponse(success=True, data=data)))  
    
  ##########################################################################
  # Saves the new informations of the config.
  # All values are editable.
  @expose()
  def saveFasConfig (self, **data):
    log.info("Saving fas config: %s" % data)
    fasData = simplejson.loads(data['fasConfig'])
    log.info("Decoded data: %s" % fasData)
    fasConfig = model.getObjectFromEditorId(fasData['fas']['nodeId'])

    if fasConfig:
      for section, values in fasData.items():
        if section == 'fasCallerName':
          if values['callerNameLocation'] == "DISABLED":
            fasConfig.set_waitFacilityDelayMs(None)
        elif section == 'fasOverlap':
          if values['overlapDigitsReception'] == "DISABLED":
            fasConfig.set_minNumberOfDigits(None)
            fasConfig.set_maxNumberOfDigits(None)
            fasConfig.set_endCollectionDigits(None)
        model.saveAttributes(fasConfig,values)

    # Write current config to file using default file location
    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))
  
  ##########################################################################
  # Returns the value of all the attributes of the r2 config.
  # The fields to be exposed to the user are:
  #   - Real ID of the config.
  #   - ID
  #   - Display Name
  #   - Switch Variant
  #   - Get Ani First
  #   - Max ani
  #   - Max dnis
  #   - Category
  #   - Skip Category
  #   - Allow collect calls
  #   - Charge calls
  #   - Imediate Accept
  #   - Double Answer
  #   - Forced Release
  #   - Mfback timeout
  #   - Metering pulse timeout
  #   - Advanced Protocol File
  #   - Call files
  @expose() # Ext.form.Action.load - expects JSON
  def getFrmCasR2Data(self, id):
    digitalR2Config = model.getObjectFromEditorId(id)

    data = {
      'idR2_text': digitalR2Config.get_ID(),
      'nodeId':id,
      'nameR2_text': digitalR2Config.get_displayName(),
      'switchVariantR2_select': digitalR2Config.get_switchVariant(),
      'getAniFirst_select': digitalR2Config.get_get_ani_first(),
      'maxAni_numField': digitalR2Config.get_max_ani(),
      'maxdnis_numField': digitalR2Config.get_max_dnis(),
      'category_select': digitalR2Config.get_category(),
      'skipCategory_select': digitalR2Config.get_skip_category(),
      'allowCollectCalls_select': digitalR2Config.get_allow_collect_calls(),
      'chargeCalls_select': digitalR2Config.get_charge_calls(),
      'immediateAccept_select': digitalR2Config.get_immediate_accept(),
      'doubleAnswer_select': digitalR2Config.get_double_answer(),
      'forcedRelease_select': digitalR2Config.get_forced_release(),
      'mfBackTimeout_numField': digitalR2Config.get_mfback_timeout(),
      'advancedProtocolFile_text': digitalR2Config.get_advanced_protocol_file(),
      'callFiles_select': digitalR2Config.get_call_files()
    }
    return(str(extjs.FormActionLoadResponse(success=True, data=data)))
  
  ##########################################################################
  # Saves the new informations of the config.
  # All values are editable.
  @expose()
  def saveR2Config (self, **data):
  
    infoData = simplejson.loads(data['r2Config'])['infoConfig']
    inboundData = simplejson.loads(data['r2Config'])['inboundConfig']
    outboundData = simplejson.loads(data['r2Config'])['outboundConfig']
    bidirectional = simplejson.loads(data['r2Config'])['bidirectional']


    r2Config = model.getObjectFromEditorId(infoData['nodeId'])
    
    if r2Config:
      
      #Bug 6539: Remove "Advanced Protocol" references from web UI 
      #Control box is removed from Web UI javascript so json will never have
      #the enableAdvancedProtocolFile information.
      #This is why it must be commented below.
      #enableAdvancedProtocolFile = bidirectional["enableAdvancedProtocolFile"]
      #del bidirectional["enableAdvancedProtocolFile"]
    
      model.saveAttributes(r2Config,infoData)
      model.saveAttributes(r2Config,inboundData)
      model.saveAttributes(r2Config,outboundData)
      model.saveAttributes(r2Config,bidirectional)
      
      #if enableAdvancedProtocolFile == False or \
      #  bidirectional["advanced_protocol_file"] == None or \
      #  len(bidirectional["advanced_protocol_file"]) == 0:
      #  model.saveAttributes(r2Config,{"advanced_protocol_file": None})

    # Write current config to file using default file location
    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))
  

  ##########################################################################
  # Returns the value of all the attributes of the nfas config.
  # The fields to be exposed to the user are:
  #   - Real ID of the config.
  #   - The display name of the config
  #   - Termination
  #   - Switch Variant
  #   - B-Channel Negotiation
  #   - Tone Set Attributes
  #   - Initiate Restart Procedure
  #   - Inband Progress Tones Generation
  #   - Inband Progress Tones Indicator
  @expose() # Ext.form.Action.load - expects JSON
  def getFrmIsdnNfasData(self, id):
    nfasGroup = model.getObjectFromEditorId(id)
    data = {
      'idNfas_text': nfasGroup.get_ID(),
      'nodeId':id,
      'nameNfas_text': nfasGroup.get_displayName(),
      'terminationNfas_select': nfasGroup.get_termination(),
      'switchVariantNfas_select': nfasGroup.get_switchVariant(),
      'BChannelNegotiationNfas_select':nfasGroup.get_BChannelNegotiation(),
      'initRestartProcNfas_select':nfasGroup.get_initiateRestartProcedure(),
      'inbandProgTonesGenNfas_select':nfasGroup.get_inbandProgressTonesGeneration(),
      'inbandProgTonesIndNfas_select':nfasGroup.get_inbandProgressTonesIndicator(),
      'callerNameLocationNfas_select':nfasGroup.get_callerNameLocation(),
      'waitFacilityDelayMsNfas_numfield':nfasGroup.get_waitFacilityDelayMs(),
      'overlapDigitsReceptionNFas_select':nfasGroup.get_overlapDigitsReception(),
      'minNumberOfDigitsNFas_numfield':nfasGroup.get_minNumberOfDigits(),
      'maxNumberOfDigitsNFas_numfield':nfasGroup.get_maxNumberOfDigits(),
      'endCollectionDigitsNFas_text':nfasGroup.get_endCollectionDigits(),
      'isdnT310TimerNfas_text':nfasGroup.get_isdnT310Timer(),
      'isdnT203TimerNfas_text':nfasGroup.get_isdnT203Timer(),
      'isdnT302TimerNfas_text':nfasGroup.get_isdnT302Timer()
    }                                  

    return(str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Saves the new informations of the config.
  # All values are editable.
  @expose()
  def saveNfasConfig (self, **data):
    log.info("Saving nfas group config: %s" % data)
    nfasData = simplejson.loads(data['nfasGroup'])
    log.info("Decoded data: %s" % nfasData)
    nfasConfig = model.getObjectFromEditorId(nfasData['nfas']['nodeId'])

    if nfasConfig:
      for section, values in nfasData.items():
        if section == 'nfasCallerName':
          if values['callerNameLocation'] == "DISABLED":
            nfasConfig.set_waitFacilityDelayMs(None)
        elif section == 'nfasOverlap':
          if values['overlapDigitsReception'] == "DISABLED":
            nfasConfig.set_minNumberOfDigits(None)
            nfasConfig.set_maxNumberOfDigits(None)
            nfasConfig.set_endCollectionDigits(None)
        model.saveAttributes(nfasConfig,values)

    # Write current config to file using default file location
    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))
  
  ##########################################################################
  # Returns the value of all the attributes of the fas config.
  # The fields to be exposed to the user are:
  #   - Real ID of the config.
  #   - The display name of the config
  #   - Termination
  #   - Switch Variant
  #   - B-Channel Negotiation
  #   - Tone Set Attributes
  #   - Initiate Restart Procedure
  #   - Inband Progress Tones Generation
  #   - Inband Progress Tones Indicator
  @expose() # Ext.form.Action.load - expects JSON
  def getFrmIsdnBriData(self, id):
    briConfig = model.getObjectFromEditorId(id)

    data = {
      'idBri_text': briConfig.get_ID(),
      'nodeId':id,
      'nameBri_text': briConfig.get_displayName(),
      'terminationBri_select': briConfig.get_termination(),
      'switchVariantBri_select': briConfig.get_switchVariant(),
      'topologyBri_select':briConfig.get_topology(),
      'teiBri_numField':briConfig.get_TEI(),
      'multipleSubscriberNumbersBri_text':briConfig.get_multipleSubscriberNumbers(),
      'BChannelNegotiationBri_select':briConfig.get_BChannelNegotiation(),
      'initRestartProcBri_select':briConfig.get_initiateRestartProcedure(),
      'inbandProgTonesGenBri_select':briConfig.get_inbandProgressTonesGeneration(),
      'inbandProgTonesIndBri_select':briConfig.get_inbandProgressTonesIndicator(),
      'callerNameLocationBri_select':briConfig.get_callerNameLocation(),
      'waitFacilityDelayMsBri_numfield':briConfig.get_waitFacilityDelayMs(),
      'aocBehaviorBri_select':briConfig.get_aocBehavior(),
      'serviceTypeBri_select':briConfig.get_serviceType(),
      'overlapDigitsReceptionBri_select':briConfig.get_overlapDigitsReception(),
      'minNumberOfDigitsBri_numfield':briConfig.get_minNumberOfDigits(),
      'maxNumberOfDigitsBri_numfield':briConfig.get_maxNumberOfDigits(),
      'endCollectionDigitsBri_text':briConfig.get_endCollectionDigits(),
      'isdnT310TimerBri_text':briConfig.get_isdnT310Timer(),
      'isdnT203TimerBri_text':briConfig.get_isdnT203Timer(),
      'isdnT302TimerBri_text':briConfig.get_isdnT302Timer()
    }                                  

    return(str(extjs.FormActionLoadResponse(success=True, data=data)))
    
  ##########################################################################
  # Saves the new informations of the config.
  # All values are editable.
  @expose()
  def saveBriConfig (self, **data):
    log.info("Saving bri config: %s" % data)
    briData = simplejson.loads(data['briConfig'])
    log.info("Decoded data: %s" % briData)
    briConfig = model.getObjectFromEditorId(briData['bri']['nodeId'])

    if briConfig:    
      for section, values in briData.items():
        if section == 'briCallerName':
          if values['callerNameLocation'] == "DISABLED":
            briConfig.set_waitFacilityDelayMs(None)
        elif section == 'briL2':
          if values['topology'] == "MULTIPOINT":
            briConfig.set_TEI(None)
        elif section == 'briOverlap':
          if values['overlapDigitsReception'] == "DISABLED":
            briConfig.set_minNumberOfDigits(None)
            briConfig.set_maxNumberOfDigits(None)
            briConfig.set_endCollectionDigits(None)
            
        model.saveAttributes(briConfig,values)

    # Write current config to file using default file location
    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))


#################################################
#                                               #
# Analog Config methods (Add/delete/get/save)   #
#                                               #
#################################################
  ##########################################################################
  @expose()
  def addAnalogConfig(self, **data): 
    analogData = simplejson.loads(data['analogConfig'])

    if analogData['analogConfigChoice'] == 'FXO':
      configParams = {}
      configParams['displayName'] = \
        'new_FXO_%s' % model.getNextAnalogConfID(model.currentConfig)
      model.addAnalogFxoConfig(model.currentConfig, configParams)
    elif analogData['analogConfigChoice'] == 'FXS':
      # Get default parameters for an Fxs config
      configParams = model.defaultFxsParams
      configParams['displayName'] = \
        'new_FXS_%s' % model.getNextAnalogConfID(model.currentConfig)
      model.addAnalogFxsConfig(model.currentConfig, configParams)
    
    model.writeCurrentConfigToFile()
    return(str(extjs.FormActionSubmitResponse(success=True)))
  
  ##########################################################################
  @expose()
  def deleteAnalogConfig(self, **data):

    log.info("Deleting analog config %s" % data['analogConfigs'])

    config = model.currentConfig

    for selection in simplejson.loads(data['analogConfigs']):
      analogConfig = model.getObjectFromEditorId(selection['nodeId'])
      if type(analogConfig) == model.analogFxoConfig:
        config.get_callControl().get_analog().\
          get_analogFxoConfig().remove(analogConfig)
      elif type(analogConfig) == model.analogFxsConfig:
        config.get_callControl().get_analog().\
          get_analogFxsConfig().remove(analogConfig)
      model.deleteObjectFromEditor(analogConfig)

    #If it was the last analog config, erase the <analog> tag from the config
    if not len(config.get_callControl().get_analog().get_analogFxoConfig()) and \
        not len(config.get_callControl().get_analog().get_analogFxsConfig()):
      config.get_callControl().set_analog(None)

    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))

  ##########################################################################
  # Returns a list of all the analogFxo and analogFxs configs available
  # in the system.
  # The fields to be exposed to the user are:
  #   - Real ID of the board in the config.
  #   - The display name of the configs
  @expose() # Ext.form.Action.load - expects JSON
  def getGridAnalogConfigurationData(self, id):
    data = []
    analog = model.currentConfig.get_callControl().get_analog()
    if analog:
      for config in (analog.get_analogFxoConfig()+analog.get_analogFxsConfig()):
        data.append({
          'id':config.get_ID(),
          'displayName':config.get_displayName(),
          'nodeId':model.getObjectId(config)
        })
    data = sorted(data, key=itemgetter('id'))
    return(str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Returns the value of all the attributes of the analog fxo config.
  # The fields to be exposed to the user are:
  #   - Real ID of the config.
  #   - The display name of the config
  #   - Tone Set Name
  #   - Inbound Call Handling
  #   - Outbound Call Handling
  #   - Connect Supervision
  #   - Disconnect Supervision
  @expose() # Ext.form.Action.load - expects JSON
  def getFrmAnalogFxoData(self, id):
    analogFxo = model.getObjectFromEditorId(id)
    notify =  analogFxo.get_inboundCallHandling().get_notifyOutboundSide()
    inbound = analogFxo.get_inboundCallHandling()
    out_dial = analogFxo.get_outboundCallHandling().get_dial()[0]
    con_super = analogFxo.get_outboundCallHandling().\
                get_connectSupervision()[0]
    ring_cadence = analogFxo.get_disconnectSupervision().\
                   get_disconnectDetectors().get_ringCadenceDetector()
    
    data = {
      'idFxo_text':analogFxo.get_ID(),
      'nodeId':id,
      'nameFxo_text':analogFxo.get_displayName(),
      'minimumOnHookDurationMs':analogFxo.get_minimumOnHookDurationMs(),
      'notifyOutbound_condition_select':notify.get_condition(),
      'numOfRings_text':notify.get_numberOfRings(),
      'notifyOutbound_delay_text':notify.get_delayInMs(),

      'notifyOutboundConnect_condition_select': \
        inbound.get_connect().get_condition(),
      'callerID_type_selection':inbound.get_callerID().get_type(),
      
      'dial_active_selection':out_dial.get_active(),
      'dial_waitForDialTone_selection': out_dial.get_waitForDialTone(),
      'intDialSpecCharacter_text':out_dial.get_InternationalDialingSpecialCharacter(),
      'intDialSpecCharacterSubs_text':out_dial.get_InternationalDialingSpecialCharacterSubstitute(),
      'preDialDelayInMs_text': out_dial.get_preDialDelayInMs(),
      'interDigitDelayInMs_text': out_dial.get_interDigitDelayInMs(),
      'dTMFDigitDurationInMs_text': out_dial.get_DTMFDigitDurationInMs(),

      'connectOutBoundSide_condition_selection': \
        con_super.get_connectOutboundSide().get_condition(),
      'disconnectTimeoutMs_text': con_super.get_connectOutboundSide().\
        get_disconnectTimeoutMs(),
      'connectOnNoRemoteRingTimeoutMs_text': con_super.get_connectOutboundSide().\
        get_connectOnNoRemoteRingTimeoutMs(),
      'connectParamOutbound_condition_selection': \
        con_super.get_connect().get_condition(),
      'connectParamOutbound_delayInMs_text': \
        con_super.get_connect().get_delayInMs(),

      'ringCadenceDetector_active_selection': \
        ring_cadence.get_active(),
      'ringCadenceDetector_OffDelayMs_text': \
        ring_cadence.get_ringCadenceOffDelayMs()
      }

    for detectorRef in analogFxo.get_disconnectSupervision().\
            get_disconnectDetectors().get_detector():
      if detectorRef.get_sequence() == 'batteryDisconnect':
        data['batteryDisconnectDetector_chk'] = True
      elif detectorRef.get_sequence() == "batteryPolarityInversion":
        data['batteryPolarityInversionDetector_chk'] = True
      elif detectorRef.get_sequence() == "dial":
        data['dialToneDetector_chk'] = True
      elif detectorRef.get_sequence() == "reorder":
        data['reorderToneDetector_chk'] = True
      elif detectorRef.get_sequence() == "busy":
        data['busyToneDetector_chk'] = True
      elif detectorRef.get_sequence() == "sit":
        data['sitToneDetector_chk'] = True
      elif detectorRef.get_sequence() == "disconnect":
        data['disconnectToneDetector_chk'] = True

    return(str(extjs.FormActionLoadResponse(success=True, data=data)))
    
  ##########################################################################
  # Saves the new informations of the config.
  # All values are editable.
  @expose()
  def saveFxoConfig (self, **data):
    log.info("Saving fxo group config: %s" % data)
    fxoData = simplejson.loads(data['frmAnalogFxo'])
    log.info("Saving fxo Data config: %s" % fxoData)
    fxoConfig = model.getObjectFromEditorId(data['nodeId'])

    notify =  fxoConfig.get_inboundCallHandling().get_notifyOutboundSide()
    inbound = fxoConfig.get_inboundCallHandling()
    out_dial = fxoConfig.get_outboundCallHandling().get_dial()[0]
    con_super = fxoConfig.get_outboundCallHandling().\
                get_connectSupervision()[0]
    ring_cadence = fxoConfig.get_disconnectSupervision().\
                   get_disconnectDetectors().get_ringCadenceDetector()
    

    # Sections are: config, analog, audio, EchoCancellation, etc.
    # Values are the values of that section wich is a dictionary
    for section, values in fxoData.items():
      if section == 'fxoConfig':
        fxoConfig.set_displayName(values['displayName'])
        fxoConfig.set_minimumOnHookDurationMs(values['minimumOnHookDurationMs'])
      elif section == 'inboundCall':
        notify.set_condition(values['notifyOutbound_condition'])
        if values['notifyOutbound_condition'] == 'AFTER-N-RING':
          notify.set_numberOfRings(values['numberOfRings'])
        else:
          notify.set_numberOfRings(None)
        notify.set_delayInMs(values['notifyOutbound_delay'])
        inbound.get_connect().set_condition(values['connect_condition'])
        inbound.get_callerID().set_type(values['callerID_type'])
      elif section == 'outboundCallDialing':
        out_dial.set_active(values['dial_active'])
        if values['dial_active'] =="YES":
          out_dial.set_waitForDialTone(values['dial_waitForDialTone'])
          out_dial.set_InternationalDialingSpecialCharacter(values['intDialSpecCharacter'])
          out_dial.set_InternationalDialingSpecialCharacterSubstitute(values['intDialSpecCharacterSubs'])
          out_dial.set_preDialDelayInMs(values['preDialDelayInMs'])
          out_dial.set_interDigitDelayInMs(values['interDigitDelayInMs'])
          out_dial.set_DTMFDigitDurationInMs(values['dTMFDigitDurationInMs'])
      elif section == 'outboundCallConnect':
        con_super.get_connectOutboundSide().\
            set_condition(values['connectOutBoundSide_condition'])
        if values['connectOutBoundSide_condition'] == "ON-REMOTE-RING-STOPPED":
          con_super.get_connectOutboundSide().\
            set_disconnectTimeoutMs(values['disconnectTimeoutMs'])
          con_super.get_connectOutboundSide().\
            set_connectOnNoRemoteRingTimeoutMs(values['connectOnNoRemoteRingTimeoutMs'])
        elif values['connectOutBoundSide_condition'] == "ON-BATTERY-POLARITY-INVERSION":
          con_super.get_connectOutboundSide().\
            set_disconnectTimeoutMs(values['disconnectTimeoutMs'])
          con_super.get_connectOutboundSide().\
            set_connectOnNoRemoteRingTimeoutMs(None)
        else:
          con_super.get_connectOutboundSide().set_disconnectTimeoutMs(None)
          con_super.get_connectOutboundSide().set_connectOnNoRemoteRingTimeoutMs(None)
        con_super.get_connect().\
            set_condition(values['connectParamOutbound_condition'])
        con_super.get_connect().\
            set_delayInMs(values['connectParamOutbound_delayInMs'])
      elif section == 'disconnectSupervision':
        ring_cadence.set_active(values['ringCadenceDetector_active'])
        if values['ringCadenceDetector_active'] == "YES":
          ring_cadence.\
              set_ringCadenceOffDelayMs(values['ringCadenceDetector_OffDelayMs'])
        else:
          ring_cadence.\
              set_ringCadenceOffDelayMs(None)

        # Apply the Detectors to the given fxo configuration.
        log.info("Applying Detectors to fxo config: %s" % data)
        fxoConfig.get_disconnectSupervision().\
         get_disconnectDetectors().set_detector([]) #Clear the detector section

        for detectorName, detectorValue in values['disconnectDetectors'].items():
          log.info("Adding detector %s=%s to fxo config" % (detectorName, detectorValue))
          if detectorValue == True:
            d = model.detector(detectorName,detectorName)
            fxoConfig.get_disconnectSupervision().\
              get_disconnectDetectors().add_detector(d)

      else:
        # Get config attribute method
        method = getattr(fxoConfig, "get_%s" % section)
        model.saveAttributes(method(),values)

    # Write current config to file using default file location
    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))

  @expose() # Ext.form.Action.load - expects JSON
  def getFrmAnalogFxsData(self, id):
    analogFxs = model.getObjectFromEditorId(id)
    inbound = analogFxs.get_inboundCallHandling()
    hotline = analogFxs.get_hotlineDialing()
    disconnectAnswer = analogFxs.get_disconnectAnswerNotification()

    data = {
      'Paraxip.config.callControl.fxs.ID':analogFxs.get_ID()
      ,'nodeId':id
      ,'Paraxip.config.callControl.fxs.displayName':analogFxs.get_displayName()
      ,'Paraxip.config.callControl.fxs.regretTime':analogFxs.get_regretTimeMs()
      ,'Paraxip.config.callControl.fxs.oosBehavior':analogFxs.get_oosBehavior()
      ,'Paraxip.config.callControl.fxs.minExpectedDigits': \
          inbound.get_minExpectedDigits()
      ,'Paraxip.config.callControl.fxs.interDigitTimeout': \
          inbound.get_interDigitTimeoutMs()
      ,'Paraxip.config.callControl.fxs.busyToneDuration': \
          inbound.get_busyToneDurationMs()
      ,'Paraxip.config.callControl.fxs.dialToneDuration': \
          inbound.get_dialToneDurationMs()
      ,'Paraxip.config.callControl.fxs.relayEarlyMedia': \
          inbound.get_relayEarlyMedia().get_value()
      ,'Paraxip.config.callControl.fxs.disconnectAnswerNotificationMode': \
          disconnectAnswer.get_mode()
      ,'Paraxip.config.callControl.fxs.hotlineDialingActive': \
          hotline.get_enable()
    }
    # TODO : Add currentDisconnectDurationMs after the disconnecterAnswerMode
    # when this feature will be available.

    if hotline.get_enable() == "YES":
      data['Paraxip.config.callControl.fxs.hotlineDelayBeforeDialing'] = \
          hotline.get_delayBeforeDialingMs()
      data['Paraxip.config.callControl.fxs.hotlineDialingNumber'] = \
          hotline.get_dialingNumber()


    return(str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Saves the new informations of the config.
  # All values are editable.
  @expose()
  def saveFxsConfig (self, **data):
    log.info("Saving fxs group config: %s" % data)
    fxsData = simplejson.loads(data['frmAnalogFxs'])
    log.info("Saving fxs Data config: %s" % fxsData)
    fxsConfig = model.getObjectFromEditorId(data['nodeId'])

    if fxsConfig is None:
      log.error("Unable to find the FXS configuration.")
      return(str(extjs.FormActionSubmitResponse(success=False)))

    hotline = fxsConfig.get_hotlineDialing()
    inbound = fxsConfig.get_inboundCallHandling()
    disconnectAnswer = fxsConfig.get_disconnectAnswerNotification()

    for section, values in fxsData.items():
      if section == "fxsConfig":
        fxsConfig.set_displayName(values['displayName'])
        fxsConfig.set_regretTimeMs(values['regretTimeMs'])
        fxsConfig.set_oosBehavior(values['oosBehavior'])
      elif section == "inboundCallHandling":
        inbound.set_minExpectedDigits(values['minExpectedDigits'])
        inbound.set_interDigitTimeoutMs(values['interDigitTimeoutMs'])
        inbound.set_busyToneDurationMs(values['busyToneDurationMs'])
        inbound.set_dialToneDurationMs(values['dialToneDurationMs'])
        inbound.get_relayEarlyMedia().set_value(values['relayEarlyMedia'])
      elif section == "hotlineDialing":
        hotline.set_enable(values['hotlineDialingActive'])
        if values['hotlineDialingActive'] == "YES":
          hotline.set_delayBeforeDialingMs(values['delayBeforeDialingMs'])
          hotline.set_dialingNumber(values['dialingNumber'])
        else:
          hotline.set_delayBeforeDialingMs(None)
          hotline.set_dialingNumber(None)
      elif section == "disconnectAnswerNotification":
        disconnectAnswer.set_mode(values['disconnectAnswerMode'])
        # TODO : save the currentDisconnectDurationMs if mode is 
        # CURRENT-DISCONNECT when this feature will be available

      else:
        # Get config attribute method
        method = getattr(fxsConfig, "get_%s" % section)
        model.saveAttributes(method(),values)

    # Write current config to file using default file location
    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))

#################################################
#                                               #
# ApplyTo FasConfigs                            #
#                                               #
#################################################
  ##########################################################################
  # Returns all the digital boards.
  @expose()
  def getBoards(self,id):
    data = []
    for board in (model.currentConfig.get_boards().get_sangomaDigitalBoard() + \
      model.currentConfig.get_boards().get_sangomaVirtualDigitalBoard()):
      data.append({
        'id': board.get_ID(),
        'displayName':board.get_displayName(),
        'nodeId': model.getObjectId(board)
      })

    data = sorted(data, key=itemgetter('id'))
    return (str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Returns all the digital interfaces of the board corresponding to the 
  # given id.
  @expose()
  def getInterfaces(self,id):
    data = []
    board = model.getBoardObject(model.currentConfig, int(id))

    for interface in board.get_interface():
      data.append({
        'id': interface.get_ID(),
        'displayName': interface.get_displayName()
      })
    data = sorted(data, key=itemgetter('id'))

    return (str(extjs.FormActionLoadResponse(success=True, data=data)))


  ##########################################################################
  # Returns all the channel groups, excluding the channel groups reseved
  # for FXS lines.
  @expose()
  def getChannelGroups(self,id):
    data = []
    boards = model.currentConfig.get_boards().get_sangomaAnalogBoard()

    for channelGroup in model.currentConfig.get_callControl().get_channelGroups().get_channelGroup():
      isAnalogChannelGroup = False
      for board in boards:
        for interface in board.get_interface():
          if interface.get_phoneNumber() == channelGroup.get_ID():
            isAnalogChannelGroup = True

      if not isAnalogChannelGroup:
        data.append({
          'id': channelGroup.get_ID()
        })

    return (str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Returns the data in the fas config applyTo section.
  # Returned values are:
  #   - Board id
  #   - Interface id
  #   - First Channel
  #   - Last Channel
  #   - Channel Group ID
  @expose()
  def getFasBChannels(self, id):
    data = []
    fasConfig = model.getISDNConfig(model.currentConfig, int(id))

    for interfaceRef in model.getInterfaceRefsForConfig(fasConfig):
      for bChannelRange in interfaceRef.get_BChannelRange():
        data.append({
          'boardId':interfaceRef.get_boardID(),
          'interfaceId':interfaceRef.get_interfaceID(),
          'firstChannel':bChannelRange.get_firstChannel(),
          'lastChannel':bChannelRange.get_lastChannel(),
          'channelGroupID':bChannelRange.get_channelGroupID(),
          'nodeId': model.getObjectId(model.getBoardObject(model.currentConfig, interfaceRef.get_boardID()))
        })

    return (str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Apply the BChannel to the given fas configuration.
  @expose() # Ext.form.Action.load - expects JSON
  def applyBChannelToFasConfig(self,**data):
    log.info("Applying BChannels to fas config: %s" % data)
    id = simplejson.loads(data['fasConfigID'])
    gridData = simplejson.loads(data['gridData'])

    fasConfig = model.getISDNConfig(model.currentConfig, int(id))
    fasConfig.set_applyTo(None) #Clear the applyTo section

    for interfaceInfo in gridData:
      #Find the board object
      board = model.getBoardObject(model.currentConfig,int(interfaceInfo['boardId']))
      #Find the interface object
      interface = model.getInterfaceObject(model.currentConfig, board.get_ID(), int(interfaceInfo['interfaceId']))
      BChannelRange = model.BChannelRange.factory(firstChannel=interfaceInfo['firstChannel'], 
        lastChannel=interfaceInfo['lastChannel'], channelGroupID=interfaceInfo['channelGroupID'])
      model.applyConfigToInterface(fasConfig,board,interface,[BChannelRange])

    model.writeCurrentConfigToFile()
    return(str(extjs.FormActionResponse(success=True)))

#################################################
#                                               #
# ApplyTo Nfas Group                            #
#                                               #
#################################################
  ##########################################################################
  # Returns the data in the nfas group members section.
  # Returned values are:
  #   - Board id
  #   - Interface id
  #   - Nfas Interface Id
  #   - DChannel Mode
  #   - First Channel
  #   - Last Channel
  #   - Channel Group ID
  @expose()
  def getNfasBChannels(self, id):
    data = []
    nfasGroup = model.getISDNConfig(model.currentConfig, int(id))

    for interfaceRef in model.getInterfaceRefsForNfasGroup(nfasGroup):
      for bChannelRange in interfaceRef.get_BChannelRange():
        data.append({
          'boardId':interfaceRef.get_boardID(),
          'interfaceId':interfaceRef.get_interfaceID(),
          'nfasInterfaceID':interfaceRef.get_nfasInterfaceID(),
          'DChannelMode':interfaceRef.get_DChannelMode(),
          'firstChannel':bChannelRange.get_firstChannel(),
          'lastChannel':bChannelRange.get_lastChannel(),
          'channelGroupID':bChannelRange.get_channelGroupID(),
          'nodeId': model.getObjectId(model.getBoardObject(model.currentConfig, interfaceRef.get_boardID()))
        })

    return (str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Returns all the digital boards.
  @expose()
  def applyBChannelToNfasConfig(self,**data):
    log.info("Applying BChannels to nfas config: %s" % data)
    id = simplejson.loads(data['nfasGroupID'])
    gridData = simplejson.loads(data['gridData'])

    nfasGroup = model.getISDNConfig(model.currentConfig, int(id))
    nfasGroup.set_members(None) #Clear the members section

    for interfaceInfo in gridData:
      #Find the board object
      board = model.getBoardObject(model.currentConfig,int(interfaceInfo['boardId']))
      #Find the interface object
      interface = model.getInterfaceObject(model.currentConfig, board.get_ID(), int(interfaceInfo['interfaceId']))
      BChannelRange = model.BChannelRange.factory(firstChannel=interfaceInfo['firstChannel'], 
        lastChannel=interfaceInfo['lastChannel'], channelGroupID=interfaceInfo['channelGroupID'])
      model.addInterfaceToNfasMembers(nfasGroup,board,interface,
          interfaceInfo['nfasInterfaceID'],interfaceInfo['DChannelMode'],[BChannelRange])

    model.writeCurrentConfigToFile()
    return(str(extjs.FormActionResponse(success=True)))
  

##########################################################################
  # Returns the data in the fas config applyTo section.
  # Returned values are:
  #   - Board id
  #   - Interface id
  #   - First Channel
  #   - Last Channel
  #   - Channel Group ID
  @expose()
  def getFasBChannels(self, id):
    data = []
    fasConfig = model.getISDNConfig(model.currentConfig, int(id))

    for interfaceRef in model.getInterfaceRefsForConfig(fasConfig):
      for bChannelRange in interfaceRef.get_BChannelRange():
        data.append({
          'boardId':interfaceRef.get_boardID(),
          'interfaceId':interfaceRef.get_interfaceID(),
          'firstChannel':bChannelRange.get_firstChannel(),
          'lastChannel':bChannelRange.get_lastChannel(),
          'channelGroupID':bChannelRange.get_channelGroupID(),
          'nodeId': model.getObjectId(model.getBoardObject(model.currentConfig, interfaceRef.get_boardID()))
        })

    return (str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Apply the BChannel to the given fas configuration.
  @expose() # Ext.form.Action.load - expects JSON
  def applyBChannelToFasConfig(self,**data):
    log.info("Applying BChannels to fas config: %s" % data)
    id = simplejson.loads(data['fasConfigID'])
    gridData = simplejson.loads(data['gridData'])

    fasConfig = model.getISDNConfig(model.currentConfig, int(id))
    fasConfig.set_applyTo(None) #Clear the applyTo section

    for interfaceInfo in gridData:
      #Find the board object
      board = model.getBoardObject(model.currentConfig,int(interfaceInfo['boardId']))
      #Find the interface object
      interface = model.getInterfaceObject(model.currentConfig, board.get_ID(), int(interfaceInfo['interfaceId']))
      BChannelRange = model.BChannelRange.factory(firstChannel=interfaceInfo['firstChannel'], 
        lastChannel=interfaceInfo['lastChannel'], channelGroupID=interfaceInfo['channelGroupID'])
      model.applyConfigToInterface(fasConfig,board,interface,[BChannelRange])

    model.writeCurrentConfigToFile()
    return(str(extjs.FormActionResponse(success=True)))
  
  ##########################################################################
  # Returns the data in the R2 config applyTo section.
  # Returned values are:
  #   - Board id
  #   - Interface id
  #   - First Channel
  #   - Last Channel
  #   - Channel Group ID
  @expose()
  def getR2BChannels(self, id):
    data = []
    r2Config = model.getCASConfig(model.currentConfig, int(id))

    for interfaceRef in model.getInterfaceRefsForConfig(r2Config):
      for bChannelRange in interfaceRef.get_BChannelRange():
        data.append({
          'boardId':interfaceRef.get_boardID(),
          'interfaceId':interfaceRef.get_interfaceID(),
          'firstChannel':bChannelRange.get_firstChannel(),
          'lastChannel':bChannelRange.get_lastChannel(),
          'channelGroupID':bChannelRange.get_channelGroupID(),
          'nodeId': model.getObjectId(model.getBoardObject(model.currentConfig, interfaceRef.get_boardID()))
        })

    return (str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Apply the BChannel to the given r2 configuration.
  @expose() # Ext.form.Action.load - expects JSON
  def applyBChannelToR2Config(self,**data):
    log.info("Applying BChannels to r2 config: %s" % data)
    id = simplejson.loads(data['r2ConfigID'])
    gridData = simplejson.loads(data['gridData'])

    r2Config = model.getCASConfig(model.currentConfig, int(id))
    r2Config.set_applyTo(None) #Clear the applyTo section

    for interfaceInfo in gridData:
      #Find the board object
      board = model.getBoardObject(model.currentConfig,int(interfaceInfo['boardId']))
      #Find the interface object
      interface = model.getInterfaceObject(model.currentConfig, board.get_ID(), int(interfaceInfo['interfaceId']))
      BChannelRange = model.BChannelRange.factory(firstChannel=interfaceInfo['firstChannel'], 
        lastChannel=interfaceInfo['lastChannel'], channelGroupID=interfaceInfo['channelGroupID'])
      model.applyConfigToInterface(r2Config,board,interface,[BChannelRange])

    model.writeCurrentConfigToFile()
    return(str(extjs.FormActionResponse(success=True)))
#################################################
#                                               #
# ApplyTo AnalogFxoConfig                       #
#                                               #
#################################################
  ##########################################################################
  # Returns all the analog boards.
  @expose()
  def getAnalogFxoBoards(self, id):
    data = []
    for board in model.currentConfig.get_boards().get_sangomaAnalogBoard():
      if model.boardHasFxoInterfaces(board):
        data.append({
          'id': board.get_ID(),
          'displayName':board.get_displayName(),
          'nodeId': model.getObjectId(board)
        })

    data = sorted(data, key=itemgetter('id'))
    return (str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Returns all the analog boards.
  @expose()
  def getAnalogFxsBoards(self, id):
    data = []
    for board in model.currentConfig.get_boards().get_sangomaAnalogBoard():
      if model.boardHasFxsInterfaces(board):
        data.append({
          'id': board.get_ID(),
          'displayName':board.get_displayName(),
          'nodeId': model.getObjectId(board)
        })

    data = sorted(data, key=itemgetter('id'))
    return (str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Returns all the analog interfaces of the board corresponding to the 
  # given id.
  @expose()
  def getAnalogFxoInterfaces(self, id):
    data = []
    board = model.getBoardObject(model.currentConfig, int(id))

    for interface in board.get_interface():
      if interface.get_type() == "FXO":
        data.append({
          'id': interface.get_ID(),
          'displayName': interface.get_displayName(),
          'phoneNumber': interface.get_phoneNumber()
        })
    data = sorted(data, key=itemgetter('id'))

    return (str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Returns all the analog interfaces of the board corresponding to the
  # given id that are of FXS type.
  @expose()
  def getAnalogFxsInterfaces(self, id):
    data = []
    board = model.getBoardObject(model.currentConfig, int(id))

    for interface in board.get_interface():
      if interface.get_type() == "FXS":
        data.append({
          'id': interface.get_ID(),
          'displayName': interface.get_displayName(),
          'phoneNumber': interface.get_phoneNumber()
        })
    data = sorted(data, key=itemgetter('id'))

    return (str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Returns the data in the fxo config applyTo section.
  # Returned values are:
  #   - Board id
  #   - Interface id
  #   - Channel Group ID
  @expose()
  def getAssignedFxoInterfaces(self, id):
    data = []
    fxoConfig = model.getAnalogConfig(model.currentConfig, int(id))

    for interfaceRef in model.getInterfaceRefsForConfig(fxoConfig):
      data.append({
        'boardId':interfaceRef.get_boardID(),
        'interfaceId':interfaceRef.get_interfaceID(),
        'channelGroupID':interfaceRef.get_channelGroupID(),
        'nodeId': model.getObjectId(model.getBoardObject(model.currentConfig, interfaceRef.get_boardID()))
      })

    return (str(extjs.FormActionLoadResponse(success=True, data=data)))
  
  ##########################################################################
  # Returns the data in the fxs config applyTo section.
  # Returned values are:
  #   - Board id
  #   - Interface id
  #   - Channel Group ID
  @expose()
  def getAssignedFxsInterfaces(self, id):
    data = []
    fxsConfig = model.getAnalogConfig(model.currentConfig, int(id))

    for interfaceRef in model.getInterfaceRefsForConfig(fxsConfig):
      data.append({
        'boardId':interfaceRef.get_boardID(),
        'interfaceId':interfaceRef.get_interfaceID(),
        'channelGroupID':interfaceRef.get_channelGroupID(),
        'nodeId': model.getObjectId(model.getBoardObject(model.currentConfig, interfaceRef.get_boardID()))
      })

    return (str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Apply the Interfaces to the given fxo configuration.
  @expose() # Ext.form.Action.load - expects JSON
  def applyInterfacesToFxoConfig(self, **data):
    log.info("Applying Interfaces to fxo config: %s" % data)
    id = simplejson.loads(data['fxoConfigID'])
    gridData = simplejson.loads(data['gridData'])

    fxoConfig = model.getAnalogConfig(model.currentConfig, int(id))
    fxoConfig.set_applyTo(None) #Clear the applyTo section

    for interfaceInfo in gridData:
      #Find the board object
      board = model.getBoardObject(model.currentConfig,int(interfaceInfo['boardId']))
      #Find the interface object
      interface = model.getInterfaceObject(model.currentConfig, board.get_ID(), int(interfaceInfo['interfaceId']))
      if interface is None:
        log.error("Unable to find interface with ID [%s] for board ID [%s]." \
                  % (interfaceInfo['interfaceId'], interfaceInfo['boardId']))
        return(str(extjs.FormActionSubmitResponse(success=False)))
      elif not interface.get_type() == "FXO":
        log.error("The interface with ID [%s] is not an FXO interface." \
                  % interfaceInfo['interfaceId'])
        return(str(extjs.FormActionSubmitResponse(success=False)))
      model.applyConfigToInterface(fxoConfig,board,interface,channelGroupID=interfaceInfo['channelGroupID'])

    model.writeCurrentConfigToFile()
    return(str(extjs.FormActionSubmitResponse(success=True)))

  ##########################################################################
  # Apply the Interfaces to the given fxs configuration.
  @expose() # Ext.form.Action.load - expects JSON
  def applyInterfacesToFxsConfig(self, **data):
    log.info("Applying Interfaces to fxs config: %s" % data)
    id = simplejson.loads(data['fxsConfigID'])
    gridData = simplejson.loads(data['gridData'])
    config = model.currentConfig

    fxsConfig = model.getAnalogConfig(config, int(id))
    fxsConfig.set_applyTo(None) #Clear the applyTo section

    for interfaceInfo in gridData:
      #Find the board object
      board = model.getBoardObject(config,int(interfaceInfo['boardId']))
      #Find the interface object
      interface = model.getInterfaceObject(config, board.get_ID(), int(interfaceInfo['interfaceId']))
      if interface is None:
        log.error("Unable to find interface with ID [%s] for board ID [%s]." \
                  % (interfaceInfo['interfaceId'], interfaceInfo['boardId']))
        return(str(extjs.FormActionSubmitResponse(success=False)))
      elif not interface.get_type() == "FXS":
        log.error("The interface with ID [%s] is not an FXS interface." \
                  % interfaceInfo['interfaceId'])
        return(str(extjs.FormActionSubmitResponse(success=False)))

      model.applyConfigToInterface(fxsConfig,board,interface,channelGroupID=interfaceInfo['channelGroupID'])

    model.writeCurrentConfigToFile()
    return(str(extjs.FormActionSubmitResponse(success=True)))

