import os, stat
import turbogears
import cherrypy
import datetime
import pstnconfig.model as model
import logging
import netborderPySangoma

from turbogears import controllers, expose
from cherrypy.lib.cptools import serve_file
from odict import odict
from wizard import WizardController
from status import StatusController
from help import HelpController
from config import ConfigController
from boards import BoardsController
from physicalConfig import PhysicalConfigController
from callControl import CallControlController
from webServices import WebServicesController
from fileBrowser import FileBrowserController
from fxoTune import FxoTuneController
from support import SupportController
from paraxip.gateway_gui_widgets.tabpstn  import TabPSTN
from netborderPyInfra import PyGlobalConfig as globalConfig
from hardwareprobe import HardwareProbeController

log = logging.getLogger("rootconfig")

class Root(controllers.RootController):

  # Determine the project path.
  strInstallDir = \
    globalConfig.getParameter("netborder.Installation.OamWebServer.Path")
  log.info("Installation directory: %s" % strInstallDir)

  if strInstallDir is None:
    log.error("netborder.Installation.OamWebServer.Path parameter not set. " \
              "Cannot determine project path")

  projectName = globalConfig.getParameter("netborder.oamWebServer.projectName")

  strProjectPath = os.path.join(strInstallDir, projectName)

  log.info("Project path: %s" % strProjectPath)

  autoDetectHardware = \
    globalConfig.getParameter("netborder.oamWebServer.enableHardwareDetection")

  model.hardwareDetectionEnabled = autoDetectHardware

  if autoDetectHardware:
    model.hardwareCfg = model.parseString(netborderPySangoma.getBoardsListString())

    log.debug("Hardware Config version: %s" % model.hardwareCfg.get_version())

    log.info("Hardware Detection Mode enabled")

  # Check if file exists
  if model.PSTNConfigFileExists():

    # PSTN config file already exists
    log.info( "PSTN config file %s already exists",
              model.getPSTNConfigFileName() )

    try:
      if model.currentConfig is None:
        model.currentConfig = model.parse(model.getPSTNConfigFileName())

      if autoDetectHardware:
        log.info("validating current config with detected hardware")
        model.validateCurrentConfig()

      activeTab='statusControlTabId'
    except:
      log.info( "PSTN config file %s is invalid",
              model.getPSTNConfigFileName() )
      activeTab='wizardTabId'

  else:
    log.info( "PSTN config file %s does not already exist",
              model.getPSTNConfigFileName() )
    activeTab='wizardTabId'

  # TODO - retrieve the url and the namespace from the config file
  #Bug 2168 - changed localhost for 127.0.0.1, because of IPV6 problem,
  #Change back if using IPV6
  webService = WebServicesController("http://127.0.0.1:18086/","urn:SoapOAM")

  wizard      = WizardController(strProjectPath)
  status      = StatusController(strProjectPath,webService)
  help        = HelpController()
  config      = ConfigController(strProjectPath,webService)
  boards      = BoardsController()
  phys        = PhysicalConfigController()
  call        = CallControlController()
  fileBrowser = FileBrowserController()
  support     = SupportController()
  fxoTune     = FxoTuneController(webService,status)
  hwprobe     = HardwareProbeController()

  #Elements that will build our tabber
  tabs = odict()
  tabs['Status and Controls'] = \
    ['','statusControlTabId',"Status and Controls of the gateway"]
  tabs['Configuration'] = \
    ['','configTabId',"Edit the routing rules and/or the pstn config"]
  tabs['Quick Setup'] = \
    [ '', 'wizardTabId',
      "Setup a new system configuration. Reset all previous modifications"]
  tabs['File Browser'] = \
    ['','fileBrowserTabId',"File Browser to edit and download system files."]
  tabs['Support'] = \
	['','supportTabId',"Generate archive file for support"]	
  tabs['Help'] = ['/static/help.html','helpTabId',"Help"]

  items = [
    {
      'title': k,
      'autoLoad': v[0],
      'id': v[1],
      'tabTip': v[2]
    } for k, v in tabs.items()
  ]

  tabList = \
    [(k.replace(" ","").lower(),v[0],v[1],v[2]) for k, v in tabs.items()]
  tabs = dict(tabs)

  strFileLocation = os.path.join(strProjectPath, "pstnconfig/static/"
                                 "javascript/widgets_src/main_pstn_tabs.js")

  # Flag to switch on the offering of the virtual boards in the wizard.
  offerVirtualBoards = \
    globalConfig.getParameter("netborder.oamWebServer.offerVirtualBoards")


  # Flag to determine if the gateway is started from the command line
  # of if it uses the service. This changes the method to get the status:
  # from the Service Control Manager of by pinging the WebServices.
  isGatewayInCLI = \
    globalConfig.getParameter("netborder.oamWebServer.isGatewayInCLI")


  extTab = TabPSTN( tabs=tabs, items=items, renderTo="pstn_tabs",
                    tabList=tabList, width=976, height=550,
                    activeTab=activeTab,
                    src_file=strFileLocation,
                    offerVirtualBoards=offerVirtualBoards,
                    isGatewayInCLI=isGatewayInCLI,
                    autoDetectHardware=autoDetectHardware,
                    children=[
                      status.getSubTabs(),
                      wizard.getWizardForm(),
                      config.getSubTabs()
                    ])

  @expose(template='pstnconfig.templates.index')
  def index(self):
    """Show the user the choice he have to edit the pstn-config.xml file."""

    gwServiceName = globalConfig.getParameter('netborder.oamWebServer.gwServiceName')
    gwVersion =globalConfig.getParameter('netborder.oamWebServer.gwVersion')
    
    # Check the product name to determine what should be written in
    # the footer of the master page
    footerText = \
        u"\xA9 %s Sangoma Technologies Corp." % datetime.date.today().year

    versionText = u"v%s" % (gwVersion)
    return dict(extTab = self.extTab, versionText = versionText, footerText = footerText  )

  @expose()
  def configFileExist(self, _dc=''):
    return {"configFileExist":model.PSTNConfigFileExists()}

  # Controller to redirect requests to the same host but different port
  #   including the remaining path and the query parameter of the original
  #   request (for use for ex. to redirect to the embedded gw web server on
  #   the same host, different port)
  # path is an array of path segments (not including the segement that
  #   triggers the call to this controller)
  # query_params is a dictionany of the query parameters
  # TODO re-useable in other projects
  def redirectToSameHostDifferentPort(self, port, *path, **query_params):

    # cherrypy.request holds the current request (different instance per thread)
    requestHeaders = cherrypy.request.headers
    # Host header is 'host [: port]' following HTTP 1.1
    assert requestHeaders.has_key("Host")
    hostAndPort = requestHeaders.get('Host')

    # Isolate host name, skip port if present
    import re
    # anything char but colon, until colon or end of string
    match = re.match('([^:]+):?', hostAndPort)
    assert match
    host = match.group(1) # first atom (what's behing parethesis in regexp)

    targeturl = '%s://%s:%d' % (cherrypy.request.scheme,
                                host,
                                port)

    # Build target URL path following remaining path segments
    for path_segment in path:
      targeturl += '/' + path_segment

    import urllib
    query_string = urllib.urlencode(query_params)
    targeturl += '?' + query_string

    log.info("Target URL for redirect is '%s'" % targeturl)

    raise cherrypy.HTTPRedirect(targeturl)

  # This method returns the value of a parameter in the Global Configuration
  @expose()
  def getGCParameterValue(self, paramName, _dc=''):
    return {"paramValue":globalConfig.getParameter(str(paramName))}

  @expose()
  def redirectToEmbeddedGWGui(self, *args, **kw):
    # HARDCODED port number
    EMBEDDED_GW_GUI_PORT = 7782
    return self.redirectToSameHostDifferentPort(EMBEDDED_GW_GUI_PORT,
                                                *args,
                                                **kw)

  # Used by static html file for startup
  @expose()
  def ping(self, *args, **kw):
    #set headers to block IE caching
    cherrypy.response.headers['Expires'] = '...'
    cherrypy.response.headers['Cache-Control'] = \
      'no-store, no-cache, must-revalidate, post-check=0, pre-check=0'
    cherrypy.response.headers['Pragma'] = 'no-cache'

    if kw.has_key('delayMs'):
      # Asked to delay response for a given ms
      # Will raise ValueError if cannot be converted to int
      delayMs = int(kw['delayMs'])
      import time
      time.sleep(delayMs / 1000.0) # Using float literal to keep decimals

    return 'pong'

  @expose()
  def releaseNotes(self):
    filePath = os.path.join(self.strInstallDir,"doc","Release_Notes.pdf")
    log.info("release notes path: %s" % filePath)
    return serve_file(filePath)

  @expose()
  def userGuide(self):
    filePath = \
      os.path.join(self.strInstallDir,"doc","TDM_Gateway_User_Guide.pdf")

    log.info("user guide path: %s" % filePath)
    return serve_file(filePath)

  @expose()
  def license(self):
    filePath = globalConfig.getParameter("netborder.LicenseFile")
    log.info("license file path: %s" % filePath)
    return serve_file(filePath)

  @expose()
  def default(self, *args, **kw):
    return 'Page not found'




