import os
import netborderPyInfra
import subprocess # used to call netborder-list-boards tool
import logging

# import pstnconfig autogenerated module contents into 'model' module
from paraxip.gateway_ui_model.pstnconfig import *

# import PstnConfigHelpers module contents into 'model' module
from paraxip.gateway_ui_model.PstnConfigHelpers import *

from netborderPyInfra import PyGlobalConfig as globalConfig

from turbogears import config

log = logging.getLogger("model")

# Public module attributes

# The current pstnConfig
currentConfig = None

# The configuration for the boards that were detected using
# netborder-list-boards tool
hardwareCfg = None

hardwareDetectionEnabled = False

strInstallDir = globalConfig.getParameter("netborder.Installation.Directory")
if strInstallDir is None:
  log.error("netborder.Installation.Directory parameter not set. Cannot "\
            "determine pstn configuration file location")

# The file location of thew PSTN config in XML format
strPstnConfigFileName = os.path.join(strInstallDir,"config/pstn-config.xml")

# Debug flag to allow validation of the config each time it is written out
# Turbogears bool configuration property "model.validateOnWrite" (default:False)
#validateOnWrite = config.get("netborder.oamWebServer.model.validateOnWrite",
#                              False)
validateOnWrite = \
  globalConfig.getParameter("netborder.oamWebServer.model.validateOnWrite")

###############################################################################
def getPSTNConfigFileName():
  return strPstnConfigFileName

def PSTNConfigFileExists ():
  log.debug("entering model.PSTNConfigFileExists")
  if os.path.isfile(getPSTNConfigFileName()):
    # PSTN config file already exists
    log.info("PSTN config file %s already exists", getPSTNConfigFileName() )
    log.debug("exiting model.PSTNConfigFileExists")
    return True
  else:
    log.info("PSTN config file %s does not already exist",
             getPSTNConfigFileName() )
    log.debug("exiting model.PSTNConfigFileExists")
    return False

###############################################################################
# Writes current config to file as XML at the default location
def writeCurrentConfigToFile():
  log.debug("entering model.writeCurrentConfigToFile")
  # Make a backup of current file
  # TODO make number of rolling backup configurable. Currently set to 5.
  try:
    bRv = netborderPyInfra.renameWithTimestamp(getPSTNConfigFileName(),5)
    if bRv == False:
      log.debug("Unable to make a backup copy of pstn "
                "configuration file (%s). This might be the file creation.",
                getPSTNConfigFileName())

    writeConfigToFile(currentConfig, getPSTNConfigFileName())
  except:
    log.error("Error: exception caught while making a backup of pstn "
              "configuration file %s ",getPSTNConfigFileName())
    log.debug("exiting model.writeCurrentConfigToFile")
    return

  log.debug("exiting model.writeCurrentConfigToFile")

###############################################################################
# Writes the config passed in parameter in the file specified by filename
def writeConfigToFile(config, filename):
  log.debug("entering model.writeConfigToFile")
  # 'w+b' is for write, truncate, binary
  # Now ready to overwrite file
  outfile = open(filename, 'w+b')
  writeConfigToFileAsXML(config, outfile)
  outfile.close()
  if validateOnWrite:
    if isConfigValid():
      log.info('PSTN config written to file "%s" is valid with respect '
        'to the XML schema' % filename)
    else:
      log.error('PSTN config written to file "%s" is NOT valid with respect '
        'to the XML schema, see paraxip log for details' %
        filename)

  log.debug("exiting model.writeConfigToFile")

def isConfigValid ():
  import netborderPyXML
  validator = netborderPyXML.XmlRelaxngValidator()
  strSchemaFile = os.path.join(strInstallDir, 'config/pstn-config.rng')
  validator.configure(strSchemaFile)
  return validator.validateXmlFile(getPSTNConfigFileName())

###############################################################################
# Validates the boards in the currentConfig with the boards detected with
# netborder-list-boards tools (hardwareCfg).
#
# Tries to match each board in the current config with a board detected using
# the board model, the pci bus number and the pci slot number. When a board is
# matched, we update the wanpipe numbers.
#
# This method is called only when the parameter
# oamWebServer.enableHardwareDetection is set to true
###############################################################################
def validateCurrentConfig():
  log.debug("entering model.validateCurrentConfig")
  # For boards in current config
  for cfgBoard in (currentConfig.get_boards().get_sangomaDigitalBoard() + \
      currentConfig.get_boards().get_sangomaAnalogBoard()):

    boardMatched = False

    # If board has no SLOT and BUS attributes
    pciBus = cfgBoard.get_pciBus()
    pciSlot = cfgBoard.get_pciSlot()
    boardModel = cfgBoard.get_model()

    if pciSlot is None and pciBus is None:
      log.warn("board [%s] was added manually. deleting board from config." \
                % cfgBoard.get_ID())
      deleteBoard(currentConfig,cfgBoard.get_ID())
      continue

    # For boards in hardwareCfg
    for board in (hardwareCfg.get_boards().get_sangomaDigitalBoard() + \
      hardwareCfg.get_boards().get_sangomaAnalogBoard()):

      # If board matches with board model, pci bus and pci slot number
      if boardModel == board.get_model() and pciBus == board.get_pciBus() and \
          pciSlot == board.get_pciSlot():
        log.debug("found matching board for board [%s]" % cfgBoard.get_ID())
        boardMatched = True

        # If Analog board, update wanpipe number
        if boardsModelInfo[boardModel]['type'] == "analog":
          cfgBoard.set_wanpipe(board.get_wanpipe())

        # Validate the wanpipe numbers for each interface
        if not validateWanpipeInterface(cfgBoard, board):
          log.warn("could not validate wanpipe. deleting board [%s]." % cfgBoard.get_ID())
          # Could happen if fxo and fxs modules were inverted.
          # delete board from the config
          deleteBoard(currentConfig,cfgBoard.get_ID())
    # for hardwareCfg boards

    # If board is not found
    if not boardMatched:
      log.warn( "could not find a matching board for board [%s], deleting board." % cfgBoard.get_ID())
      deleteBoard(currentConfig,cfgBoard.get_ID())

  # for currentConfig boards
  writeCurrentConfigToFile()
  log.debug("exiting model.validateCurrentConfig")

###############################################################################
# For each interface on the board, we reassign the wanpipe number or wanpipe
# interface number (digital or analog board) in the currentConfig with the
# value that was detected using netborder-list-boards tool (hardwareCfg).
#
# When validating interfaces of an anlog board, we also check that the
# interface type (FXO or FXS) has not been modified. It could happen if the
# user replaced a module on the card with another one not of the same type.
# If it is the case, we return False which will remove the board from the
# current configuration since it is not valid. The user may re-add the board by
# using the WebUI.
def validateWanpipeInterface(configBoard, hwBoard):
  log.debug("entering model.validateWanpipeInterface")
  isAnalog = ( boardsModelInfo[configBoard.get_model()]['type'] == "analog" )

  for configInterface in configBoard.get_interface():
    for hwInterface in hwBoard.get_interface():
      if configInterface.get_ID() == hwInterface.get_ID():
        if isAnalog:
          if not (configInterface.get_type() == hwInterface.get_type()):
            log.warn( "interface type switched from %s to %s." \
                      % (configInterface.get_type(),hwInterface.get_type()))
            return False

          configInterface.set_wanpipeInterface(
            hwInterface.get_wanpipeInterface()
          )

        else:
          configInterface.set_wanpipe(hwInterface.get_wanpipe())
    #for hwBoard
  #for configBoard

  log.debug("exiting model.validateWanpipeInterface")

  return True