//-----------------------------------------------
// Analog Physical Config form
// -----------------------------------------------
Ext.namespace("Paraxip.config.physConfig");

var frmAnalogPhysConf = new Ext.FormPanel({
  title         : "Analog Configuration",
  frame         : true,
  header        : false,
  id            : "frmAnalogPhysConf",
  autoScroll    : true,
  layout        : 'accordion',
  layoutConfig  : {
    activeItem  : 0
  },
  defaults      : {
    layout: 'form',
    style: 'text-align:left;',
    labelWidth: 150,
    bodyStyle: 'padding:5px;',
    width: 500,
    autoHeight: true,
    fill: true
  },
  buttonAlign   : "right",
  buttons       : [{
    id: "Paraxip.config.physConfig.assignedBoardsButton",
    text: 'Assigned board(s)',
    listeners: {'click':function(){
        configId = id_textAnalog.getValue();
        unassignedBoardsGrid.getStore().load({
           params: {'id':configId}
        });
        assignedBoardsGrid.getStore().load({
           params: {'id':configId}
        });
        assignBoardsWindow.show();
      }
    }
  },{
    id: "Paraxip.config.physConfig.saveAssignedBoardsButton",
    text: 'Save',
    listeners: {'click':saveConfig}
  },{
    id: "Paraxip.config.physConfig.cancelAsssignedBoardsButton",
    text: 'Cancel',
    listeners: {'click':physConfCancel}
  }]
});

var analogConfigPanel = new Ext.Panel({
  id    : 'analogConfigPanel',
  name  : 'config',
  title : 'Configuration Info',
  items : [{
    xtype: 'textfield',
    id: 'analogNodeId',
    name: 'nodeId',
    hidden: true,
    hideLabel: true
  }]
});

var analogPanel = new ParaxipPanelWithHelp({
  id    : 'analogPanel',
  name  : 'analog',
  title : 'Analog Configuration'
});

var analogFxoPanel = new ParaxipPanelWithHelp({
  id    : 'analogFxoPanel',
  name  : 'analogFxo',
  title : 'Analog FXO Configuration'
});

var analogFxsPanel = new ParaxipPanelWithHelp({
  id    : 'analogFxsPanel',
  name  : 'analogFxs',
  title : 'Analog FXS Configuration'
});

var audioAnalogPanel = new ParaxipPanelWithHelp({
  id    : 'audioAnalogPanel',
  name  : 'audio',
  title : 'Audio Configuration'
});

var echoAnalogPanel = new ParaxipPanelWithHelp({
  id    : 'echoAnalogPanel',
  name  : 'EchoCancellation',
  title : 'Echo Cancellation Configuration',
  helpText : "HardwareEchoCancellationEchoCancellation"
});

var acousticAnalogPanel = new ParaxipPanelWithHelp({
  id    : 'acousticAnalogPanel',
  name  : 'AcousticEchoCancellation',
  title : 'Acoustic Echo Cancellation Configuration',
  helpText : "HardwareAcousticEchoCancellationAcousticEchoCancellation"
});

var levelAnalogPanel = new ParaxipPanelWithHelp({
  id    : 'levelAnalogPanel',
  name  : 'LevelControl',
  title : 'Level Control Configuration',
  helpText : "HardwareLevelControlLevelControl"
});

var backgroundAnalogPanel = new ParaxipPanelWithHelp({
  id    : 'backgroundAnalogPanel',
  name  : 'BackgroundNoiseAttenuation',
  title : 'Background Noise Attenuation Configuration',
  helpText : "HardwareBackgroundNoiseRemovalBackgroundNoiseAttenuation"
});

var dtmfAnalogPanel = new ParaxipPanelWithHelp({
  id    : 'dtmfAnalogPanel',
  name  : 'DTMFRemoval',
  title : 'DTMF Removal Configuration',
  helpText : "HardwareDTMFRemovalDTMFRemoval"
});

var id_textAnalog= new Ext.form.TextField({
  id          : 'id_textAnalog',
  fieldLabel  : "ID",
  name        : "id",
  disabled    : true,
  width       : 162
});

var name_textAnalog = new Ext.form.TextField({
  id          : 'name_textAnalog',
  fieldLabel  : "Display Name",
  name        : "displayName",
  width       : 162
});

var analogVariant_select = new Ext.form.ComboBox({
  id              : 'variantAnalog_selection',
  name            : 'variant',
  fieldLabel      : "Analog Variant",
  store           : Paraxip.countryList,
  forceSelection  : true,
  triggerAction   : 'all',
  mode            : 'local',
  typeAhead       : true,
  selectOnFocus   : true,
  width           : 145,
  minListWidth    : 145
});

var analogClocking_select = new Ext.form.ComboBox({
  id              : 'analogClocking_selection'
  ,name           : 'clocking'
  ,fieldLabel     : "Analog Clocking"
  ,store          : ["EXTERNAL","INTERNAL"]
  ,forceSelection : true
  ,triggerAction  : 'all'
  ,mode           : 'local'
  ,typeAhead      : true
  ,selectOnFocus  : true
  ,width          : 145
  ,minListWidth   : 145
  ,helpText       : "analogClockingAttributeclocking"
});

var analogFxoRxGain_text = new Ext.form.NumberField({
  id             : 'analogFxoRxGain_text',
  fieldLabel     : "Receiver Gain (in dB)",
  name           : 'fxoRxGainIndB',
  helpText       : "analogFxoRxGainAttributefxoRxGainIndB",
  width          : 162,
  allowDecimals  : true,
  allowNegative  : true,
  decimalPrecision : 1,
  minValue       : -15.0,
  maxValue       : 12.0,
  allowBlank     : false
});

var analogFxoTxGain_text = new Ext.form.NumberField({
  id             : 'analogFxoTxGain_text',
  fieldLabel     : "Transmitter Gain (in dB)",
  name           : 'fxoTxGainIndB',
  helpText       : "analogFxoTxGainAttributefxoTxGainIndB",
  width          : 162,
  allowDecimals  : true,
  allowNegative  : true,
  decimalPrecision : 1,
  minValue       : -15.0,
  maxValue       : 12.0,
  allowBlank     : false
});

var analogFxsRxGain_select = new Ext.form.ComboBox({
  id             : 'analogFxsRxGain_select',
  fieldLabel     : "Receiver Gain (in dB)",
  name           : 'fxsRxGainIndB',
  helpText       : "analogFxsRxGainAttributefxsRxGainIndB",
  store          : [-3.5, 0.0, 3.5],
  forceSelection : true,
  triggerAction  : 'all',
  mode           : 'local',
  typeAhead      : true,
  selectOnFocus  : true,
  width          : 145,
  minListWidth   : 145
});

var analogFxsTxGain_select = new Ext.form.ComboBox({
  id             : 'analogFxsTxGain_select',
  fieldLabel     : "Transmitter Gain (in dB)",
  name           : 'fxsTxGainIndB',
  helpText       : "analogFxsTxGainAttributefxsTxGainIndB",
  store          : [-3.5, 0.0, 3.5],
  forceSelection : true,
  triggerAction  : 'all',
  mode           : 'local',
  typeAhead      : true,
  selectOnFocus  : true,
  width          : 145,
  minListWidth   : 145
});

var analogFxoBatteryDebounce_text = new Ext.form.NumberField({
  id             : 'analogFxoBatteryDebounce_text',
  fieldLabel     : "Battery Debounce Time (in milliseconds)",
  name           : 'fxoBatteryDebounceTimeInMs',
  helpText       : "analogFxoBatteryDebounceTimeAttributefxoBatteryDebounceTimeInMs",
  width          : 162,
  allowDecimals  : true,
  allowNegative  : false,
  minValue       : 0,
  maxValue       : 80,
  allowBlank     : false
});

var analogFxoBatteryThreshold_text = new Ext.form.NumberField({
  id             : 'analogFxoBatteryThreshold_text',
  fieldLabel     : "Battery Threshold (in volts)",
  name           : 'fxoBatteryThresholdInVolts',
  helpText       : "analogFxoBatteryThresholdAttributefxoBatteryThresholdInVolts",
  width          : 162,
  allowDecimals  : true,
  allowNegative  : false,
  minValue       : 1,
  maxValue       : 5,
  allowBlank     : false
});

function isFxoFakePolarityEnabled()
{
	if (this.id == "analogFxoFakePolarityEnabled_selection") {
		if (this.getValue() == "YES") {
		
			analogFxoFAkePolarityThreshold_text.enable();
		} else {
			 analogFxoFAkePolarityThreshold_text.disable();
			 analogFxoFAkePolarityThreshold_text.clearInvalid();
		}
	
	}
}

var analogFxoFAkePolarityThreshold_text = new Ext.form.NumberField({
  id             : 'analogFxoFAkePolarityThreshold_text',
  fieldLabel     : "Fake Polarity Threshold (in DBoV)",
  name           : 'fxoPolarityThresholdInDbov',
  helpText       : "analogFxoFakePolarityThresholdAttributefxoPolarityThresholdInDbov",
  width          : 162,
  allowDecimals  : false,
  allowNegative  : true,
  allowPositive  : false,
  minValue       : -18,
  maxValue       : 0,
  allowBlank     : false
});


var analogFxoFakePolarityEnabled_selection = new Ext.form.ComboBox({
  id              : 'analogFxoFakePolarityEnabled_selection',
  name            : 'analogFxoFakePolarityEnabled',
  helpText        : "analogFxoFakePolarityThresholdAttributeanalogFxoFakePolarityEnabled",
  fieldLabel      : "Enable Fake Polarity",
  store           : ['NO', 'YES'],
  forceSelection  : true,
  triggerAction   : 'all',
  mode            : 'local',
  typeAhead       : true,
  selectOnFocus   : true,
  width           : 145,
  minListWidth    : 145,
  listeners : { 
      'select':isFxoFakePolarityEnabled,
       scope  :analogFxoFakePolarityEnabled_selection
	}
  
});



var defaultPcmLawAnalog_select =  createPcmLawSelection({
  id : 'defaultPcmLawAnalog_selection'
});

var voicePacketLengthInMsAnalog_text = createVoicePacketText({
  id : 'voicePacketLengthInMsAnalog_text'
});

var optimalTxBufferLevelAnalog_text = createOptimalTxBufferLevelText({
  id : 'optimalTxBufferLevelAnalog_text'
});

var toneSetAttributesAnalog_select = createToneSetAtributesSelection({
  id : 'toneSetAttributesAnalog_select'
});

var ecAnalog_type_select = createEcTypeSelect({
  id : 'ecAnalog_type_selection'
});

var ecAnalog_mode_select = createEcModeSelect({
  id : 'ecAnalog_mode_selection'
});

var ecAnalog_comfortNoiseMode_select = createEcComfortNoiseSelect({
  id : 'ecAnalog_comfortNoiseMode_selection'
});

var ecAnalog_tailDisplacementInMs_text = createEcTailText({
  id : 'ecAnalog_tailDisplacementInMs_text'
});

var ecAnalog_doubleTalkBehavior_select = createEcDoubleTalkSelect({
  id : 'ecAnalog_doubleTalkBehavior_selection'
});

var aecAnalog_type_select = createAecTypeSelect({
  id : 'aecAnalog_type_selection'
});

var aecAnalog_mode_select = createAecModeSelect({
  id : 'aecAnalog_mode_selection'
});

var levelAnalog_type_select = createLevelTypeSelect({
  id : 'levelAnalog_type_selection'
});

var levelAnalog_direction_select = createLevelDirectionSelect({
  id : 'levelAnalog_direction_selection'
});

var levelAnalog_mode_select = createLevelModeSelect({
  id        : 'levelAnalog_mode_selection',
  listeners : {
    'select': levelControlVisibleFields,
    scope   : frmAnalogPhysConf
  }
});

var levelAnalog_targetLevelIndBM0_text = createLevelTargetText({
  id : 'levelAnalog_targetLevelIndBM0_text'
});

var levelAnalog_gainIndB_text = createLevelGainText({
  id : 'levelAnalog_gainIndB_text'
});

var bnaAnalog_type_select = createBnaTypeSelect({
  id : 'bnaAnalog_type_selection'
});

var bnaAnalog_direction_select = createBnaDirectionSelect({
  id  : 'bnaAnalog_direction_selection'
});

var bnaAnalog_mode_select = createBnaModeSelect({
  id : 'bnaAnalog_mode_selection'
});

var bnaAnalog_attenuationIndB_select = createBnaAttenuationSelect({
  id : 'bnaAnalog_attenuationIndB_selection'
});

var dtmfAnalog_type_select = createDtmfTypeSelect({
  id : 'dtmfAnalog_type_selection'
});

var dtmfAnalog_direction_select = createDtmfDirectionSelect({
  id : 'dtmfAnalog_direction_selection'
});

var dtmfAnalog_mode_select = createDtmfModeSelect({
  id : 'dtmfAnalog_mode_selection'
});

analogConfigPanel.add(id_textAnalog);
analogConfigPanel.add(name_textAnalog);
frmAnalogPhysConf.add(analogConfigPanel);

// !-! jpboily : those gain are not relevant after all. Keeping them here for
// now if for some reason we need to enable them easilly.
//analogFxoPanel.add(analogFxoRxGain_text);
//analogFxoPanel.add(analogFxoTxGain_text);
analogFxoPanel.add(analogFxoBatteryDebounce_text);
analogFxoPanel.add(analogFxoBatteryThreshold_text);
 
analogFxoPanel.add(analogFxoFakePolarityEnabled_selection);
analogFxoPanel.add(analogFxoFAkePolarityThreshold_text);
frmAnalogPhysConf.add(analogFxoPanel);

// !-! jpboily : those gain are not relevant after all. Keeping them here for
// now if for some reason we need to enable them easilly.
//analogFxsPanel.add(analogFxsRxGain_select);
//analogFxsPanel.add(analogFxsTxGain_select);
//frmAnalogPhysConf.add(analogFxsPanel);

analogPanel.add(analogVariant_select);
analogPanel.add(analogClocking_select);
frmAnalogPhysConf.add(analogPanel);

audioAnalogPanel.add(defaultPcmLawAnalog_select);
audioAnalogPanel.add(voicePacketLengthInMsAnalog_text);
audioAnalogPanel.add(toneSetAttributesAnalog_select);
audioAnalogPanel.add(optimalTxBufferLevelAnalog_text);
frmAnalogPhysConf.add(audioAnalogPanel);

echoAnalogPanel.add(ecAnalog_type_select);
echoAnalogPanel.add(ecAnalog_mode_select);
echoAnalogPanel.add(ecAnalog_comfortNoiseMode_select);
echoAnalogPanel.add(ecAnalog_tailDisplacementInMs_text);
echoAnalogPanel.add(ecAnalog_doubleTalkBehavior_select);
frmAnalogPhysConf.add(echoAnalogPanel);

acousticAnalogPanel.add(aecAnalog_type_select);
acousticAnalogPanel.add(aecAnalog_mode_select);
frmAnalogPhysConf.add(acousticAnalogPanel);

levelAnalogPanel.add(levelAnalog_type_select);
levelAnalogPanel.add(levelAnalog_direction_select);
levelAnalogPanel.add(levelAnalog_mode_select);
levelAnalogPanel.add(levelAnalog_targetLevelIndBM0_text);
levelAnalogPanel.add(levelAnalog_gainIndB_text);
frmAnalogPhysConf.add(levelAnalogPanel);

backgroundAnalogPanel.add(bnaAnalog_type_select);
backgroundAnalogPanel.add(bnaAnalog_direction_select);
backgroundAnalogPanel.add(bnaAnalog_mode_select);
backgroundAnalogPanel.add(bnaAnalog_attenuationIndB_select);
frmAnalogPhysConf.add(backgroundAnalogPanel);

dtmfAnalogPanel.add(dtmfAnalog_type_select);
dtmfAnalogPanel.add(dtmfAnalog_direction_select);
dtmfAnalogPanel.add(dtmfAnalog_mode_select);
frmAnalogPhysConf.add(dtmfAnalogPanel);

