//-----------------------------------------------
// Add/Delete CAS configuration methods
//-----------------------------------------------
var r2ID = '';

Ext.namespace("Paraxip.config.callControl");

var addCasConfHandler = function(){
  addCasConfWindow.show();
};

function addCasConfig(){
  var form = addCasConfigForm.getForm();
  var el  = form.getEl();

  el.dom.action = Paraxip.baseUrl + "call/addCasConfig";
  el.dom.method = 'POST';
  var casConfig = Ext.util.JSON.encode(form.getValues());
  form.submit({
    params: {'casConfig':casConfig},
    success: function(){
      Paraxip.isConfigOutOfSynch(true);
      addCasConfWindow.hide();
      gridCASConfiguration.store.reload();
      reloadTree(true);
    },
    failure: Paraxip.failureMessage
  });
};

var deleteCasConfHandler = function(){
  var selectedRecords = gridCASConfiguration.getSelectionModel().getSelections();

  if (selectedRecords !== []) {
    var casConfigs = new Array();
    for (var i = 0; i < selectedRecords.length; i++) {
      casConfigs.push(selectedRecords[i].data);
    }
    casConfigs = Ext.util.JSON.encode(casConfigs);
    Ext.Ajax.request({
      url: Paraxip.baseUrl + "call/deleteCasConfig",
      method: 'POST',
      disableCaching: false,
      params: {"casConfigs":casConfigs},
      success: function(result)
      {
        var deleteResponse = Ext.decode(result.responseText);
        if(deleteResponse.success){
          Paraxip.isConfigOutOfSynch(true);
          for (var index = 0; index < selectedRecords.length; index++) {
            gridCASConfiguration.store.remove(selectedRecords[index]);
          }
          reloadTree(true);
        }
        else{
          Ext.Msg.alert("Can not delete last configuration!",
          "At least one configuration is necessary for the system to work properly. " +
          "Please edit current configuration or create a new one before deleting " +
          "this one");
        }
      },
      failure: Paraxip.failureMessage
    });
  }
};

var casConfigChoice = new Ext.form.ComboBox({
  id              : 'casConfigChoice',
  fieldLabel      : "CAS Config Type",
  store           : ['R2'],
  forceSelection  : true,
  triggerAction   : 'all',
  mode            : 'local',
  typeAhead       : true,
  value           : 'R2',
  selectOnFocus   : true,
  width           : 100
//  disabled        : true
});



var addCasConfigForm = new Ext.FormPanel({
  id          : "addCasConfigForm",
  region      : 'center',
  labelWidth  : 175,
  labelPad    : 10,
  bodyStyle   : 'padding:10px;',
  frame       : false,
  header      : false,
  footer      : false,
  buttonAlign : "right",
  buttons     : [{
    id: "Paraxip.config.callControl.AddCasConfigButton",
    text: "Add",
    listeners: {'click':addCasConfig}
  },{
    id: "Paraxip.config.callControl.cancelAddCasConfigButton",
    text: "Cancel",
    listeners: {'click':function(){addCasConfWindow.hide();}}
  }]
});

addCasConfigForm.add(casConfigChoice);

var addCasConfWindow = new Ext.Window({
  id          : "addCasWindow",
  layout      : 'border',
  width       : 505,
  height      : 150,
  closeAction : 'hide',
  plain       : true
});

addCasConfWindow.add(addCasConfigForm);

var gridCASConfiguration = new Ext.grid.GridPanel({
  title       : 'CAS Configuration Grid',
  id          : 'gridCASConfiguration',
  autoScroll  : true,
  layout      : 'absolute',
  store       : new Ext.data.JsonStore({
    id:'storeCASConfiguration',
    url: Paraxip.baseUrl + "call/getGridCASConfigurationData",
    root: 'data',
    fields: [
      {name: 'id'},
      {name: 'displayName'},
      {name: 'nodeId'} //not displayed
    ]
  }),
  columns     : [
    {header: "ID", width: 35, sortable: true, dataIndex: 'id', css:'padding:5px;'},
    {header: "Name", width: 200, sortable: true, dataIndex: 'displayName', css:'padding:5px;'}
  ],
  viewConfig  : {forceFit:true}
});

gridCASConfiguration.addButton("Add CAS Config",addCasConfHandler,
                                gridCASConfiguration);
gridCASConfiguration.addButton("Remove CAS Config(s)",deleteCasConfHandler,
                                gridCASConfiguration);

gridCASConfiguration.on('rowdblclick', function(grid, rowIndex){
  var record = grid.store.getAt(rowIndex);
  goToNode(record.get('nodeId'));
});

var r2ID = '';

var frmCasR2 = new Ext.FormPanel({
  title         : "CAS R2 config",
  frame         : true,
  header        : false,
  id            : "frmCasR2",
  autoScroll    : true,
  layout        : 'accordion',
  layoutConfig  : {
    activeItem  : 0
  },
  defaults      : {
    layout: 'form',
    style: 'text-align:left;',
    labelWidth: 150,
    bodyStyle: 'padding:5px;',
    width: 500,
    autoHeight: true,
    fill: true
  },

  listeners       :{"actioncomplete":function(){
    //initialize the Advanced Protocol File Path Field
    var isEnable = advancedProtocolFile_text.getValue() != undefined && advancedProtocolFile_text.getValue().length > 0;
    enableAdvancedProtocoleFile.setValue(isEnable); 
    advancedProtocolFile_text.setDisabled(!isEnable);
  }},

  buttonAlign     : "right",
  buttons         : [{
    id: "Paraxip.config.callControl.assignedR2Button",
    text: 'Assigned Channel(s)',
    listeners: {
      'click':function(){
        r2ID = idR2_text.getValue();
        r2BChannelGrid.getStore().load({
          url: Paraxip.baseUrl + "call/getR2BChannels",
          params: {'id':r2ID}
        });
        storeFasBoards.load({
          url: Paraxip.baseUrl + "call/getBoards",
          params: {'id':briID}
        });
        storeFasChannelGroup.load({
          url: Paraxip.baseUrl + "call/getChannelGroups",
          params: {'id':briID}
        });
        assignedChannelR2Window.show();
      }
    }
  },{
    id: "Paraxip.config.callControl.saveR2Button",
    text: 'Save',
    listeners: {'click':saveR2Config}
  },{
    id: "Paraxip.config.callControl.cancelR2Button",
    text: 'Cancel',
    listeners: {'click':callControlCancel}
  }]
});
var R2Panel = new ParaxipPanelWithHelp({
  id    : 'R2Panel',
  name  : 'r2',
  title : 'R2 Configuration',
  helpText: "",
  items : [{
   xtype: 'textfield',
   name: 'nodeId',
   hidden: true,
   hideLabel: true
  }]
 });

function saveR2Config(e){
  var owner = e.ownerCt;
  var form = owner.getForm();
  var el = form.getEl();

  el.dom.action = Paraxip.baseUrl + "call/saveR2Config";
  el.dom.method = 'POST';

  var r2Config = owner.getItemsValueAsJson();

  form.submit({
   params:{
     'r2Config': r2Config
   },
   success: function(){
     Paraxip.isConfigOutOfSynch(true);
     reloadTree(true);
   },
   failure: Paraxip.failureMessage
  });
}


//FIELDS

var idR2_text = createIdText({
  id  : 'idR2_text'
});
var nameR2_text = createNameText({
  helpText: "casCallControlConfigcasdigitalR2ConfigdisplayName",
  id  : 'nameR2_text'
});
var switchVariantR2_select = new Ext.form.ComboBox( Ext.apply({
  id              : 'switchVariantR2_select'
  ,store           : ['Argentina',
                     'Brazil',
                     'Czech Republic',
                     'Colombia',
                     'Ecuador',
                     'Mexico',
                     'Philippines',
                     'Venezuela',
                     'ITU Standard']
  ,name            : 'switchVariant'
  ,fieldLabel      : "Country"
  ,helpText        : "casCallControlConfigcasdigitalR2ConfigswitchVariant"
  ,displayField    : 'variant_name'
  ,valueField      : 'value'
  ,hiddenName      : 'switchVariant'
  ,forceSelection  : true
  ,triggerAction   : 'all'
  ,mode            : 'local'
  ,typeAhead       : true
  ,selectOnFocus   : true
  ,width           : 145
  ,minListWidth    : 145
}));

var getAniFirst_select = new Ext.form.ComboBox({
  id                : 'getAniFirst_select'
  ,fieldLabel       : "Digits reception sequence"
  ,name             : 'get_ani_first'
  ,store            : [['TRUE','ANI, DNIS'],['FALSE','DNIS, ANI']]
  ,forceSelection   : true
  ,helpText         : "casCallControlConfigcasdigitalR2Configget_ani_first"
  ,triggerAction    : 'all'
  ,mode             : 'local'
  ,typeAhead        : true
  ,selectOnFocus    : true
  ,width            : 145
  ,minListWidth     : 145
  ,value            : 'NO'
});

var maxAni_numField = new Ext.form.NumberField( Ext.apply({
  id              : 'maxAni_numField',
  name            : "max_ani",
  fieldLabel      : "Max ANI digits",
  helpText        : "casCallControlConfigcasdigitalR2Configmax_ani",
  allowNegative   : false,
  allowBlank      : false,
  allowDecimal    : false,
  width           : 145,
  minListWidth    : 145
}));

var maxdnis_numField = new Ext.form.NumberField( Ext.apply({
  id              : 'maxdnis_numField',
  name            : "max_dnis",
  fieldLabel      : "Max DNIS digits",
  helpText        : "casCallControlConfigcasdigitalR2Configmax_dnis",
  allowNegative   : false,
  allowBlank      : false,
  allowDecimal    : false,
  width           : 145,
  minListWidth    : 145
}));

var category_select  = new Ext.form.ComboBox({
  id                : 'category_select'
  ,fieldLabel       : "Calling Party Category"
  ,name             : 'category'
  ,store            : ['National Subscriber',
                       'National Priority Subscriber',
                       'International Subscriber',
                       'International Priority Subscriber',
                       'Collect Call',
                       'Test Equipment']
  ,forceSelection   : true
  ,helpText         : "casCallControlConfigcasdigitalR2Configcategory"
  ,triggerAction    : 'all'
  ,mode             : 'local'
  ,typeAhead        : true
  ,selectOnFocus    : true
  ,width            : 145
  ,minListWidth     : 145
  ,value            : 'National Subscriber'
});

var skipCategory_select = new Ext.form.ComboBox({
  id                : 'skipCategory_select'
  ,fieldLabel       : "Skip Reception of Calling Party Category"
  ,name             : 'skip_category'
  ,store            : [['TRUE','YES'],['FALSE','NO']]
  ,forceSelection   : true
  ,helpText         : "casCallControlConfigcasdigitalR2Configskip_category"
  ,triggerAction    : 'all'
  ,mode             : 'local'
  ,typeAhead        : true
  ,selectOnFocus    : true
  ,width            : 145
  ,minListWidth     : 145
  ,value            : 'NO'
}); 

var allowCollectCalls_select= new Ext.form.ComboBox({
  id                : 'allowCollectCalls_select'
  ,fieldLabel       : "Collect calls"
  ,name             : 'allow_collect_calls'
  ,store            : [['TRUE','ALLOW'],['FALSE','REJECT']]
  ,forceSelection   : true
  ,helpText         : "casCallControlConfigcasdigitalR2Configallow_collect_calls"
  ,triggerAction    : 'all'
  ,mode             : 'local'
  ,typeAhead        : true
  ,selectOnFocus    : true
  ,width            : 145
  ,minListWidth     : 145
  ,value            : 'NO'
});

var chargeCalls_select = new Ext.form.ComboBox({
  id                : 'chargeCalls_select'
  ,fieldLabel       : "Notify calls as charged"
  ,name             : 'charge_calls'
  ,store            : [['TRUE','YES'],['FALSE','NO']]
  ,forceSelection   : true
  ,helpText         : "casCallControlConfigcasdigitalR2Configcharge_calls"
  ,triggerAction    : 'all'
  ,mode             : 'local'
  ,typeAhead        : true
  ,selectOnFocus    : true
  ,width            : 145
  ,minListWidth     : 145
  ,value            : 'NO'
});

var immediateAccept_select = new Ext.form.ComboBox({
  id                : 'immediateAccept_select'
  ,fieldLabel       : "Bypass Group B tones"
  ,name             : 'immediate_accept'
  ,store            : [['TRUE','YES'],['FALSE','NO']]
  ,forceSelection   : true
  ,helpText         : "casCallControlConfigcasdigitalR2Configimmediate_accept"
  ,triggerAction    : 'all'
  ,mode             : 'local'
  ,typeAhead        : true
  ,selectOnFocus    : true
  ,width            : 145
  ,minListWidth     : 145
  ,value            : 'NO'
});
var doubleAnswer_select = new Ext.form.ComboBox({
  id                : 'doubleAnswer_select'
  ,fieldLabel       : "Double Answer"
  ,name             : 'double_answer'
  ,store            : [['TRUE','YES'],['FALSE','NO']]
  ,forceSelection   : true
  ,helpText         : "casCallControlConfigcasdigitalR2Configdouble_answer"
  ,triggerAction    : 'all'
  ,mode             : 'local'
  ,typeAhead        : true
  ,selectOnFocus    : true
  ,width            : 145
  ,minListWidth     : 145
  ,value            : 'NO'
});
var forcedRelease_select = new Ext.form.ComboBox({
  id                : 'forcedRelease_select'
  ,fieldLabel       : "Send Forced Release signal"
  ,name             : 'forced_release'
  ,store            : [['TRUE','YES'],['FALSE','NO']]
  ,forceSelection   : true
  ,helpText         : "casCallControlConfigcasdigitalR2Configforced_release"
  ,triggerAction    : 'all'
  ,mode             : 'local'
  ,typeAhead        : true
  ,selectOnFocus    : true
  ,width            : 145
  ,minListWidth     : 145
  ,value            : 'NO'
});
var mfBackTimeout_numField = new Ext.form.NumberField( Ext.apply({
  id              : 'mfBackTimeout_numField',
  name            : "mfback_timeout",
  fieldLabel      : "MF Backward timeout (ms)",
  helpText        : "casCallControlConfigcasdigitalR2Configmfback_timeout",
  allowNegative   : false,
  allowBlank      : false,
  allowDecimal    : false,
  width           : 145,
  minListWidth    : 145
}));



var callFiles_select = new Ext.form.ComboBox({
  id                : 'callFiles_select'
  ,fieldLabel       : "Activate Call Logging"
  ,name             : 'call_files'
  ,store            : [['TRUE','YES'],['FALSE','NO']]
  ,forceSelection   : true
  ,helpText         : "casCallControlConfigcasdigitalR2Configcall_files"
  ,triggerAction    : 'all'
  ,mode             : 'local'
  ,typeAhead        : true
  ,selectOnFocus    : true
  ,width            : 145
  ,minListWidth     : 145
  ,value            : 'NO'
});

enableAdvancedProtocoleFile = new Ext.form.Checkbox({
  id           : "enableAdvancedProtocolFile"
  ,name        : "enableAdvancedProtocolFile"
  ,fieldLabel  : "Enable Advanced Protocol File Path"
  ,checked     : false
  ,listeners   : {
    'check':function() {
      advancedProtocolFile_text.setDisabled(!this.getValue());
    }
  }
});

var advancedProtocolFile_text = new Ext.form.TextField({
  id           : 'advancedProtocolFile_text'
  ,fieldLabel  : "Advanced Protocol File Path"
  ,name        : "advanced_protocol_file"
  ,helpText    : "casCallControlConfigcasdigitalR2Configadvanced_protocol_file"
  ,width       : 162
});

var infoConfig = new ParaxipPanelWithHelp({
  id    : 'infoConfig',
  name  : 'infoConfig',
  title : 'Configuration Info',
  helpText: "",
  items : [{
   xtype: 'textfield',
   name: 'nodeId',
   hidden: true,
   hideLabel: true
  }]
});

var inboundConfig = new ParaxipPanelWithHelp({
  id    : 'inboundConfig',
  name  : 'inboundConfig',
  title : 'R2 Inbound Configuration',
  helpText: "",
  items : [{
   xtype: 'textfield',
   name: 'nodeId',
   hidden: true,
   hideLabel: true
  }]
});
var outboundConfig = new ParaxipPanelWithHelp({
  id    : 'outboundConfig',
  name  : 'outboundConfig',
  title : 'R2 Outbound Configuration',
  helpText: "",
  items : [{
   xtype: 'textfield',
   name: 'nodeId',
   hidden: true,
   hideLabel: true
  }]
});
var bidirectional = new ParaxipPanelWithHelp({
  id    : 'bidirectional',
  name  : 'bidirectional',
  title : 'R2 Bidirectional Configuration',
  helpText: "",
  items : [{
   xtype: 'textfield',
   name: 'nodeId',
   hidden: true,
   hideLabel: true
  }]
});

infoConfig.add(idR2_text);
infoConfig.add(nameR2_text);

inboundConfig.add(getAniFirst_select);
inboundConfig.add(maxAni_numField);
inboundConfig.add(maxdnis_numField);
inboundConfig.add(skipCategory_select);
inboundConfig.add(allowCollectCalls_select);

inboundConfig.add(chargeCalls_select);
//Bug 6529: Removed field that has no effect on the gateway
//since not actually implemented in openr2.
//inboundConfig.add(immediateAccept_select);
inboundConfig.add(doubleAnswer_select);
inboundConfig.add(mfBackTimeout_numField);

outboundConfig.add(category_select);

bidirectional.add(switchVariantR2_select);
//bidirectional.add(forcedRelease_select);
bidirectional.add(callFiles_select);
//Bug 6539: Remove "Advanced Protocol" references from web UI 
//bidirectional.add(enableAdvancedProtocoleFile);
//bidirectional.add(advancedProtocolFile_text);

frmCasR2.add(infoConfig);
frmCasR2.add(inboundConfig);
frmCasR2.add(outboundConfig);
frmCasR2.add(bidirectional);

//ASSIGNED INTERFACES TO THE CAS CONFIG

var assignedChannelR2Window = new Ext.Window({
  title       : 'Assign Channels to the R2 Config',
  layout      : 'fit',
  layoutConfig: {
     autoWidth:false
  },
  width       : 600,
  height      : 500,
  closeAction : 'hide',
  resizable   : false,
  plain       : true,
  id          : "assignedChannelsR2Window"
});

var r2BChannelGrid = new Ext.grid.EditorGridPanel({
  title       : 'Channels assigned to config',
  id          : 'r2BChannelGrid',
  autoScroll  : true,
  width       : 550,
  height      : 400,
  enableColumnMove : false,
  selModel    : new Ext.grid.RowSelectionModel(),
  store       : new Ext.data.JsonStore({
    id:'storer2BChannelGrid',
    url: Paraxip.baseUrl + "call/getR2BChannels",
    root: 'data',
    fields: [
      {name: 'boardId'},
      {name: 'interfaceId'},
      {name: 'firstChannel'},
      {name: 'lastChannel'},
      {name: 'channelGroupID'},
      {name: 'nodeId'} //not displayed
    ]
  }),
  columns     : [
    {
      header: "Board ID",
      width: 110,
      sortable: true,
      dataIndex: 'boardId',
      renderer: renderEditableColumn,
      editor: new Ext.form.ComboBox({
        forceSelection  : true,
        triggerAction   : 'all',
        mode            : 'local',
        selectOnFocus   : true,
        valueField      : 'id',
        displayField    : 'id',
        store           : storeFasBoards,
        listClass       : 'x-combo-list-small'
      })
    },
    {
      header: "Interface ID",
      width: 110,
      sortable: true,
      dataIndex: 'interfaceId',
      renderer: renderEditableColumn,
      editor: new Ext.form.ComboBox({
        forceSelection  : true,
        triggerAction   : 'all',
        mode            : 'local',
        selectOnFocus   : true,
        valueField      : 'id',
        displayField    : 'id',
        store           : storeFasInterfaces,
        listClass       : 'x-combo-list-small'
      })
    },
    {
      header: "First Channel",
      width: 110,
      sortable: true,
      editable: false,
      dataIndex: 'firstChannel',
      renderer: renderEditableColumn,
      editor: new Ext.form.NumberField({
        allowNegative : false,
        allowBlank  : false
      })
    },
    {
      header: "Last Channel",
      width: 110,
      sortable: true,
      editable: false,
      dataIndex: 'lastChannel',
      renderer: renderEditableColumn,
      editor: new Ext.form.NumberField({
        allowNegative : false,
        allowBlank  : false
      })
    },
    {
      header: "Channel Group ID",
      width: 110,
      sortable: true,
      dataIndex: 'channelGroupID',
      renderer: renderEditableColumn,
      editor: new Ext.form.ComboBox({
        forceSelection  : true,
        triggerAction   : 'all',
        mode            : 'local',
        selectOnFocus   : true,
        valueField      : 'id',
        displayField    : 'id',
        store           : storeFasChannelGroup,
        listClass       : 'x-combo-list-small'
      })
    }
  ],
  listeners       : {
    'beforeedit':function(e){
      if( e.field == "interfaceId" ) {
        var colModel = e.grid.getColumnModel();
        var selectedRecord = e.grid.getSelectionModel().getSelected();
        var interfaceCellEditor = colModel.getCellEditor(e.column,e.row);
        interfaceCellEditor.field.store.load({
          url: Paraxip.baseUrl + "call/getInterfaces",
          disableCaching: false,
          params: {'id':selectedRecord.data.boardId}
        });
      }
    }
  },
  viewConfig:{forceFit:true}
});



r2BChannelGrid.addButton("Add Channel",addBchannel,r2BChannelGrid);
r2BChannelGrid.addButton("Delete",deleteBchannels,r2BChannelGrid);
r2BChannelGrid.addButton("Save",applyToFas,r2BChannelGrid);
r2BChannelGrid.addButton("Cancel",function(){assignedChannelR2Window.hide();});

assignedChannelR2Window.add(r2BChannelGrid);
