/*
 * Paraxip overrides to Ext JS Library 2.2
 * Copyright(c) 2008, Paraxip Technologies
 */

/*
  Ext.form.Field override to allow hiding the whole container, including
    the associated dynamic label. Use once layout is done for the field.
  From "timb" at http://extjs.com/forum/showthread.php?t=17396
*/

// turn on validation errors beside the field globally
Ext.form.Field.prototype.msgTarget = 'side';
Ext.form.Field.prototype.anchor = '60%';
// to enable quicktips
Ext.QuickTips.init();

Ext.override(Ext.form.Field, {

  showContainer: function() {
    this.enable();
    // show entire container and children (including label if applicable)
    this.getEl().up('.x-form-item').setDisplayed(true);
    this.show();
  },

  hideContainer: function() {
    this.disable(); // for validation
    // hide container and children (including label if applicable)
    this.getEl().up('.x-form-item').setDisplayed(false);
    this.hide();
  },

  setContainerVisible: function(visible) {
    if (visible) {
      this.showContainer();
    } else {
      this.hideContainer();
    }
    return this;
  },

  alignErrorIcon : function(){
    if (this.helpImage){
      this.errorIcon.alignTo(this.el, 'tl-tr', [20, 0]);
    }else{
      this.errorIcon.alignTo(this.el, 'tl-tr', [2, 0]);
    }
  },

  alignHelpImage : function(){
    this.helpImage.alignTo(this.el, 'tl-tr', [2, 0]);
  },

  afterRender : function() {
    var label = findLabel(this);

    if(label && (this.helpText ||this.helpUrl ))
    {
      var elp = this.el.findParent('.x-form-element', 5, true);
      this.helpImage = elp.createChild({
        style : "width:16px;height:18px;position:absolute;left:0;top:0;" +
                "display:block;background:transparent " +
                "url(/static/images/help.gif) no-repeat 0 2px;" +
                "cursor:pointer!important;"
      });
      this.alignHelpImage();
      this.on('resize', this.alignHelpImage, this);
      this.on('show', this.alignHelpImage, this);

      this.helpImage.on("click", function()
      {

        //this.ownerCt.add(helpWindow);
        helpWindow.setTitle(this.fieldLabel + " Help");
        helpWindow.show();

        if(this.helpText){
          helpTextArea.body.load({
            url: "/static/help/pstn-config/" +
                Paraxip.doc.pstnConfigNodeDocText[this.helpText]
                });
        }else{
          helpTextArea.body.load({
            url: "/static/help/" + this.helpUrl
                });
        }


      }, this);//scope
    }

    Ext.form.Field.superclass.afterRender.call(this);
    this.initEvents();
    this.initValue();
  },

  findParentBy: function(fn) {
    var p = this;
    do {
      p = p.ownerCt;
    } while (p !== null && !fn(p));
    return p;
  },

  findParentByType: function(xtype) {
    return typeof xtype == 'function' ?
      this.findParentBy(function(p) {
        return p.constructor === xtype;
      }) :
      this.findParentBy(function(p) {
        var type = p.constructor.xtype;
        return type === xtype;
      });
  }

});

// Override ComboBox for the help and error
Ext.override(Ext.form.ComboBox, {

  alignErrorIcon : function(){
    if (this.helpImage){
      this.errorIcon.alignTo(this.wrap, 'tl-tr', [20, 0]);
    }else{
      this.errorIcon.alignTo(this.wrap, 'tl-tr', [2, 0]);
    }
  },

  alignHelpImage : function(){
    this.helpImage.alignTo(this.wrap, 'tl-tr', [2, 0]);
  }
});

// Override afterRender method of FieldSet to be able to add a help image next
// to FieldSet's lable
Ext.override(Ext.form.FieldSet, {
  // Aligns the help image. User can specify an horizontal and vertical offset
  // to apply to the standard alignment.
  alignHelpImage : function(){
    this.helpImage.alignTo(
      this.el.child("legend").child(".x-fieldset-header-text"),
      'tl-tr',
      [2, -3]
    );
  },

  afterRender : function() {
    if( this.helpText ||this.helpUrl )
    {
      var child = this.el.child("legend");
      this.helpImage = child.createChild({
        style : "width:16px;height:18px;position:absolute;left:0;top:0;" +
                "display:block;background:transparent " +
                "url(/static/images/help.gif) no-repeat 0 2px;" +
                "cursor:pointer!important;"
      });
      this.alignHelpImage();
      this.on('resize', this.alignHelpImage, this);
      this.on('show', this.alignHelpImage, this);
      this.on('expand', this.alignHelpImage, this);
      this.on('collapse', this.alignHelpImage, this);

      this.helpImage.on("click", function()
      {

        //this.ownerCt.add(helpWindow);
        helpWindow.setTitle(this.title + " Help");
        helpWindow.show();

        if(this.helpText){
          helpTextArea.body.load({
            url: "/static/help/pstn-config/" +
                Paraxip.doc.pstnConfigNodeDocText[this.helpText]
                });
        }else{
          helpTextArea.body.load({
            url: "/static/help/" + this.helpUrl
                });
        }


      }, this);//scope
    }
    Ext.form.FieldSet.superclass.afterRender.call(this);

  }
});

// Override the data Store to allow an automatic refresh.
// Useful to automatically refresh grids.
Ext.override(Ext.data.Store, {
  startAutoRefresh : function(interval, params, callback, refreshNow){
    if(refreshNow){
      this.load({params:params, callback:callback});
    }
    if(this.autoRefreshProcId){
      clearInterval(this.autoRefreshProcId);
    }
    this.autoRefreshProcId =
      setInterval(
        this.load.createDelegate(
          this, [{params:params, callback:callback}]
        )
        , interval*1000
      );
  },
  stopAutoRefresh : function(){
    if(this.autoRefreshProcId){
      clearInterval(this.autoRefreshProcId);
    }
  }
});


// this fixes the problem of the static load from an ajax source
// see http://extjs.com/forum/showthread.php?t=34902

Ext.override( Ext.tree.TreeLoader, {
  processResponse : function(response, node, callback){
    var json = response.responseText;
    try {
      var o = eval('('+json+')');
      node.beginUpdate();
      for(var i = 0, len = o.length; i < len; i++){
        var n = this.createNode(o[i]);
        if(n){
          node.appendChild(n);
          if(this.preloadChildren){
            this.doPreload(n);
          }
        }
      }
      node.endUpdate();
      if(typeof callback == 'function'){
        callback(this, node);
      }
    }catch(e){
      this.handleFailure(response);
    }
  }
});

/* Ext.Container
 *  getItemsValueAsJson()
 *    Recurse thru "items" and return their names and native values as JSON, ex.
        {digitalT1: {framing: 'ESF',
                     lineEncoding: 'AMI'},
         audio: {echo: true,
                 tap: 2}}
 *
 */
Ext.namespace('Paraxip'); // Ensure Paraxip namespace is defined
Paraxip.getItemsValue = function(items) {
  var itemsValues = {};
  items.each(function(item) {
    if(item.items === undefined) {
      // Simple item (field): getValue for Ext.form.Field subclasses return
      //  native object (int, float, etc.)
      if( !item.disabled && item.getXType() != 'label' ) {
        //----------------------------------------------------------------------
        // !+! jpboily - Bug 2628: by forcing the panel to render, it will also
        // render is children. We need this to be sure that the checkbox group
        // for the disconnect detectors is rendered when we SAVE the config.
        // Otherwise, all detectors will be set to false, loosing the current
        // configuration.
        if( item.getXType() == "checkbox" ) {
          itemsValues[item.name] = item.checked;
        } else {
          itemsValues[item.name] = item.getValue();
        }
        // end of bug 2491
        //----------------------------------------------------------------------
      }
    } else {
      // The item itself contains items - recursive call
      itemsValues[item.name] = Paraxip.getItemsValue(item.items);
    }
  });
  return itemsValues;
};

Ext.override(Ext.Container, {
  getItemsValueAsJson: function() {
    return Ext.util.JSON.encode(Paraxip.getItemsValue(this.items));
  }
});

var helpWindow = new Ext.Window({
  id          : "helpWindow",
  layout      : 'border',
  width       : 505,
  height      : 300,
  closeAction : 'hide',
  plain       : true
});

/*
 * Field Override for text files
 *
 * */
var helpTextArea= new Ext.Panel({
  id          : 'helpText_area',
  region      : 'center',
  grow        : true,
  collapsible : false,
  bodyStyle   : 'padding: 10px',
  autoScroll  : true
});

helpWindow.add(helpTextArea);

var findLabel = function(field) {
  var wrapDiv = null;
  var label = null;

  //find form-item and label
  wrapDiv = field.getEl().up('div.x-form-item');
  if(wrapDiv)
  {
    label = wrapDiv.child('label');
  }
  if(label) {
    return label;
  }
};

// Panel with help button that shows contextual help
Ext.override(Ext.Panel, {
  constructor: function(config) {
  // Add help button if required
  if( config.helpText ||
      config.helpUrl){
    config = Ext.apply({
      tools: [{
        id:'help',
        qtip: "Get help",
        hidden : false,
        handler: function(event, toolEl, panel){
          helpWindow.setTitle(panel.title + " Help");
          helpWindow.show();
          if(config.helpText){
            helpTextArea.body.load({
              url: "/static/help/pstn-config/" +
                  Paraxip.doc.pstnConfigNodeDocText[config.helpText]
            });
          }else{
            helpTextArea.body.load({
              url: "/static/help/" +config.helpUrl});
          }
        }
      }]
    }, config);
  }
    Ext.Panel.superclass.constructor.call(this, config);
  }
});

ParaxipPanelWithHelp =
Ext.extend(Ext.Panel,{
 constructor: function(config) {
  // Add help button if required
  if( config.helpText ||
      config.helpUrl){
    config = Ext.apply({
      tools: [{
        id:'help',
        qtip: "Get help",
        hidden : false,
        handler: function(event, toolEl, panel){
          helpWindow.setTitle(panel.title + " Help");
          helpWindow.show();
          if(config.helpText){
            helpTextArea.body.load({
              url: "/static/help/pstn-config/" +
                  Paraxip.doc.pstnConfigNodeDocText[config.helpText]
            });
          }else{
            helpTextArea.body.load({
              url: "/static/help/" +config.helpUrl});
          }
        }
      }]
    }, config);
  }
  ParaxipPanelWithHelp.superclass.constructor.call(this, config);
 }
});


ParaxipEditorGridPanelWithHelp
= Ext.extend(Ext.grid.EditorGridPanel,{
 constructor: function(config) {
  // Add help button if required
  if( config.helpText ||
      config.helpUrl){
    config = Ext.apply({
      tools: [{
        id:'help',
        qtip: "Get help",
        hidden : false,
        handler: function(event, toolEl, panel){
          helpWindow.setTitle(panel.title + " Help");
          helpWindow.show();
          if(config.helpText){
            helpTextArea.body.load({
              url: "/static/help/pstn-config/" +
                  Paraxip.doc.pstnConfigNodeDocText[config.helpText]
            });
          }else{
            helpTextArea.body.load({
              url: "/static/help/" +config.helpUrl});
          }
        }
      }]
    }, config);
  }
 ParaxipEditorGridPanelWithHelp.superclass.constructor.call(this, config);
 }
});

ParaxipFormPanelWithHelp
= Ext.extend(Ext.FormPanel,{
 constructor: function(config) {
  // Add help button if required
  if( config.helpText ||
      config.helpUrl){
    config = Ext.apply({
      tools: [{
        id:'help',
        qtip: "Get help",
        hidden : false,
        handler: function(event, toolEl, panel){
          helpWindow.setTitle(panel.title + " Help");
          helpWindow.show();
          if(config.helpText){
            helpTextArea.body.load({
              url: "/static/help/pstn-config/" +
                  Paraxip.doc.pstnConfigNodeDocText[config.helpText]
            });
          }else{
            helpTextArea.body.load({
              url: "/static/help/" +config.helpUrl});
          }
        }
      }]
    }, config);
  }
  ParaxipFormPanelWithHelp.superclass.constructor.call(this, config);
 }
});

// It is imposible to a form to set a value to an input into a
// checkboxgroup or radiogroup, using a record + a reader
// + calling load() method.
// I have seen this is because checkboxgroups (and radiogroups
//as it is a subclass), for a FormPanel, are considered as a single
//field, so when it tries to collect all the fields to manage,
//it doesn't go inside to collect the checkboxes or radios they keep.
//
//Other problem is the checkboxgroup initilize the objects inside
//at rendering not at own initialization (in fact there is no own
//initilization, it inheritate the form.Field initComponent)
//
// Source: http://www.extjs.com/forum/showthread.php?t=47243
Ext.form.CheckboxGroup.prototype.initComponent = function(){
  Ext.form.CheckboxGroup.superclass.initComponent.call(this);
  var panelCfg = {
    cls: this.groupCls,
    layout: 'column',
    border: false
  };
  var colCfg = {
    defaultType: this.defaultType,
    layout: 'form',
    border: false,
    defaults: {
      hideLabel: true,
      anchor: '100%'
    }
  };

  if(this.items[0].items){
    Ext.apply(panelCfg, {
      layoutConfig: {columns: this.items.length},
      defaults: this.defaults,
      items: this.items
    });
    for(var i=0, len=this.items.length; i<len; i++){
      Ext.applyIf(this.items[i], colCfg);
    };
  }else{
    var numCols, cols = [];

    if(typeof this.columns == 'string'){
      this.columns = this.items.length;
    }
    if(!Ext.isArray(this.columns)){
      var cs = [];
      for(var i=0; i<this.columns; i++){
        cs.push((100/this.columns)*.01);
      };
      this.columns = cs;
    }

    numCols = this.columns.length;

    for(var i=0; i<numCols; i++){
      var cc = Ext.apply({items:[]}, colCfg);
      cc[this.columns[i] <= 1 ? 'columnWidth' : 'width'] = this.columns[i];
      if(this.defaults){
        cc.defaults = Ext.apply(cc.defaults || {}, this.defaults);
      }
      cols.push(cc);
    };

    if(this.vertical){
    var rows = Math.ceil(this.items.length / numCols), ri = 0;
      for(var i=0, len=this.items.length; i<len; i++){
        if(i>0 && i%rows==0){
          ri++;
        }
        if(this.items[i].fieldLabel){
          this.items[i].hideLabel = false;
        }
        cols[ri].items.push(this.items[i]);
      };
    }else{
      for(var i=0, len=this.items.length; i<len; i++){
        var ci = i % numCols;
        if(this.items[i].fieldLabel){
          this.items[i].hideLabel = false;
        }
        cols[ci].items.push(this.items[i]);
      };
    }

    Ext.apply(panelCfg, {
      layoutConfig: {columns: numCols},
      items: cols
    });
  }

  this.panel = new Ext.Panel(panelCfg);

  if(this.forId && this.itemCls){
    var l = this.el.up(this.itemCls).child('label', true);
    if(l){
      l.setAttribute('htmlFor', this.forId);
    }
  }

  var fields = this.panel.findBy(function(c){
    return c.isFormField;
  }, this);

  this.items = new Ext.util.MixedCollection();
  this.items.addAll(fields);
};

Ext.form.CheckboxGroup.prototype.onRender = function(ct, position){
  if(!this.el){
    this.panel.render(ct, position);
    this.el = this.panel.getEl();
  }
  Ext.form.CheckboxGroup.superclass.onRender.call(this, ct, position);
};

Ext.FormPanel.prototype.initFields = function(){
  var f = this.form;
  var formPanel = this;
  var fn = function(c){
    if(c.isFormField){
      if (!c.items){
        f.add(c);
      }else{
        c.items.each(fn);
      }
    }else if(c.doLayout && c != formPanel){
      Ext.applyIf(c, {
        labelAlign: c.ownerCt.labelAlign,
                      labelWidth: c.ownerCt.labelWidth,
                      itemCls: c.ownerCt.itemCls
                      });
      if(c.items){
        c.items.each(fn);
      }
    }
  };
  this.items.each(fn);
};

// The initFields function failed when the checkbox is not renderred
// prior to calling setValue
// source: http://www.extjs.com/forum/showthread.php?p=204841#post204841
Ext.form.Checkbox.override({
  setValue : function(v) {
    var checked = this.checked;
    this.checked = (v === true || v === 'true' || v == '1'
                    || String(v).toLowerCase() == 'on');

    if(this.rendered){
      this.el.dom.checked = this.checked;
      this.el.dom.defaultChecked = this.checked;
      this.wrap[this.checked? 'addClass' : 'removeClass'](this.checkedCls);
    }

    if(checked != this.checked){
      this.fireEvent("check", this, this.checked);
      if(this.handler){
        this.handler.call(this.scope || this, this, this.checked);
      }
    }
  },
  afterRender : function(){
    Ext.form.Checkbox.superclass.afterRender.call(this);
    this.wrap[this.checked? 'addClass' : 'removeClass'](this.checkedCls);
  }
});

Ext.grid.ParaxipCheckboxSelectionModel = Ext.extend(Ext.grid.CheckboxSelectionModel, {
  constructor : function(){
    Ext.grid.CheckboxSelectionModel.superclass.constructor.apply(this, arguments);

    if(this.checkOnly){
        this.handleMouseDown = this.handleMouseDownCheckOnly;
    }
  },

  // private
  handleMouseDownCheckOnly : function(g, rowIndex, e){
      if(e.button !== 0 || this.isLocked()){
          return;
      }
      var view = this.grid.getView();
      view.focusRow(rowIndex); // !important
  }
});
// Override EditorGridPanel to edit cell on first mouse click
Ext.override(Ext.grid.EditorGridPanel, {
  clicksToEdit: 1
});

// Override CheckboxSelectionModel so header cell follows the selection of
// individual rows so it is more user friendly. Header cell selected only when
// all rows are selected and un-selected when any rows are not selected.
Ext.apply(Ext.grid.CheckboxSelectionModel.prototype, {

  onMouseDown: function(e, t) {

    // If using left mouse button and simpleSelect or checkbox was clicked
    if(e.button === 0 && (t.className == 'x-grid3-row-checker' )) {
      e.stopEvent();

      var row = e.getTarget('.x-grid3-row');
      if(row) {
        var index = row.rowIndex;
        if(this.isSelected(index)) {
          this.deselectRow(index);

          // Deselect checkbox header if selected
          var hd = Ext.DomQuery.selectNode('.x-grid3-hd-checker', this.grid.view.innerHd);
          if(hd)
            Ext.fly(hd).removeClass('x-grid3-hd-checker-on');
        } else {
          this.selectRow(index, true);
          // Select checkbox header if all rows are selected
          if (this.grid.getStore().getRange().length == this.getCount()) {
            var hd = Ext.DomQuery.selectNode('.x-grid3-hd-checker', this.grid.view.innerHd);
            if(hd)
                Ext.fly(hd).addClass('x-grid3-hd-checker-on');
          }
        }
      }
    }
  }
});

function renderEditableColumn(val){
  return val;
  //return val + 
  //  ' <span style="font-size:xx-small;color:grey">[editable]</span>';
}
