//-----------------------------------------------
// SIP form
//-----------------------------------------------
Ext.namespace("Paraxip.wizard");

var labelWidth = 275;

var sip_form = new Ext.Panel({
  title           : "SIP Configuration"
  ,layout         : 'form'
  ,autoWidth      : true
  ,autoHeigth     : true
  ,autoScroll     : true
  ,frame          : true
  ,standardSubmit : true
  ,header         : false
  ,id             : "Paraxip.wizard.sip_form"
  ,labelWidth     : labelWidth
  ,bodyStyle      : "text-align:left;padding-left:15px;"
  ,defaults       : {
    anchor: '65%'
  }
  ,items: [{}]
});

// Static widths
var transportPortWidth = 290;
var comboSelectWidth = 17;
var helpImageWidth = 20;
var sipUrlsPadding = 167;

var SipDescriptionLabel = new Ext.form.Label({
  id      : "Paraxip.wizard.SipDescriptionLabel"
  ,text   : "Select the SIP configuration"
  ,style  : 'font-size:120%;font-weight:bold; line-height:60px;'
});

Paraxip.wizard.enableUdp = new Ext.form.Checkbox({
  id          : "Paraxip.wizard.enableUdp"
  ,name        : "enableUdp"
  ,checked     : true
  ,listeners   : {
    'check':function() {
      Paraxip.wizard.udpPort.setDisabled(!this.getValue());
      if (this.getValue()) {
        Paraxip.wizard.enableTls.setValue(false);
      }
    }
  }
});

Paraxip.wizard.udpPort = new Ext.form.NumberField({
  id      : "Paraxip.wizard.udpPort"
  ,name   : "udpPort"
  ,width  : transportPortWidth
});

Paraxip.wizard.enableTcp = new Ext.form.Checkbox({
  id          : "Paraxip.wizard.enableTcp"
  ,name       : "enableTcp"
  ,checked    : true
  ,listeners  : {
    'check':function() {
      Paraxip.wizard.tcpPort.setDisabled(!this.getValue());
      if (this.getValue()) {
        Paraxip.wizard.enableTls.setValue(false);
      }
    }
  }
});

Paraxip.wizard.tcpPort = new Ext.form.NumberField({
  id      : "Paraxip.wizard.tcpPort"
  ,name   : "tcpPort"
  ,width  : transportPortWidth
});

Paraxip.wizard.enableTls = new Ext.form.Checkbox({
  id          : "Paraxip.wizard.enableTls"
  ,name       : "enableTls"
  ,checked    : false
  ,listeners  : {
    'check':function() {
      Paraxip.wizard.tlsPort.setDisabled(!this.getValue());
      if (this.getValue()) {
        Paraxip.wizard.udpPort.setDisabled(true);
        Paraxip.wizard.tcpPort.setDisabled(true);
        Paraxip.wizard.enableUdp.setValue(false);
        Paraxip.wizard.enableTcp.setValue(false);
      }
    }
  }
});

Paraxip.wizard.tlsPort = new Ext.form.NumberField({
  id        : "Paraxip.wizard.tlsPort"
  ,name     : "tlsPort"
  ,width    : transportPortWidth
  ,disabled : true
});

var transportLabelStyle = "padding-top:5px;padding-bottom:5px;" +
                          "padding-left:15px;padding-right:56px;";

Paraxip.wizard.transport = new Ext.form.FieldSet({
  id            : "Paraxip.wizard.transport"
  ,title        : "Transport"
  ,autoHeight   : true
  ,defaults     : {anchor: '-20'}
  ,layout       : 'table'
  ,layoutConfig : {
    columns: 3
  }
  ,defaults     : {
    style       : "font-size:120%;"
  }
  ,items        : [
    Paraxip.wizard.enableUdp
    ,{
      layout      :'form'
      ,bodyStyle  : transportLabelStyle
      ,items      :
      {
        xtype   : 'label'
        ,id     : 'Paraxip.wizard.enableUdpLabel'
        ,text   : " Enable UDP on local listening port: "
      }
    }
    ,Paraxip.wizard.udpPort
    ,// Line 2
    Paraxip.wizard.enableTcp
    ,{
      layout      :'form'
      ,bodyStyle  : transportLabelStyle
      ,items      :
      {
        xtype   : 'label'
        ,id     : 'Paraxip.wizard.enableTcpLabel'
        ,text   : " Enable TCP on local listening port: "
      }
    }
    ,Paraxip.wizard.tcpPort
    ,// Line 3
    Paraxip.wizard.enableTls
    ,{
      layout      :'form'
      ,bodyStyle   : transportLabelStyle
      ,items:
      {
        xtype   : 'label'
        ,id     : 'Paraxip.wizard.enableTlsLabel'
        ,text   : " Enable TLS on local listening port: "
      }
    }
    ,Paraxip.wizard.tlsPort
  ]
});

var ipAddressStore = new Ext.data.JsonStore({
  id          : "ipAddressStore"
  ,url        : Paraxip.baseUrl + "wizard/getIPAddress"
  ,root       : "data"
  ,baseParams : {'js':'used_to_force_post'}
  ,autoLoad   : true
  ,fields     : [{name:"ip",type:"string"}]
});

var primaryIPAddress = new Ext.form.ComboBox({
  id              : 'primaryIPAddress'
  ,fieldLabel     : "Primary IP Address"
  ,store          : ipAddressStore
  ,valueField     : 'ip'
  ,displayField   : 'ip'
  ,forceSelection : true
  ,triggerAction  : 'all'
  ,mode           : 'local'
  ,typeAhead      : true
  ,selectOnFocus  : true
  ,allowBlank     : false
  ,style          : ".x-form-text {margin: 0px 0;}"
  ,helpUrl        : "quick-setup/qs-sip-primary-ip-address.html"
});

var srtpModeBeforeSelect = 'Support encryption';

Paraxip.wizard.srtpMode = new Ext.form.ComboBox({
  id              : "Paraxip.wizard.srtpMode"
  ,name           : "srtpMode"
  ,fieldLabel     : "Media encryption level"
  ,store          : [
    'Support encryption',
    'Require Encryption',
    'Do not support encryption'
  ]
  ,valueField     : 'srtp'
  ,displayField   : 'srtp'
  ,forceSelection : true
  ,triggerAction  : 'all'
  ,mode           : 'local'
  ,typeAhead      : true
  ,selectOnFocus  : true
  ,style          : ".x-form-text {margin: 0px 0;}"
  ,helpUrl        : "quick-setup/qs-media-encryption-level.html"
  ,listeners      : {
    'beforeselect': function(){
      srtpModeBeforeSelect = this.getValue();
    }
    ,'select': function(){
      if( this.getValue() == 'Require Encryption'
            && !Paraxip.wizard.enableTls.getValue()) {
        Ext.Msg.alert(
          "Invalid SIP Transport",
          "For security reasons it is not possible to select Require " +
          "Encryption when SIP Transport is not TLS"
        );
        this.setValue(srtpModeBeforeSelect);
      }
    }
    ,'render': function(){
      this.setValue('Support encryption');

    }
  }
});



Paraxip.wizard.sipMiscellaneous = new Ext.form.FieldSet({
  id          : "Paraxip.wiza primaryIPAddressrd.sipMiscellaneous"
  ,title      : "Sip Miscellaneous"
  ,autoHeight : true
  ,layout     : 'form'
  ,defaults   : {anchor: '-20'}
  ,items      : [
     primaryIPAddress
    ,Paraxip.wizard.srtpMode
  ]
});

      

sip_form.add(SipDescriptionLabel);
sip_form.add(Paraxip.wizard.transport);
sip_form.add(Paraxip.wizard.sipMiscellaneous);
